/*
 * Copyright (C) 2002,2003 Daniel Heck
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *  
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License along
 * with this program; if not, write to the Free Software Foundation, Inc.,
 * 59 Temple Place, Suite 330, Boston, MA 02111-1307, USA.
 *
 * $Id: d_models.cc,v 1.2 2003/01/19 17:28:33 dheck Exp $
 */
#include "d_models.hh"

// Functions in display.cc
namespace display
{
    void add_active_model(Model *m);
    void remove_active_model(Model *m);
}


//======================================================================
// DEFINING AND CREATING MODELS
//======================================================================

namespace
{
    SurfaceCache       surface_cache;
    SurfaceCache_Alpha surface_cache_alpha;

    px::Dict<Model*>   model_templates(1069);
    vector<Surface *>  image_pile;
}

//----------------------------------------------------------------------
// MODELS
//----------------------------------------------------------------------

void
display::InitModels()
{
    lua::Dofile("models-2d.lua");
    surface_cache_alpha.clear();

    enigma::Log << "# models: " << model_templates.size() << endl;
    enigma::Log << "# images: " << surface_cache.size() << endl;
}

void
display::ShutdownModels()
{
    delete_map(model_templates.begin(),
               model_templates.end());
    delete_sequence(image_pile.begin(), image_pile.end());
}


void 
display::ReloadModels()
{
    NewWorld(0,0);
    surface_cache.clear();
    surface_cache_alpha.clear();
    delete_map(model_templates.begin(),
               model_templates.end());
    model_templates.clear();
    lua::Dofile("models-2d.lua");
}

/* Register a new model template `m' under the name `name'. */
void 
display::DefineModel (const char *name, Model *m)
{
    string n(name);
    if (model_templates.has_key(n))
    {
        enigma::Log << "Redefinition of model '"<<name<<"'\n";
        delete m;
    } 
    else
    {
        model_templates.insert(n, m);
    }
}

/* Create a new model of type `name' by cloning the corresponding
   entry in model_templates. 

   This implementation remembers the most recently cloned object. This
   helps performance when filling big levels with identical floor
   tiles.
*/
Model * 
display::MakeModel (const string &name)
{
    static Model *last_templ=0;
    static string last_name;

    if (name!=last_name) {
        px::Dict<Model*>::iterator i;
        
        i = model_templates.find(name);
        if (i != model_templates.end()) {
            last_templ = i->second;
            last_name = name;
        }
    }

    Model *m = 0;
    if (last_templ)
        m = last_templ->clone();
    if (!m)
        fprintf(stderr, "Unknown model '%s'\n", name.c_str());
    return m;
}


void 
display::DefineImage(const char *name, const char *fname,
                     int xoff, int yoff)
{
    if (px::Surface *sfc = surface_cache.get(fname))
        DefineModel(name, new ImageModel(sfc, xoff, yoff));
}

void 
display::DefineSubImage(const char *name, const char *fname,
                        int xoff, int yoff, px::Rect subrect)
{
    if (px::Surface *sfc = surface_cache.get(fname))
        DefineModel(name, new ImageModel(sfc, subrect, xoff, yoff));
}

void
display::DefineRandModel(const char *name, int n, const char **names)
{
    RandomModel *m = new RandomModel();
    for (int i=0; i<n; i++) 
        m->add_model(names[i]);
    DefineModel(name, m);
}

void
display::DefineShadedModel (
    const char *name, 
    const char *model, 
    const char *shade)
{
    DefineModel(name, new ShadedModel(MakeModel(model),
                                       MakeModel(shade)));
}


/* Create an image by overlaying several other images.  The first entry in
   `images' is the name of the background image, the following images are
   drawn on top of it. */
void
display::DefineOverlayImage (const char *name, int n, const char **images)
{
    Surface *sfc = Duplicate(surface_cache.get(images[0]));
    if (sfc) // Surface *sfc = enigma::LoadImage(images[0]))
    {
	for (int i=1; i<n; i++) 
	    sfc->blit(0,0, surface_cache_alpha.get(images[i]));
	DefineModel(name, new ImageModel(sfc, 0,0));
        image_pile.push_back(sfc); // make sure it gets destructed
    }
}

void
display::DefineComposite (const char *name, 
                          const char *bgname,
                          const char *fgname)
{
    DefineModel(name, new CompositeModel(MakeModel(bgname),
                                          MakeModel(fgname)));
}

namespace
{
    string anim_templ_name;
    Anim2d *anim_templ = 0;
}

void
display::DefineAnim (const char *name, bool loop_p)
{
    anim_templ = new Anim2d(loop_p);
    DefineModel(name, anim_templ);
    anim_templ_name = name;
}

void
display::AddFrame (const char *name, const char *model, double time)
{
    if (anim_templ_name != name) 
        fprintf(stderr, "AddFrame: Cannot add frames to completed animations.");
    else
        anim_templ->add_frame(MakeModel(model), 
                              time / 1000.0);
}

void 
display::DefineAlias (const char *name, const char *othername)
{
    DefineModel(name, new AliasModel(othername));
}



//======================================================================
// MODEL IMPLEMENTATION
//======================================================================
Model *
RandomModel::clone() 
{
    if (!modelnames.empty()) {
	int r = int(modelnames.size() * (rand()/(RAND_MAX+1.0)));
	return MakeModel(modelnames[r]);
    } else {
	fprintf(stderr, "display_2d.cc: empty RandomModel\n");
	return 0;
    }
}

Model *
AliasModel::clone()
{
    return MakeModel(name);
}

Anim2d::Anim2d(AnimRep *r) 
    : rep(r), curframe(0), frametime(0), 
      finishedp(false), changedp(false), reversep(false),
      is_sprite(false),
      callback(0)
{
    rep->refcount++;
}

Anim2d::~Anim2d() 
{
    if (--rep->refcount == 0)
        delete rep;
    deactivate();
}

void Anim2d::add_frame(Model *m, double duration)
{
    rep->frames.push_back(new AnimFrame(m, duration));
}

double Anim2d::duration() const
{
    return rep->duration();
}

void Anim2d::draw(px::GC &gc, int x, int y) 
{
    if (!finishedp) {
        AnimFrame *f =rep->frames[curframe];
        f->model->draw(gc,x,y);
        changedp = false;
    }
}
        
void 
Anim2d::draw_shade (px::GC &gc, int x, int y) 
{
    if (!finishedp) {
        AnimFrame *f =rep->frames[curframe];
        f->model->draw_shade(gc,x,y);
    }
}

void 
Anim2d::activate (double worldx, double worldy, bool is_sprite)
{
    add_active_model(this);
    posx=worldx;
    posy=worldy;
    this->is_sprite = is_sprite;
}

void Anim2d::deactivate()
{
    remove_active_model(this);
}

bool 
Anim2d::has_changed(Rect &changed_region)
{
    if (changedp) 
        changed_region = is_sprite 
            ? Rect(int(posx-1), int(posy-1), 3, 3)
            : Rect(int(posx), int(posy), 1, 1);
    return changedp;
}

void 
Anim2d::tick (double dtime) 
{
    assert(curframe < rep->frames.size());
    frametime += dtime;
    double framedur = rep->frames[curframe]->duration;

    if (frametime >= framedur) {
        frametime -= framedur;
        changedp = true;

        if (reversep) {
            if (curframe >= 1)
                curframe--;
            else if (rep->loop)
                curframe = rep->frames.size()-1;
            else
                finishedp = true;
        } 
        else {
            if (curframe+1 < rep->frames.size())
                curframe++;
            else if (rep->loop)
                curframe = 0;
            else 
                finishedp = true;
        }
        if (finishedp && callback!=0)
            callback->animcb();
    }
}
