/*
 * Copyright (C) 2002,2003 Daniel Heck
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License along
 * with this program; if not, write to the Free Software Foundation, Inc.,
 * 59 Temple Place, Suite 330, Boston, MA 02111-1307, USA.
 *
 * $Id: display.cc,v 1.10 2003/01/30 20:38:14 dheck Exp $
 */

/*
 * This file contains the code renders the graphics during the game.
 * This includes displaying the current landscape with all its objects
 * and the inventory at the bottom of the screen.
 */

#include "display.hh"
#include "objects.hh"
#include "video.hh"
#include "options.hh"

#include "px/sdl.hh"
#include "px/px.hh"

#include <vector>
#include <algorithm>
#include <functional>
#include <list>
#include <cmath>
#include <cassert>

using namespace std;
using namespace px;
using namespace display;

//#include "display_impl.hh"

typedef Rect ScreenArea;
typedef Rect WorldArea;

/*
 * Function prototypes
 */
static void world_to_screen (const V3 & pos, int *x, int *y);
static WorldArea screen_to_world (const ScreenArea &sa);
static void mark_redraw_area (const WorldArea &wa);
static void mark_update_area (const WorldArea &wa);
static void mark_redraw_screen();
static void draw_sprites(SpriteLayer l, bool shades, GC &gc);
static void follow_center();

namespace
{
    class Window {
    public:
        Window() {}
        Window (const ScreenArea &area) : m_area(area) 
        {}

        const ScreenArea &get_area() const { return m_area; }
    private:
        ScreenArea m_area;
    };


    struct Field {
        static const int NUM_LAYERS = 3;
    
        Field() { 
            fill_n(layers, NUM_LAYERS, (Model*)0); 
            redrawp = updatep = false;
        }
        ~Field() { delete_sequence(layers, layers+NUM_LAYERS); }

        Model* layers[NUM_LAYERS];
        bool redrawp;
        bool updatep;
    };

    class Sprite : public Nocopy {
    public:
        Model *model;
        V3 pos;
        SpriteLayer layer;
        bool visible;

        Sprite(const V3 & p, SpriteLayer l, Model *m) 
            : model(m), pos(p), layer(l)
        {}
        ~Sprite() { delete model; }
    };

    struct Rubber {
        V2 start,end;
        V2 oldstart, oldend;

        Rubber(const V2 &s, const V2 &e) :start(s), end(e) {}
        Rubber() {}
    };

    typedef Array2<Field>   FieldArray;
    typedef list<Model*>   ModelList;

    //typedef map<SpriteId, Sprite*> SpriteList;
    typedef vector<Sprite*> SpriteList;
    typedef vector<Rubber> RubberList;
}


//======================================================================
// STATUS BAR
//======================================================================
namespace
{
    class TextDisplay {
    public:
        TextDisplay(Rect a, Font &f);

        void set_text(const string &t, TextMode m);

        void deactivate() { active = false; }
        bool is_active() const { return active; }

        void tick(double dtime);
        bool has_changed() { return changedp; }

        void draw(px::GC &gc, const Rect &r);
    private:
        string text;
        bool active, changedp;
        TextMode mode;
        Rect area;
        double xoff;
        const double scrollspeed; // pixels per second
        std::auto_ptr<Surface> textsurface;
        px::Font &font;
        double time, maxtime;
    };
    
    class StatusBarImpl : public StatusBar, public Window {
    public:
        StatusBarImpl (const ScreenArea &area);
        ~StatusBarImpl();

        bool has_changed() const { return m_changedp; }
        void redraw (px::GC &gc, const ScreenArea &r);
        void tick (double dtime);
        void new_world();

        // Window interface.

        // StatusBar interface.
        void set_inventory (Inventory *inv);
        void update_inventory (Inventory *inv);

        void show_text (const std::string &str, 
                        TextMode m=TEXT_SCROLLING);
        void hide_text() { m_textview.deactivate(); }
    private:
        ScreenArea       m_itemarea;
        ScreenArea       m_textarea;
        vector<Model*>   m_models;
        bool             m_changedp;
        Inventory *      m_inventory;
        TextDisplay      m_textview;

        double m_leveltime;
        bool   m_showtime_p;
    };
}

StatusBarImpl::StatusBarImpl (const ScreenArea &area) 
    : Window(area), 
      m_itemarea (area.x+142, area.y+area.h-64+5, area.w-147, 64-10),
      m_textarea (area.x+147, area.y+area.h-64+20, area.w-157, 32),
      m_models(),
      m_changedp(false),
      m_inventory(0),
      m_textview (m_textarea, *enigma::GetFont("dreamorp24")),
      m_leveltime (0),
      m_showtime_p (true)
{}

StatusBarImpl::~StatusBarImpl()
{}

void 
StatusBarImpl::redraw (px::GC &gc, const ScreenArea &r)
{
    ScreenArea a = get_area();
    clip(gc, intersect(a, r));

    blit(gc, a.x, a.y, enigma::GetImage ("inventory"));

    if (m_showtime_p)
    {
        char time[8];
        int minutes = static_cast<int>(m_leveltime/60) % 100;
        int seconds = static_cast<int>(m_leveltime) % 60;
        Font *f = enigma::GetFont ("timefont");

        sprintf (time, "%02d : %02d", minutes, seconds);
        Surface *s = f->render (time);
        blit(gc, a.x + 30, a.y+13, s);
        delete s;
    }
    else
        blit(gc, a.x+5, a.y+10, enigma::GetImage ("enigma_logo2_small"));


    if (m_textview.is_active()) {
        m_textview.draw (gc, r);
    }
    else 
    {
        for (unsigned i=0; i<m_models.size(); ++i)
        {
            Model *m = m_models[i];
            m->draw(gc, m_itemarea.x + 10+i*40, m_itemarea.y + 11);
        }
    }
    m_changedp = false;
}

void 
StatusBarImpl::set_inventory (Inventory *inv)
{
    m_inventory = inv;
    update_inventory (inv);
}

void 
StatusBarImpl::update_inventory (Inventory *inv)
{
    if (inv && inv == m_inventory)
    {
        px::delete_sequence(m_models.begin(), m_models.end());
        m_models.clear();

        for (int i=0; i<inv->size(); ++i) {
            world::Item *it = inv->get_item(i);
            m_models.push_back(MakeModel(it->get_inventory_model ()));
        }
        m_changedp = true;
    }
}
    
void 
StatusBarImpl::show_text (const std::string &str, TextMode m)
{
    m_textview.set_text (str, m);
}

void 
StatusBarImpl::tick (double dtime)
{
    // Update text display
    if (m_textview.is_active()) 
    {
        m_textview.tick (dtime);
        m_changedp = m_textview.has_changed();
    }

    double oldtime=m_leveltime;
    m_leveltime += dtime;
    if (m_showtime_p && int(m_leveltime) - int(oldtime) >= 1)
        m_changedp = true;      // update clock
}

void 
StatusBarImpl::new_world() 
{
    m_leveltime = 0;

    delete_sequence(m_models.begin(), m_models.end());
    m_models.clear();
    m_textview.deactivate();
}



//----------------------------------------
// TextDisplay
//----------------------------------------
TextDisplay::TextDisplay(ScreenArea a, Font &f) 
    : text(), active(false), changedp(false), 
      mode(TEXT_SCROLLING),
      area(a), xoff(0), scrollspeed(200),
      textsurface(0), font(f)
{
    time = maxtime = 0;
}

void 
TextDisplay::set_text(const string &t, TextMode m) 
{ 
    text = t;
    mode = m;
    textsurface.reset(font.render(text.c_str()));

    time = 0;
    maxtime = 1e20;             // this should be close to ``indefinitely''
    switch (mode) {
    case TEXT_SCROLLING:
        xoff = -area.w;
        break;
    case TEXT_2SECONDS:
        maxtime = 2.0;          // 2 seconds
        // fall through
    case TEXT_STATIC:
        xoff = -(area.w - textsurface->width())/2;
        break;
    }
    changedp = active = true;
}

void 
TextDisplay::tick(double dtime) 
{
    time += dtime;
    if (time > maxtime) {
        active = false;
        changedp = true;
    } 
    else {
        int oldxoff = int(xoff);
        if (mode == TEXT_SCROLLING) {
            xoff += dtime * scrollspeed;
            changedp = int(xoff) != oldxoff;
            if (xoff >= textsurface->width()) {
                active = false;
                changedp = true;
            }
        }
    }
}

void 
TextDisplay::draw(px::GC &gc, const ScreenArea &r) 
{
    if (active) {
        clip(gc, intersect(area, r));
        set_color(gc, 0,0,0);
        box(gc, area);
        blit(gc, area.x-int(xoff), area.y, textsurface.get());
    }
}



//----------------------------------------
// Local variables
//----------------------------------------
namespace 
{
    struct Level {
    public: 
        Level(int w, int h) : width(w), height(h) {}
        ~Level() {
        }
        int width, height;
    };

    Level *level;
}

static DisplayFlags    display_flags = SHOW_ALL;

/* The absolute offset of the visible area inside the world.  Under normal
   circumstances, these variables contain integers, but when softscrolling is
   active, they can temporarily assume fractional values. */
static double ScrollX = 0;
static double ScrollY = 0;


/* The width and height of a tile in pixels.  Together with ScrollX, ScrollY,
   these two variables are used to convert world coordinates to screen
   coordinates.  */
static int TileWidth = 32;
static int TileHeight = 32;

/* The current and maximum number of effect sprites. */
static int NumEffectSprites = 0;
static int MaxEffectSprites = 10;

static bool RedrawEverything=false;

namespace
{
    typedef AssocList<unsigned, Rubber> RubberMap;

    // Do not change the order of the following two variables!
    ModelList       active_models;
    FieldArray      fields(0,0);
    SpriteList      sprites;
    SpriteList      effects;
    RubberMap       rubbers;

    StatusBarImpl   *status_bar = 0;

    // Screen layout
    ScreenArea gamearea(0,0,640, 480-64);
    ScreenArea inventoryarea(0, 480-64, 640, 64);
}

StatusBar *
display::GetStatusBar()
{
    return status_bar;
}

static void
mark_redraw_area (const WorldArea &wa)
{
    int x2 = Min(level->width, wa.x+wa.w);
    int y2 = Min(level->height, wa.y+wa.h);
    for (int x=Max(0,wa.x); x <x2; x++)
        for (int y=Max(0,wa.y); y<y2; y++)
            fields(x,y).redrawp = true;
}

static void
mark_redraw_screen()
{
    mark_redraw_area(screen_to_world(gamearea));
}

static void
mark_update_world()
{
    mark_update_area(WorldArea(0,0, level->width, level->height));
}

static void
mark_update_area (const WorldArea &wa)
{
    int x2 = Min(level->width, wa.x+wa.w);
    int y2 = Min(level->height, wa.y+wa.h);
    for (int x=Max(0,wa.x); x <x2; x++)
        for (int y=Max(0,wa.y); y<y2; y++)
            fields(x,y).updatep = true;
}

static WorldArea
screen_to_world (const ScreenArea &a)
{
    int worldw = level->width;
    int worldh = level->height;

    int sx = int(ScrollX*TileWidth);
    int sy = int(ScrollY*TileHeight);

    int x1 = Max(0, (sx + a.x) / TileWidth);
    int y1 = Max(0, (sy + a.y) / TileHeight);
    int x2 = Min(worldw, (sx + a.x+a.w+TileWidth-1) / TileWidth);
    int y2 = Min(worldh, (sy + a.y+a.h+TileHeight-1) / TileHeight);

    return WorldArea (x1, y1, x2-x1, y2-y1);
}

static void 
world_to_screen (const V3 & pos, int *x, int *y)
{
    *x = int((pos[0] - ScrollX)*TileWidth);
    *y = int((pos[1] - ScrollY)*TileHeight);
}




//======================================================================
// DISPLAY ENGINE
//======================================================================

//----------------------------------------
// Rubber bands
//----------------------------------------

/* Mark the screen region occupied by a rubber band for redraw.
   Problem is: what region is that exactly?  What pixels on the screen
   will the line rasterizer touch?  Hard to tell, especially when
   anti-aliasing is used.

   This function constructs a list of rectangles that completely
   enclose the line by subdividing the line into n segments and
   constructing the bounding box for each of these segments.  To
   account for the (effective) finite width of the line, these boxes
   need to be enlarged by a small amount to make them overlap a bit.

   The number n of subdivision depends on the length of the line.  n=1
   would of course do, but we want to redraw as little of the screen
   as possible.  'maxboxsize' 
*/
static void
mark_redraw_rubber (const Rubber &r)
{
    const double maxboxsize = 1.0; //0.5;

    double w0 = r.start[0]-r.end[0];
    double h0 = r.start[1]-r.end[1];
    int n = int (max(abs(w0),abs(h0)) / maxboxsize)+1;

    double w = w0/n;
    double h = h0/n;

    double overlap = 0.5;

    double x = r.end[0];
    double y = r.end[1];
    
    WorldArea wa(int(x-overlap), int(y-overlap), 
                 int(abs(w)+2*overlap+1), int(abs(h)+2*overlap+1));

    for (int i=0; i<n; ++i) {
        wa.x = int(x-overlap);
        wa.y = int(y-overlap);
//         if (wa.x != int(x) || wa.y !=int(y))
//             wa.w = wa.h = 2;
//         else
//             wa.w = wa.h = 1;

        mark_redraw_area (wa);

        x += w;
        y += h;
    }
}


RubberID
display::AddRubber (const V2 &p1, const V2 &p2)
{
    static unsigned id=1;
    rubbers[id] = Rubber(p1, p2);
    mark_redraw_rubber (rubbers[id]);
    return RubberID(id++);
}

RubberID::RubberID(unsigned id_): id(id_)
{
}

void 
RubberID::update_first(const V2 &p1)
{
//    assert(0 <= id && id < rubbers.size());
    mark_redraw_rubber (rubbers[id]);
    rubbers[id].start = p1;
    mark_redraw_rubber (rubbers[id]);
}

void 
RubberID::update_second(const V2 &p2)
{
//    assert(0 <= id && id < rubbers.size());
    mark_redraw_rubber (rubbers[id]);
    rubbers[id].end = p2;
    mark_redraw_rubber (rubbers[id]);
}

void RubberID::kill()
{
//    assert(0 <= id && id < rubbers.size());
    mark_redraw_rubber (rubbers[id]);
    RubberMap::iterator i=rubbers.find(id);
    if (i != rubbers.end())
        rubbers.erase(i);
}

//----------------------------------------
// Shadow rendering
//----------------------------------------

namespace
{
    class ShadowLayer {
    public:
        virtual ~ShadowLayer() {}
        virtual void new_world(int w, int h) {}
        virtual void update (int x, int y) = 0;
        virtual void draw(GC &gc, int xpos, int ypos, int x, int y) = 0;
    };

    class ShadowLayer_Alpha : public ShadowLayer {
    public:
        ShadowLayer_Alpha();
        ~ShadowLayer_Alpha();

        void new_world(int w, int h);
        void update (int x, int y);
        void draw(GC &gc, int xpos, int ypos, int x, int y);
    private:
        void shadow_blit (px::Surface *scr, int x, int y, 
                          px::Surface *shadows, px::Rect r);

        Surface *new_surface();
        void     dispose_surface(Surface *s);
        Surface *get_cache(int x, int y);
        bool     draw_stone_shade(GC &gc, int x, int y, Field &f);

        // Variables.
        Surface           *shadow_surface;
        Uint32             shadow_ckey; // Color key
        Array2<Surface *>  m_gridcache;
        Array2<bool>       has_shadow; 
        vector<Surface *>  m_surface_avail;
        Surface           *buffer;
    };

    class ShadowLayer_Stippled : public ShadowLayer {
    public:
        ShadowLayer_Stippled() {}
        void update (int x, int y) {}
        void draw(GC &gc, int xpos, int ypos, int x, int y)
        {}
    private:

    };

    ShadowLayer *shadow_layer;
}

ShadowLayer_Alpha::ShadowLayer_Alpha() 
    : m_gridcache()
{
    SDL_Surface *ss = SDL_CreateRGBSurface(SDL_SWSURFACE,
                                           TileWidth, TileHeight,
                                           16,0,0,0,0);
    SDL_SetColorKey(ss, SDL_SRCCOLORKEY,
                    SDL_MapRGB(ss->format, 255,255,255));
    SDL_SetAlpha(ss, SDL_SRCALPHA, 128);
    buffer = new Surface(ss);
}

ShadowLayer_Alpha::~ShadowLayer_Alpha()
{
    delete_sequence(m_gridcache.begin(), m_gridcache.end());
    delete_sequence(m_surface_avail.begin(), m_surface_avail.end());
    delete buffer;
}

void 
ShadowLayer_Alpha::new_world(int w, int h) 
{
    Array2<Surface*>::iterator i;
    for (i=m_gridcache.begin();i!=m_gridcache.end(); ++i)
        dispose_surface(*i);
    m_gridcache.resize(w,h);
    has_shadow.resize(w,h,false);
}

Surface *
ShadowLayer_Alpha::new_surface()
{
    Surface *s = 0;
    if (m_surface_avail.empty()) {
        SDL_Surface *ss = SDL_CreateRGBSurface(SDL_SWSURFACE|SDL_RLEACCEL,
                                               TileWidth, TileHeight,
                                               16,0,0,0,0);
// #if 0
//         SDL_SetColorKey(ss, SDL_SRCCOLORKEY | SDL_RLEACCEL,
//                         SDL_MapRGB(ss->format, 255,255,255));
//         SDL_SetAlpha(ss, SDL_SRCALPHA | SDL_RLEACCEL, 128);
// #else
//         SDL_SetColorKey(ss, SDL_SRCCOLORKEY,
//                         SDL_MapRGB(ss->format, 255,255,255));
//         SDL_SetAlpha(ss, SDL_SRCALPHA, 128);
// #endif
        s = new Surface(SDL_DisplayFormat(ss));
        SDL_FreeSurface(ss);
    } else {
        s = m_surface_avail.back();
        m_surface_avail.pop_back();
    }
    return s;
}

void 
ShadowLayer_Alpha::dispose_surface(Surface *s)
{
    if (s)
        m_surface_avail.push_back(s);
}


Surface *
ShadowLayer_Alpha::get_cache(int x, int y)
{
    Surface *s= m_gridcache(x,y);
    if (!s) 
        m_gridcache(x,y) = s = new_surface();
    return s;
}

bool
ShadowLayer_Alpha::draw_stone_shade(GC &gc, int x, int y, Field &f)
{
    Model *m=f.layers[enigma::GRID_STONES];
    if (m)
        m->draw_shade(gc, x, y);
    return m!=0;
}

void
ShadowLayer_Alpha::update (int x, int y)
{
//     int xpos = static_cast<int>((x - ScrollX)*TileWidth);
//     int ypos = static_cast<int>((y - ScrollY)*TileHeight);
    int has_sh = false;

    Surface *s=get_cache(x,y);

    GC gc(s);
    set_color (gc, 255,255,255);
    box(gc, s->size());

    if (x>0&&y>0)
        has_sh |= draw_stone_shade(gc, -TileWidth, -TileHeight, 
                                   fields(x-1,y-1));
    if (x>0)
        has_sh |= draw_stone_shade(gc, -TileWidth, 0, fields(x-1,y));
    if (y>0)
        has_sh |= draw_stone_shade(gc, 0, -TileHeight, fields(x,y-1));
    has_sh |= draw_stone_shade(gc, 0, 0, fields(x,y));
    has_shadow (x,y) = has_sh;

    if (!has_sh) {
//        m_gridcache(x,y) = 0;
//        dispose(s);
    }
}

void 
ShadowLayer_Alpha::draw(GC &gc, int xpos, int ypos, int x, int y)
{
    bool has_actor = false;

    // Is there an actor that is partially inside this field?
    for (unsigned i=0; i<sprites.size(); ++i) {
        Sprite *s = sprites[i];
        if (s && s->layer == SPRITE_ACTOR) {
            Rect r(int(s->pos[0]-0.3), int(s->pos[1]-0.3),2,2);
            if (r.contains(x,y)) {
                has_actor = true;
                break;
            }
        }
    }

    if (has_actor || has_shadow(x,y)) {
        Surface *s=get_cache(x,y);
        
        if (!has_actor) {
            SDL_Surface *ss = s->get_surface();
            SDL_SetColorKey(ss, SDL_SRCCOLORKEY,
                            SDL_MapRGB(ss->format, 255,255,255));
            SDL_SetAlpha (ss, SDL_SRCALPHA, 128);
            blit (gc, xpos,ypos,s);
            SDL_SetAlpha (ss, 0, 0);
            SDL_SetColorKey(ss, 0,0);
        }
        else {
            GC gc2(buffer);
            blit (gc2, 0,0,s);
            for (unsigned i=0; i<sprites.size(); ++i) {
                Sprite *s = sprites[i];
                if (s && s->layer == SPRITE_ACTOR){
                    int sx = int(s->pos[0]*TileWidth) - x*TileWidth;
                    int sy = int(s->pos[1]*TileWidth) - y*TileHeight;
                    s->model->draw_shade(gc2, sx, sy);
                }
            }
            blit(gc, xpos, ypos, buffer);
        }
    }
}

void
ShadowLayer_Alpha::shadow_blit (px::Surface *scr, int x, int y, 
                                px::Surface *shadows, px::Rect r)
{
    r.intersect(scr->size());
    x=r.x;
    y=r.y;

    SDL_Surface *s = shadows->get_surface(); // source
    SDL_Surface *d = scr->get_surface();

    SDL_LockSurface(s);
    SDL_LockSurface(d);
    
    assert (s->format->BytesPerPixel==2 &&
            d->format->BytesPerPixel==2);


    /* This bitmask is used to erase the highest bit in every color
       field.  Every color value can be effectively halved by (a)
       shifting everything one bit to the right, and (b) applying the
       mask to the resulting value. */
    Uint32 mask;
    mask  = (d->format->Rmask >> d->format->Rshift+1) << d->format->Rshift;
    mask |= (d->format->Gmask >> d->format->Gshift+1) << d->format->Gshift;
    mask |= (d->format->Bmask >> d->format->Bshift+1) << d->format->Bshift;

    register Uint16 skey = (Uint16)shadow_ckey;
    for (int h=r.h; h; --h) 
    {
        Uint16 *sp = static_cast<Uint16*>(shadows->pixel_pointer(r.x, r.y));
        Uint16 *dp = static_cast<Uint16*>(scr->pixel_pointer(x, y));
        for (int w=r.w; w; --w)
        {
            if (*sp != skey)
                *dp = (*dp >> 1) & mask;
            dp++;
            sp++;
        }
        y++;
        r.y++;
    }

    SDL_UnlockSurface(s);
    SDL_UnlockSurface(d);
}



void 
display::SetShadowType(ShadowType st)
{
    delete shadow_layer;
    if (options::BitsPerPixel == 16)
        shadow_layer = new ShadowLayer_Alpha;
    else
        shadow_layer = new ShadowLayer_Stippled;
}

void 
display::Init()
{
    InitModels();
    status_bar = new StatusBarImpl (inventoryarea);

    SetShadowType(SHADOW_ALPHA);
}

void 
display::Shutdown()
{
    delete_sequence(sprites.begin(), sprites.end());
    delete_sequence(effects.begin(), effects.end());
    active_models.clear();
    delete status_bar;
    delete shadow_layer;

    ShutdownModels();
}

void 
display::NewWorld(int w, int h)
{
    SetFollowMode(FOLLOW_NONE);
    delete level;
    level = new Level(w, h);

    active_models.clear();

    fields = FieldArray(w, h);
    ScrollX = ScrollY = 0;

    delete_sequence(sprites.begin(), sprites.end());
    delete_sequence(effects.begin(), effects.end());
    sprites.clear();
    effects.clear();
    NumEffectSprites = 0;

    status_bar->new_world();
    shadow_layer->new_world(w,h);
    mark_update_world();
}

//----------------------------------------------------------------------
// Sprite following code
//----------------------------------------------------------------------

namespace
{
    class Follow {
    public:
        virtual ~Follow() {}
        virtual void tick(double dtime) = 0;
        virtual void center() =0;
    };

    class Follow_Screen : public Follow {
    public:
        void tick(double dtime);
        void center();
    };

    class Follow_Scrolling : public Follow {
    public:
        Follow_Scrolling();
        void tick(double dtime);
        void center();
    private:
        bool   currently_scrolling;
        V3     curpos, destpos;
        V3     dir;
        double scrollspeed;
        double resttime;
    };
}

namespace
{
    const SpriteId  magic_spriteid = 1000000;
    SpriteId        follow_sprite  = magic_spriteid;
    FollowMode      follow_mode    = FOLLOW_NONE;
    Follow         *follower       = 0;
}

static void
SetScreenOffset(double xoff, double yoff)
{
    int oldx = int(ScrollX * TileWidth);
    int oldy = int(ScrollY * TileHeight);
    int newx = int(xoff * TileWidth);
    int newy = int(yoff * TileHeight);

    if (newx != oldx || newy != oldy)
    {
        Rect newarea(newx, newy, gamearea.w, gamearea.h);
        Rect oldarea(oldx, oldy, gamearea.w, gamearea.h);
        Rect common = intersect(newarea, oldarea);

        Screen *scr = video::GetScreen();
        Rect blitrect(common.x-oldx+gamearea.x,
                      common.y-oldy+gamearea.y,
                      common.w, common.h);
        scr->blit(common.x-newx+gamearea.x,
                  common.y-newy+gamearea.y,
                  scr,
                  blitrect);
        blitrect.x = common.x-newx+gamearea.x;
        blitrect.y = common.y-newy+gamearea.y;
        scr->update_rect(blitrect);

        ScrollX = newx/double(TileWidth);
        ScrollY = newy/double(TileHeight);

        RectList rl;
        rl.push_back(gamearea);
        rl.sub(blitrect);

        for (RectList::iterator i=rl.begin(); i!=rl.end(); ++i) {
            Rect r = screen_to_world(*i);
            mark_redraw_area (r);
        }
    }
}


void 
Follow_Screen::center()
{
    tick(0);
}

/* Determine whether the screen must be scrolled or not, and change
   the coordinate origin of the screen accordingly. */
void
Follow_Screen::tick(double)
{
    assert(follow_sprite != magic_spriteid);
    Sprite *s = sprites[follow_sprite];

    bool done=false;
    while (!done) {
        int sx, sy;
        world_to_screen(s->pos, &sx, &sy);

        done=true;
        if (sx < gamearea.x + TileWidth/2) {
            ScrollX -= 19;
            mark_redraw_screen();
            done=false;
        }
        if (sy < gamearea.y + TileHeight/2) {
            ScrollY -= 12;
            mark_redraw_screen();
            done=false;
        }

        if (sx >= gamearea.x + gamearea.w-TileWidth/2) {
            ScrollX += 19;
            mark_redraw_screen();
            done=false;
        }
        if (sy >= gamearea.y + gamearea.h-TileHeight/2) {
            ScrollY += 12;
            mark_redraw_screen();
            done=false;
        }
    }
}



Follow_Scrolling::Follow_Scrolling()
    : currently_scrolling(false),
      scrollspeed(0), resttime(0)
{}

void
Follow_Scrolling::center()
{
    tick(1000);
    tick(1000);
}

void 
Follow_Scrolling::tick(double dtime)
{
    assert(follow_sprite != magic_spriteid);
    Sprite *s = sprites[follow_sprite];

    int sx, sy;
    world_to_screen(s->pos, &sx, &sy);

    if (currently_scrolling) {
        resttime -= dtime;
        if (resttime <= 0) {
            SetScreenOffset(destpos[0], destpos[1]);
            currently_scrolling = false;
        } else {
            curpos += dir * scrollspeed*dtime;
            SetScreenOffset(curpos[0], curpos[1]);
        }
    } else {
        const int border = 1;
        bool scrollx_p = (sx < gamearea.x + TileWidth*border) 
            || (sx >= gamearea.x + gamearea.w - TileWidth*border);

        bool scrolly_p = (sy < gamearea.y + TileHeight*border)
            || (sy >= gamearea.y + gamearea.h - TileHeight*border);

        if (scrollx_p || scrolly_p)
        {
            currently_scrolling = true;
            curpos = V3(ScrollX, ScrollY, 0);
            destpos = curpos + s->pos - V3(ScrollX+gamearea.w/TileWidth/2, 
                                           ScrollY+gamearea.h/TileHeight/2,
                                           0);
            destpos[0] = int(destpos[0]*TileWidth)/TileWidth;
            destpos[1] = int(destpos[1]*TileHeight)/TileHeight;
            destpos[0] = max(0, min((int) destpos[0], 
                                    level->width-gamearea.w/TileWidth));
            destpos[1] = max(0, min((int) destpos[1], 
                                    level->height-gamearea.h/TileHeight));
            if (!scrollx_p)
                destpos[0] = ScrollX;
            if (!scrolly_p)
                destpos[1] = ScrollY;

            scrollspeed = Max(30, 0);
            resttime = length(destpos - curpos)/scrollspeed;
            dir = normalize(destpos - curpos);
        }
    }
}

static void
follow_doit(double dtime)
{
    if (follow_sprite != magic_spriteid) {
        follower->tick(dtime);
    }
}

static void
follow_center()
{
    if (follow_mode == FOLLOW_NONE)
        SetFollowMode(FOLLOW_SCREEN);
    if (follow_sprite != magic_spriteid) {
        follower->center();
    }
}

static void
set_follower (Follow *f)
{
    delete follower;
    follower = f;
}


void 
display::FollowSprite(SpriteId id)
{
    follow_sprite = id;
    follow_center();
}


FollowMode 
display::SetFollowMode(FollowMode m)
{
    FollowMode old=follow_mode;
    if (m != follow_mode)
    {
        follow_mode = m;
        switch (follow_mode) {
        case FOLLOW_NONE:      
            set_follower(0); 
            break;
        case FOLLOW_SCROLLING: 
            set_follower (new Follow_Scrolling); 
            break;
        case FOLLOW_SCREEN:    
            set_follower (new Follow_Screen); 
            break;
        };
    }
    return old;
}

//----------------------------------------
// Handling active models
//----------------------------------------

namespace display
{
    void 
    add_active_model(Model *m)
    {
        list<Model*> &am = ::active_models;
        if (find(am.begin(), am.end(), m) ==am.end()) {
            // if not already in the list...
            am.push_back(m);
        }
    }

    void remove_active_model(Model *m)
    {
        ::active_models.remove(m);
    }
}

void maybe_redraw_model(Model *m)
{
    WorldArea wa;
    if (m->has_changed(wa)) {
        mark_update_area(wa);
    }
}

void redraw_sprite_region(SpriteId id)
{
    Sprite *s = sprites[id];
    int x = int(s->pos[0]-0.5);
    int y = int(s->pos[1]-0.5);
//    if (s->layer == SPRITE_ACTOR)
//        mark_update_area (Rect(x-1, y-1, 3, 3));
//    else
    mark_redraw_area (Rect(x, y, 3, 3));
}

void 
display::Tick (double dtime)
{
    ModelList &am = active_models;
    am.remove_if(mem_fun(&Model::is_garbage));

    /* for_each does not work; animation may remove itself during
       tick() */
    for (ModelList::iterator i=am.begin(); i!=am.end(); ) {
        ModelList::iterator n=next(i);
        (*i)->tick(dtime);
        i=n;
    }
    for_each(am.begin(), am.end(), &maybe_redraw_model);

    follow_doit(dtime);
    
    status_bar->tick (dtime);
}

static void 
mark_change(const GridLoc &l)
{
    if (l.layer == enigma::GRID_STONES)
        mark_update_area (Rect(l.pos.x, l.pos.y, 2,2));
    else
        mark_redraw_area (Rect(l.pos.x,l.pos.y, 1,1));
}

static Model*
set_model(const GridLoc & l, Model* m)
{
    int x = l.pos.x, y=l.pos.y;
    Field& f = fields(x, y);
    delete f.layers[l.layer];
    f.layers[l.layer] = m;
    if (m)
        m->activate(double(x), double(y), false);
    mark_change(l);
    return m;
}

Model*
display::SetModel(const GridLoc &l, Model *m)
{
    return set_model(l, dynamic_cast<Model*>(m));
}

Model *
display::SetModel(const GridLoc &l, const string &modelname)
{
    return set_model(l, MakeModel(modelname));
}

void
display::KillModel(const GridLoc & l)
{
    delete YieldModel(l);
}

Model *
display::GetModel(const GridLoc &l)
{
    Field& f = fields(l.pos.x, l.pos.y);
    return f.layers[l.layer];
}

Model *
display::YieldModel(const GridLoc &l)
{
    Field& f = fields(l.pos.x, l.pos.y);
    Model *m = f.layers[l.layer];
    f.layers[l.layer] = 0;
    mark_change(l);
    return m;
}

SpriteId
display::AddSprite (SpriteLayer l, const V3& pos, const char *modelname)
{
    SpriteId id;

    if (l == SPRITE_EFFECT) {
	if (NumEffectSprites >= MaxEffectSprites)
	    return magic_spriteid;
	else
	    NumEffectSprites += 1;
    }

    Model *m = modelname ? MakeModel(modelname) : 0;

    switch (l) {
    case SPRITE_ACTOR:
    case SPRITE_DEBRIS:
        id = sprites.size();
        sprites.push_back(new Sprite(pos,l,m));
        break;
    case SPRITE_EFFECT:
        {
            SpriteList &sl = sprites; //effects;
            // Find the first empty slot
            SpriteList::iterator i = find(sl.begin(), sl.end(), (Sprite*)0);
            if (i == sl.end()) {
                id = sl.size();
                sl.push_back(new Sprite(pos, l, m));
            }
            else {
                id = distance(sl.begin(), i);
                *i = new Sprite(pos, l, m);
            }
        }
        break;
    }
    if (m) {
        m->activate(pos[0], pos[1], true);
        redraw_sprite_region(id);
    }
    return id;
}

void
display::MoveSprite(SpriteId id, const px::V3& newpos)
{
    if (id == magic_spriteid)
        return;
    Sprite *sprite = sprites[id];
    if ((fabs(newpos[0]-sprite->pos[0]) > 1.0/TileWidth) ||
	(fabs(newpos[1]-sprite->pos[1]) > 1.0/TileHeight))
    {
	redraw_sprite_region(id);
	sprite->pos = newpos;
	redraw_sprite_region(id);
    }
}

void 
display::ReplaceSprite (SpriteId id, const char *modelname)
{
    if (id == magic_spriteid)
        return;
    Sprite *sprite = sprites[id];
    delete sprite->model;
    sprite->model = MakeModel(modelname);
    if (!sprite->model)
        fprintf(stderr,"Could not set sprite model  `%s'\n", modelname);
    else
        sprite->model->activate(sprite->pos[0], sprite->pos[1], true);
    redraw_sprite_region(id);
}


void
display::KillSprite(SpriteId id) 
{
    if (id == magic_spriteid)
        return;
    if (Sprite *sprite = sprites[id]) {
        sprites[id] = 0;
	if (sprite->layer == SPRITE_EFFECT)
	    NumEffectSprites -= 1;
        delete sprite;
    }
}

void 
display::SetCallback(SpriteId id, ModelCallback *cb)
{
    if (Model *m = GetModel(id))
        m->set_callback(cb);
}

Model *
display::GetModel(SpriteId id)
{
    if (id == magic_spriteid)
        return 0;
    return sprites[id]->model;
}

void
display::SetTileSize(int w, int h)
{
    TileWidth = w;
    TileHeight = h;
}

//======================================================================
// DRAWING CODE
//======================================================================
static void
draw_sprites(SpriteLayer l, bool shades, GC &gc)
{
//    SpriteList &sl = (l==SPRITE_ACTOR) ? sprites :effects;
    SpriteList &sl = sprites;//(l==SPRITE_ACTOR) ? sprites :effects;
    
    for (unsigned i=0; i<sl.size(); ++i)
    {
        Sprite *s = sl[i];
        if (!(s && s->layer == l && s->model))
            continue;

        if (s->model->is_garbage()  && s->layer==SPRITE_EFFECT) {
            // Only remove effect sprites -- actor sprites remain in
            // the world all the time
            KillSprite(i);
        } 
        else {
            int sx, sy;
            world_to_screen(s->pos, &sx, &sy);
            if (shades)
                s->model->draw_shade(gc, sx, sy);
            else
                s->model->draw(gc, sx, sy);
        }
    }
}

#include "SDL_gfxPrimitives.h"

static void
draw_rubbers (GC &gc)
{
    SDL_Surface *surf = video::GetScreen()->get_surface();

    Rect r = gamearea; //gc.get_cliprect();
    SDL_Rect s;
    sdl::copy_rect(s, r);
    SDL_SetClipRect (surf, &s);

    for (RubberMap::iterator i=rubbers.begin(); i!= rubbers.end(); ++i)
    {
        aalineRGBA(surf, 
                   int ((i->second.start[0]-ScrollX)*TileWidth),
                   int ((i->second.start[1]-ScrollY)*TileHeight),
                   int ((i->second.end[0]-ScrollX)*TileWidth),
                   int ((i->second.end[1]-ScrollY)*TileHeight),
                   240,140,20,255);
    }
    SDL_SetClipRect (surf, NULL);
}

void 
display::ToggleFlag(DisplayFlags flag)
{
    display_flags = DisplayFlags(display_flags ^ flag);
}

static inline void
draw_model(GC &gc, int x, int y, Model *m)
{
    if (m)
        m->draw(gc, x, y);
}

static void
draw_field(GC &gc, Field &f, int x, int y)
{
    int xpos = static_cast<int>((x - ScrollX)*TileWidth);
    int ypos = static_cast<int>((y - ScrollY)*TileHeight);

    using namespace enigma;

    if (display_flags & SHOW_FLOOR) draw_model(gc, xpos, ypos, f.layers[GRID_FLOOR]);
    if (display_flags & SHOW_ITEMS) draw_model(gc, xpos, ypos, f.layers[GRID_ITEMS]);
    if (display_flags & SHOW_SPRITES) draw_sprites(SPRITE_DEBRIS, false, gc);
    if (display_flags & SHOW_SHADES) shadow_layer->draw(gc, xpos, ypos, x, y);
    if (display_flags & SHOW_SPRITES) draw_sprites(SPRITE_ACTOR, false, gc);
    if (display_flags & SHOW_STONES) draw_model(gc, xpos, ypos, f.layers[GRID_STONES]);
    if (display_flags & SHOW_SPRITES) draw_sprites(SPRITE_EFFECT, false, gc);
}


static Uint32 last_frame_time = 0;

void
display::DrawAll(px::GC &gc)
{
    WorldArea wa = screen_to_world (gamearea);

    int x2 = wa.x+wa.w;
    int y2 = wa.y+wa.h;
    for (int x=wa.x; x<x2; x++) {
        for (int y=wa.y; y<y2; y++) {
            Field &f = fields(x,y);
            if (f.updatep) {
                shadow_layer->update(x,y);
                f.redrawp = true;
                f.updatep = false;
            }
            ScreenArea sa(int((x-ScrollX)*TileWidth), 
                          int((y-ScrollY)*TileHeight),
                          TileWidth, 
                          TileHeight);
            sa.intersect(gamearea);
            clip(gc, sa);
            draw_field(gc, f, x,y);
        }
    }
    draw_rubbers (gc);
    status_bar->redraw (gc, inventoryarea);
}

void
display::RedrawAll(Screen *scr)
{
    mark_redraw_screen();
    RedrawEverything = true;
    Redraw(scr);
}

void
display::Redraw(Screen *screen)
{
    GC gc(screen);
    clip(gc, gamearea);
    WorldArea wa = screen_to_world (gamearea);

    int x2 = wa.x+wa.w;
    int y2 = wa.y+wa.h;
    for (int x=wa.x; x<x2; x++) {
        for (int y=wa.y; y<y2; y++) {
            Field &f = fields(x,y);
            if (f.updatep) {
                shadow_layer->update(x,y);
                f.redrawp = true;
                f.updatep = false;
            }
            if (f.redrawp) {
                ScreenArea sa(int((x-ScrollX)*TileWidth), 
                              int((y-ScrollY)*TileHeight),
                              TileWidth, 
                              TileHeight);
                sa.intersect(gamearea);
                clip(gc, sa);
                draw_field(gc, f, x,y);
                f.redrawp = false;
                screen->update_rect(sa);
            }
        }
    }
    draw_rubbers (gc);
    if (status_bar->has_changed() || RedrawEverything) {
        status_bar->redraw (gc, inventoryarea);
        screen->update_rect(inventoryarea);
    }
    if (options::ShowFPS) {
        char fps[20];
        sprintf (fps,"fps: %d\n", int(1000.0/(SDL_GetTicks()-last_frame_time)));
        last_frame_time = SDL_GetTicks();
        Font *f = enigma::GetFont("levelmenu");

        gc.noclip();
        Rect area(0,0,80,20);
        set_color(gc, 0,0,0);
        box(gc, area);
        f->render(gc, 0,0, fps);
            
        screen->update_rect(area);
    }
    screen->flush_updates();
    RedrawEverything = false;
}
