/*
 * Copyright (C) 2002 Daniel Heck
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *  
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *  
 * You should have received a copy of the GNU General Public License along
 * with this program; if not, write to the Free Software Foundation, Inc.,
 * 59 Temple Place, Suite 330, Boston, MA 02111-1307, USA.
 */

#include "gui.hh"
#include "enigma.hh"
#include "sound.hh"
#include "video.hh"
#include "px/px.hh"
#include <memory>

using namespace gui;
using namespace px;

//----------------------------------------
// Widget implementation.
//----------------------------------------
Widget::Widget(Widget *parent) 
    : area(), menu(0), m_parent(parent) 
{}

void 
Widget::redraw() 
{ 
    if (menu) menu->invalidate_area(get_area());
}

void 
Widget::redraw(const px::Rect &r) 
{ 
    if (menu) menu->invalidate_area(r);
}

void
Widget::invoke_listener()
{
    if (listener)
        listener->on_action(this);
}

void Image::draw (px::GC &gc, const px::Rect &r)
{
    if (px::Surface *s = enigma::GetImage(imgname.c_str()))
        blit(gc, get_x(), get_y(), s);
}


//----------------------------------------
// Container implementation
//----------------------------------------
Container::~Container()
{
    delete_sequence(m_widgets.begin(), m_widgets.end());
}

void 
Container::add_child (Widget *w) 
{ 
    if (w) {
        m_widgets.push_back(w);
        w->set_parent(this);
        w->move (get_x()+w->get_x(),
                 get_y()+w->get_y());
    }
}


void 
Container::draw (px::GC& gc, const px::Rect &r)
{
    for (iterator i=begin(); i!=end(); ++i) {
        Widget *w = *i;
        Rect rr = intersect(r, w->get_area());
        clip(gc, rr);
        w->draw(gc,rr);
    }
}

Widget *
Container::find_widget(int x, int y)
{
    for (iterator i=begin(); i!=end(); ++i) {
        Widget *w = *i;
        if (w->get_area().contains(x,y))
            return w;
    }
    return 0;
}

void 
Container::redraw_child (Widget *w, const px::Rect &r)
{
}

void
Container::move (int x, int y)
{
    int dx = x-get_x();
    int dy = y-get_y();

    for (iterator i=begin(); i!=end(); ++i) {
        Widget *w = *i;
        w->move(dx + w->get_x(), dy+w->get_y());
    }
}

px::Rect
Container::boundingbox()
{ 
    if (!m_widgets.empty()) {
        iterator i=begin();
        Rect bbox=(*i)->get_area();
        for (++i; i!=end(); ++i) 
            bbox = px::boundingbox(bbox, (*i)->get_area());
        return bbox;
    } else
        return get_area();
}


//----------------------------------------
// Button implementation
//----------------------------------------
void 
Button::draw(px::GC &gc, const px::Rect &r) 
{
    char bordername[50];
    sprintf(bordername, "b%dx%d%s", get_w(), get_h(),
            highlighted ? "h" : "");
    if (px::Surface *s = enigma::GetImage(bordername))
        blit(gc, get_x(), get_y(), s);
}

//----------------------------------------
// PushButton implementation
//----------------------------------------

PushButton::PushButton() 
: pressed(false) 
{}

bool
PushButton::on_event(const SDL_Event &e)
{
    switch (e.type) {
    case SDL_MOUSEBUTTONDOWN:
        pressed=true;
        redraw();
        return true;
    case SDL_MOUSEBUTTONUP:
        if (pressed)
        {
            pressed = false;
            redraw();
            sound::PlaySound("menuok");
            invoke_listener();
        }
        return true;
    default:
        return false;
    }
}

void 
PushButton::deactivate() 
{
    pressed=false; 
    redraw(); 
    Button::deactivate(); 
}


//----------------------------------------
// TextButton implementation.
//----------------------------------------
px::Font *TextButton::menufont = 0;
px::Font *TextButton::menufont_pressed = 0;

TextButton::TextButton(const string &t, ActionListener *al) 
    : text(t)
{
    if (menufont == 0) {
        menufont = enigma::GetFont("menufont"); //menufont");
        menufont_pressed = enigma::GetFont("menufontsel");
    }
    set_listener(al);
}

void 
TextButton::set_text(const std::string &t)
{
    text = t;
    redraw();
}

void 
TextButton::draw(px::GC &gc, const px::Rect &r) 
{
    Button::draw(gc,r);
    Font *f = is_pressed() ? menufont_pressed : menufont;
    int h = f->get_height();
    int w = f->get_width(text.c_str());
    int x = get_x() + (get_w()-w)/2;
    int y = get_y() + (get_h()-h)/2;

    f->render (gc, x, y, text.c_str());
}

//----------------------------------------
// ImageButton implementation.
//----------------------------------------
ImageButton::ImageButton(const std::string &unselected,
                         const std::string &selected)
    : fname_sel(selected),
      fname_unsel(unselected)
{}

void
ImageButton::draw(px::GC &gc, const px::Rect &r)
{
    Button::draw(gc, r);
    string &fname = is_pressed() ? fname_sel : fname_unsel;
        
    if (Surface *s = enigma::GetImage(fname.c_str())) {
        int w=s->width();
        int h=s->height();
        int x = get_x() + (get_w()-w)/2;
        int y = get_y() + (get_h()-h)/2;
        blit(gc, x, y, s);
    }
}

//----------------------------------------
// Menu implementation.
//----------------------------------------

Menu::Menu() 
: active_widget(0)
, quitp(false)
, abortp(false) 
{}

Menu::~Menu() 
{}

void 
Menu::add(Widget *w) 
{
    Container::add_child(w); w->set_menu(this); 
}

void 
Menu::add(Widget *w, px::Rect r) 
{
    w->set_area(r);
    add(w);
}


void Menu::quit() { quitp=true; }
void Menu::abort() { abortp=true; }

void 
Menu::invalidate_area(const px::Rect &r)
{
    dirtyrects.add(r);
}

void 
Menu::invalidate_all()
{
    dirtyrects.clear();
    dirtyrects.push_back(screen->size());
}


bool 
Menu::manage(px::Screen *scr)
{
    screen=scr;
    draw_all(scr);
    quitp=abortp=false;
    while (!(quitp || abortp)) {
        screen->flush_updates();
        SDL_Event e;
        while (SDL_PollEvent(&e)) {
            handle_event(e);
        }
        if (!dirtyrects.empty()) {
            RectList::iterator i = dirtyrects.begin();
            RectList::iterator end = dirtyrects.end();
            for (; i!=end; ++i)
                redraw_area(*i);
            dirtyrects.clear();
        }
	SDL_Delay(10);
    }
    return !abortp;
}

void 
Menu::handle_event(const SDL_Event &e)
{
    if (on_event(e))
        return;
    
    switch (e.type) {
    case SDL_QUIT:
        abort();
        break;
    case SDL_MOUSEMOTION:
        track_active_widget( e.motion.x, e.motion.y );
        break;
    case SDL_KEYDOWN:
        if (e.key.keysym.sym == SDLK_ESCAPE) {
            abort();
        }
        else if (active_widget) {
            active_widget->on_event(e);
        }
        break;
    case SDL_MOUSEBUTTONDOWN:
    case SDL_MOUSEBUTTONUP:
        track_active_widget( e.button.x, e.button.y );
        if (active_widget) active_widget->on_event(e);
        break;
    default:
        if (active_widget) active_widget->on_event(e);
    }
}

void
Menu::track_active_widget( int x, int y )
{
    Widget *w = find_widget(x,y);
    if (w != active_widget) 
    {
        if (active_widget) active_widget->deactivate();
        if (w) w->activate();
        active_widget = w;
    }
}


void 
Menu::center(px::Screen *scr)
{
    if (m_widgets.size() > 0) 
    {
        using std::min;
        using std::max;

        Rect a = m_widgets[0]->get_area();
        for (unsigned i=1; i<m_widgets.size(); ++i)
        {
            Rect r = m_widgets[i]->get_area();
            a.x = min(r.x, a.x);
            a.y = min(r.y, a.y);
            a.w += max(0, r.x+r.w-a.x-a.w);
            a.h += max(0, r.y+r.h-a.y-a.h);
        }
        Rect c=px::center(scr->size(), a);
        int dx = c.x-a.x;
        int dy = c.y-a.y;

        for (unsigned i=0; i<m_widgets.size(); ++i)
        {
            Rect r = m_widgets[i]->get_area();
            r.x += dx;
            r.y += dy;
            m_widgets[i]->set_area(r);
        }
    }    
}


void 
Menu::redraw_area(const px::Rect &r)
{
    video::HideMouse();
    GC gc(screen->get_drawable());
    clip(gc, r);
    draw_background(gc);

    Container::draw(gc,r);

    screen->update_rect(r);
    video::ShowMouse();
}

void 
Menu::draw_all(px::Screen *scr)
{
    redraw_area(scr->size());
    scr->update_all();
}
