//======================================================================
// Copyright (C) 2002 Daniel Heck
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either version 2
// of the License, or (at your option) any later version.
//  
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//  
// You should have received a copy of the GNU General Public License along
// with this program; if not, write to the Free Software Foundation, Inc.,
// 59 Temple Place, Suite 330, Boston, MA 02111-1307, USA.
//======================================================================
#ifndef GUI_HH_INCLUDED
#define GUI_HH_INCLUDED

#include "px/pxfwd.hh"
#include "px/geom.hh"
#include "SDL.h"
#include <vector>
#include <string>

//======================================================================
// EVENTS AND EVENT HANDLERS
//======================================================================
namespace gui
{
    class Widget;

    class ActionListener {
    public:
        virtual ~ActionListener() {}
        virtual void on_action(Widget *w) {};
    };
}

//======================================================================
// GUI WIDGETS
//======================================================================
namespace gui
{
    class Menu;
    class Container;

    class Widget : public ActionListener {
    public:
        virtual ~Widget() {}

        virtual bool has_changed() const { return false; }
        virtual void draw(px::GC &gc, const px::Rect &r) = 0;
        virtual void activate() {}
        virtual void deactivate() {}

        virtual void move(int x, int y) { area.x = x; area.y = y; }

        virtual bool on_event(const SDL_Event &e) { return false; }

        void set_size(int w, int h) {area.w = w; area.h = h;}

        px::Rect get_area() const { return area; }
        void set_area(const px::Rect &r) { area = r; }
        int get_x() const { return area.x; }
        int get_y() const { return area.y; }
        int get_w() const { return area.w; }
        int get_h() const { return area.h; }
        Menu * get_menu() const { return menu; }

        void set_parent(Widget *parent) { m_parent = parent; }

        void set_listener(ActionListener *al) {
            listener = al; 
        }
    protected:
        Widget(Widget *parent=0);
        void redraw();
        void redraw(const px::Rect &r);

        void invoke_listener();
    private:

        friend class Menu;
        void set_menu(Menu *m) { menu=m; }

        px::Rect        area;
        Menu           *menu;
        Widget         *m_parent;
        ActionListener *listener;
    };

//----------------------------------------
// List, HList, VList
//----------------------------------------

    class Container : public Widget {
    public:
        Container() {}
        ~Container();
        void add_child (Widget *w);
        void redraw_child (Widget *w, const px::Rect &r);

        // Widget interface.
        void draw (px::GC& gc, const px::Rect &r);
        void move (int x, int y);

        Widget *find_widget(int x, int y);

    protected:
        typedef std::vector<Widget *> WidgetList;
        typedef WidgetList::iterator iterator;

        iterator begin() { return m_widgets.begin(); }
        iterator end() { return m_widgets.end(); }
        WidgetList m_widgets;

    private:
        px::Rect boundingbox();
    };

    class List : public Container {
    public:
        void set_spacing (int pixels);
    protected:
        List(int spacing=5)
            : m_spacing(spacing)
        {}
    private:
        int m_spacing;          // # of pixels between container items
    };

    class HList : public List {
    public:
        HList() : List() {}
        void add_back(Widget *w) {
            Widget *lastw = 0;
            if (m_widgets.size() > 0)
                lastw = m_widgets[m_widgets.size()-1];
        }
    private:

    };

    class VList : public List {
    public:

    private:

    };

    class Image : public Widget {
    public:
        Image (const std::string &iname) : imgname(iname) {}
        void draw (px::GC &gc, const px::Rect &r);
    private:
        std::string imgname;
    };

    class Button : public Widget {
    protected:
        Button() : highlighted(false) {}

        void draw(px::GC &gc, const px::Rect &r);
    protected:
        // Widget interface.
        void activate() { highlighted=true; redraw(); }
        void deactivate() { highlighted=false; redraw(); }
    private:

        bool highlighted;
    };


    class PushButton : public Button {
    public:
        PushButton();
        bool is_pressed() { return pressed; }
    protected:
        bool on_event(const SDL_Event &e);
        void deactivate();
    private:
        bool pressed;
    };

//----------------------------------------
// TextButton.
//----------------------------------------
    class TextButton : public PushButton {
    public:
        TextButton(const std::string &t, ActionListener *al=0);
        void set_text(const std::string &t);
    private:
        // Widget interface.
        void draw(px::GC &gc, const px::Rect &r);

        // Variables.
        static px::Font *menufont, *menufont_pressed;

        std::string text;
    };

//----------------------------------------
// ImageButton.
//----------------------------------------
    class ImageButton : public PushButton {
    public:
        ImageButton(const std::string &unselected,
                    const std::string &selected);
    private:
        // Widget interface.
        void draw(px::GC &gc, const px::Rect &r);
        std::string fname_sel, fname_unsel;
    };

    class Menu : public Container {
    public:
        Menu();
        virtual ~Menu();

        /* true: ok, false: menu aborted by user */
        bool manage(px::Screen *scr);

        void add(Widget *w);
        void add(Widget *w, px::Rect r);
        void center(px::Screen *scr);

        void draw_all(px::Screen *scr);

        void quit();
        void abort();

        void invalidate_area(const px::Rect &r);
        void invalidate_all();

    private:
        void handle_event(const SDL_Event &e);
        void redraw_area(const px::Rect &r);
        void track_active_widget( int x, int y );
        
        // Menu interface.
        virtual void draw_background(px::GC &gc) {}
        
        // Variables.
        Widget *active_widget;
        bool quitp, abortp;
        px::RectList dirtyrects;
        px::Screen *screen;
    };
}
#endif
