/*
 * Copyright (C) 2002,2003 Daniel Heck
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *  
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *  
 * You should have received a copy of the GNU General Public License along
 * with this program; if not, write to the Free Software Foundation, Inc.,
 * 59 Temple Place, Suite 330, Boston, MA 02111-1307, USA.
 *
 * $Id: items.cc,v 1.8 2003/02/02 18:20:49 dheck Exp $
 */

#include "items.hh"
#include "game.hh"
#include "object_mixins.hh"
#include "display.hh"
#include "sound.hh"
#include <string>
#include <vector>
#include <algorithm>
#include <math.h>

using namespace world;
using std::string;

//----------------------------------------
// Item implementation
//----------------------------------------
Item::Item(const char *name) 
    : TGridObject<GRID_ITEMS>(name)
{}

string 
Item::get_inventory_model() 
{
    return get_kind();
}

void Item::stone_change (Stone *st) {}
void Item::on_stonehit (Stone *st) {}
void Item::on_drop (Actor *a) {}
void Item::on_pickup (Actor *a) {}

ItemAction Item::activate(Actor* a, GridPos p)
{ return ITEM_DROP; }


px::V3 Item::get_force(Actor *a)
{ return px::V3(); }

bool 
Item::actor_hit(Actor *actor)
{
    const double ITEM_RADIUS = 0.3;
    px::V3 item_center(get_pos().x + 0.5,
                       get_pos().y + 0.5,
                       0);
    double dist = length(actor->get_pos()-item_center);
    return (dist < ITEM_RADIUS);
}

//----------------------------------------
// Various simple items
//----------------------------------------
#define DEF_ITEM(classname, kindname)           \
    class classname : public Item {             \
        CLONEOBJ(classname);                    \
    public: classname() : Item(kindname) {}     \
    }

namespace
{
    DEF_ITEM(MagicWand, "it-magicwand");
    DEF_ITEM(Umbrella, "it-umbrella");
    DEF_ITEM(Floppy, "it-floppy");
    DEF_ITEM(Brush, "it-brush");

    class Sword : public Item {
        CLONEOBJ(Sword);
        bool on_laserhit(Direction d) {
            play_sound("st-magic");
            world::SetItem(get_pos(), MakeItem("it-hammer"));
            return false; 
        }
    public:
        Sword() : Item("it-sword") {}
    };

    class Hammer : public Item {
        CLONEOBJ(Hammer);
        bool on_laserhit(Direction d) {
            play_sound("st-magic");
            world::SetItem(get_pos(), MakeItem("it-sword"));
            return false; 
        }
    public:
        Hammer() : Item("it-hammer") {}
    };

    class ExtraLife : public Item {
        CLONEOBJ(ExtraLife);
        std::string get_inventory_model() { 
            if (player::CurrentPlayer()==0)
                return "inv-blackball";
            else
                return "inv-whiteball";
        }
    public:
        ExtraLife() : Item("it-extralife") {}
    };

    class Key : public Item {
        CLONEOBJ(Key);
    public:
    	Key() : Item("it-key") { set_attrib("keycode", 0.0); }
    };
    class Key_a : public Item {
        CLONEOBJ(Key_a);
    public:
    	Key_a() : Item("it-key_a") { set_attrib("keycode", 1.0); }
    };
    class Key_b : public Item {
        CLONEOBJ(Key_b);
    public:
    	Key_b() : Item("it-key_b") { set_attrib("keycode", 2.0); }
    };
    class Key_c : public Item {
        CLONEOBJ(Key_c);
    public:
    	Key_c() : Item("it-key_c") { set_attrib("keycode", 3.0); }
    };
}


//----------------------------------------
// Coins
//
// :value    1,2,4: how many time-units this coin buys
//----------------------------------------
namespace
{
    class Coin : public Item {
        CLONEOBJ(Coin);
    public:
        Coin() : Item("it-coin") { set_attrib("value", 1.0); }
    private:
        void init_model() { set_model(get_inventory_model()); }

        bool on_laserhit(Direction d) {
            play_sound("st-magic");
            switch(get_value()) {
            case 1: SetItem (get_pos(), MakeItem("it-umbrella")); break;
            case 2: SetItem (get_pos(), MakeItem("it-hammer")); break;
            default: SetItem (get_pos(), MakeItem("it-extralife")); break;
            }
            return false;
        }

        void on_stonehit(Stone *st) {
            switch(get_value()) {
            case 1: set_attrib("value", 2.0); init_model(); break;
            case 2: set_attrib("value", 4.0); init_model(); break;
            default: break;
            }
        }

        string get_inventory_model() 
        {
            switch(get_value()) {
            case 1: return "it-coin1";
            case 2: return "it-coin2";
            default: return "it-coin4";
            }
        }

        int get_value() { 
            int value = int_attrib("value"); 
//            assert (value==1 || value==2 || value==4);
            return value;
        }
    };
}

//----------------------------------------
// Hollows & Hills
//----------------------------------------

/** \page it-hills Hills and Hollows

Hills and hollows are placed on the floor and can
make the movement difficult.

\subsection hillsm Messages
- \b trigger	will convert a hill to a hollow and vice versa
- \b shovel     decreases the size of the hill/hollow

\image html it-hill.png
*/
namespace
{
    class HillHollow : public Item {
    protected:
        enum Type { HILL, HOLLOW, TINYHILL, TINYHOLLOW };
        
        HillHollow(const char *name, Type t);

        void transmute(Type newtype);
        V3 vec_to_center (V3 v);
        double get_radius() const { return m_radius[m_type]; }

    private:
        double get_forcefac() const { return m_forcefac[m_type]; }
        void shovel();
        
        // Item interface
        px::V3 get_force(Actor *a);
        bool actor_hit(Actor *a) { return false; }
        void on_stonehit(Stone *st);

        // Object interface.
        void message(const string &m, const Value &);

        // Variables.
        static double m_radius[4], m_forcefac[4];
        Type m_type;
    };
}

double HillHollow::m_radius[4] = {0.5, 0.5, 0.3, 0.3};
double HillHollow::m_forcefac[4] = {90,-90, 90, -90};


HillHollow::HillHollow(const char *name, Type t) 
: Item(name), m_type(t) 
{}

void HillHollow::on_stonehit(Stone *st)
{
    shovel();
}

void HillHollow::shovel()
{
    if (m_type==HOLLOW) 
        transmute (TINYHOLLOW);
    else if (m_type==HILL) 
        transmute (TINYHILL);
    else
        KillItem(get_pos());
}


void HillHollow::message(const string &m, const Value &)
{
    if (m=="trigger") {
        Type flippedkind[] = {HOLLOW,HILL, TINYHOLLOW,TINYHILL};
        transmute(flippedkind[m_type]);
    }
    else if (m=="shovel") 
        shovel();
}

V3 HillHollow::vec_to_center (V3 v)
{
    GridPos p = get_pos();
    return v - V3(p.x+0.5, p.y+0.5,0);
}

V3 HillHollow::get_force(Actor *a) 
{
    V3 v = vec_to_center(a->get_pos());
    if (length(v) <= get_radius()) 
        return get_forcefac()*v;
    else
        return px::V3(); 
}

void HillHollow::transmute(Type newtype)
{
    m_type = newtype;
    static char *modelname[] = { "it-hill", "it-hollow",
                                 "it-tinyhill", "it-tinyhollow" };
    set_model(modelname[m_type]);
}

namespace
{
    class Hill : public HillHollow {
        CLONEOBJ(Hill);
    public:
        Hill() : HillHollow("it-hill", HILL) {}
    };
    class TinyHill : public HillHollow {
        CLONEOBJ(TinyHill);
    public:
        TinyHill() : HillHollow("it-tinyhill", TINYHILL) {}
    };

    class TinyHollow : public HillHollow {
        CLONEOBJ(TinyHollow);
    public:
        TinyHollow() : HillHollow("it-tinyhollow", TINYHOLLOW) {}
    };

    /*
     * Hollows are special in that they can end the current level
     * if they have each a small white marble inside them.
     */
    class Hollow : public HillHollow {
        INSTANCELISTOBJ(Hollow);
    public:
        Hollow();
    private:        
        // Item interface.
        bool actor_hit(Actor *a);

        // Functions.
        bool near_center_p (Actor *a);
        void check_if_level_finished();
        
        // Variables.
        Actor *whiteball;       // The small white ball that is currently being tracked
        Uint32 enter_time;      // ... when did it enter the hollow?
    };
}

Hollow::InstanceList Hollow::instances;

Hollow::Hollow() : HillHollow("it-hollow", HOLLOW), whiteball(0) 
{}

bool 
Hollow::near_center_p (Actor *a) 
{
    return (length(vec_to_center(a->get_pos())) < get_radius()/2);
}

bool 
Hollow::actor_hit(Actor *a)
{
    if (whiteball==0 && 0==strcmp(a->get_kind(),"ac-whiteball-small")
        && near_center_p(a))
    {
        whiteball = a;
        enter_time = SDL_GetTicks();
    }
    else if (whiteball==a)
    {
        if (!near_center_p(a))
            whiteball=0;
        else
            check_if_level_finished();
    }
    return false;
}

/* If every hollow in the level (a) contains a small white ball and
   (b) this ball has been inside the hollow for at least MINTIME
   milliseconds, finish the level. */
void 
Hollow::check_if_level_finished() 
{
    const unsigned MINTIME=1000; 
    for (unsigned i=0; i<instances.size(); ++i) {
        Hollow *h=(Hollow*) instances[i];
        if (!h->whiteball || SDL_GetTicks() - h->enter_time < MINTIME)
            return;
    }
    enigma::FinishLevel();
}

//----------------------------------------
// Springs
//----------------------------------------

/** \page it-spring Spring

Activating a spring will make the marble jump.
A jumping marble does not fall into abyss or water.

Springs come in two flavors: it-spring1 stays in the inventory,
whereas it-spring2 drops to the floor when you activate it.

\image html it-spring1.png
*/
namespace
{
    class Spring1 : public Item {
        CLONEOBJ(Spring1);
    public:
        Spring1() : Item("it-spring1") {}
    private:
        ItemAction activate(Actor *a, GridPos)
        {
            SendMessage(a, "jump");
            return ITEM_KEEP;
        }
    };

    class Spring2 : public Item {
        CLONEOBJ(Spring2);
    public:
        Spring2() : Item("it-spring2") {}
    private:
        ItemAction activate(Actor *a, GridPos)
        {
            SendMessage(a, "jump");
            return ITEM_DROP;
        }
    };
}

//----------------------------------------
// Explosion
//----------------------------------------
namespace
{
    class Explosion : public Item {
        CLONEOBJ(Explosion);
    public:
        enum Strength { 
            WEAK,               // Explode but do nothing else.
            MEDIUM,             // Explode and leave a hole in the ground.
            STRONG              // Explode and shatter the floor.
        };
        Explosion(Strength st=WEAK);

    private:
        void init_model() {set_anim("expl");}
        bool actor_hit(Actor *actor) {
            SendMessage(actor, "shatter");
            return false;
        }
        void animcb();

        // Variables.
        Strength strength;
    };
}

Explosion::Explosion(Strength st) 
    : Item("it-explosion"), strength(st) 
{}


void 
Explosion::animcb() 
{
    switch (strength) {
    case WEAK:
        world::KillItem(get_pos());
        break;
    case MEDIUM:
        SetItem(get_pos(), MakeItem("it-hollow"));
        break;
    case STRONG:
//        SendMessage(GetFloor(get_pos()), "explode");
//        world::KillItem(get_pos());
        SetItem(get_pos(), MakeItem("it-debris"));
        break;
    }
}


//----------------------------------------
// Document.
//----------------------------------------

/** \page it-document Document Item

This item looks like a piece of paper and contains text messages
that can be displayed by activating the item.
 
\subsection documenta Attributes
 
\b text:     The message to be displayed.
 
\subsection documente Example
\verbatim
set_item("it-document", 1,1, {text="Hello World!"})
\endverbatim

or, equivalently,
\verbatim
document(1,1, "Hello World")
\endverbatim
\image html it-document.png
*/

namespace
{
    class Document : public Item {
        CLONEOBJ(Document);
    public:
        Document() : Item("it-document") {
            set_attrib("text", "");
        }

        ItemAction activate(Actor *a, GridPos)
        {
            string txt;
            if (string_attrib ("text", &txt))
                display::GetStatusBar()->show_text(txt);
            return ITEM_KILL;	       // remove from inventory
        }
        void message(const string &msg, const Value &val) {
            if (msg == "ignite")
                SetItem(get_pos(), new Explosion(Explosion::WEAK));
        }
    };
}

//----------------------------------------
// Dynamite.
//----------------------------------------
namespace
{
    class Dynamite : public Item {
        CLONEOBJ(Dynamite);
    public:
        Dynamite() : Item("it-dynamite"), state(IDLE) {}
    private:
        enum State { IDLE, BURNING };
        State state;

        void change_state(State newstate) {
            if (newstate==BURNING && state==IDLE) {
                state = BURNING;
                set_anim("it-dynamite-burning");
            }
        }

        void animcb() { explode(); }
        void explode ();

        void message(const string &msg, const Value &val) {
            if (msg == "ignite")
                change_state(BURNING);
            else if (msg == "explode")
                explode();
        }

        bool on_laserhit(Direction) {
            change_state(BURNING);
            return false;           // block light
        }

        void on_drop(Actor *a) {change_state(BURNING);}

        bool actor_hit(Actor *a) {
            // don't pick up burning dynamite
            return (state == IDLE); 
        }
    };
}

void 
Dynamite::explode() 
{
    GridPos p = get_pos();

    // ignite nearby objects
    SendMessage(GetItem(move(p,enigma::NORTH)), "ignite");
    SendMessage(GetItem(move(p,enigma::SOUTH)), "ignite");
    SendMessage(GetItem(move(p,enigma::EAST)), "ignite");
    SendMessage(GetItem(move(p,enigma::WEST)), "ignite");

    play_sound("explosion2");
    SetItem(p, new Explosion(Explosion::MEDIUM));
}

//----------------------------------------
// BlackBomb.
//----------------------------------------

/** \page it-blackbomb Black Bomb

When black bombs explode, they destroy the floor tile underneath them.

\image html it-blackbomb.png
*/

namespace
{
    class BlackBomb : public Item  {
        CLONEOBJ(BlackBomb);
    public:
        BlackBomb() : Item("it-blackbomb"), state(IDLE) {}

    private:
        enum State { IDLE, BURNING } state;

        void animcb() { explode (); }
        void explode();
        void change_state(State newstate) {
            if (newstate==BURNING && state==IDLE) {
                state = BURNING;
                set_anim("it-blackbomb-burning");
            }
        }

        void message(const string &msg, const Value &val) {
            if (msg == "ignite")
                change_state(BURNING);
            else if (msg == "explode")
                explode();
        }

        bool on_laserhit(Direction) {
            change_state(BURNING);
            return false;           // block light
        }
        bool actor_hit(Actor *a) { return false; }
    };
}

void 
BlackBomb::explode() 
{
    GridPos p = get_pos();
    SendMessage(GetItem(move(p,enigma::NORTH)), "ignite");
    SendMessage(GetItem(move(p,enigma::SOUTH)), "ignite");
    SendMessage(GetItem(move(p,enigma::EAST)), "ignite");
    SendMessage(GetItem(move(p,enigma::WEST)), "ignite");

    play_sound("explosion1");
    SetItem(p, new Explosion(Explosion::STRONG));
}

//----------------------------------------
// Trigger.
//----------------------------------------
namespace
{
    class Trigger : public Item {
        CLONEOBJ(Trigger);
    public:
        Trigger() : Item("it-trigger"), objectcnt(0), have_stone(false) {
            set_attrib("invisible", 0.0);
        }
    private:
        int objectcnt;          // number of objects on the trigger
	bool have_stone;

        bool actor_hit(Actor *a) { return false; }


        void init_model() {
            if (int_attrib("invisible"))
                set_model("invisible");
            else if (objectcnt > 0)
                set_model("it-trigger1");
            else
                set_model("it-trigger");
        }

        void add_object() {
            if (++objectcnt == 1) {
                init_model();
                world::PerformAction(this, true);
                play_sound("it-triggerdown");
            }
        }
        void remove_object() {
            if (--objectcnt == 0) {
                init_model();
                world::PerformAction(this, false);
                play_sound("it-triggerup");
            }
        }

        void actor_enter(Actor *) {
            add_object();
        }
        void actor_leave(Actor *) {
            remove_object();
        }
        void stone_change(Stone *st);
    };
}

void 
Trigger::stone_change(Stone *st) 
{
    if (st == 0 && have_stone){
        have_stone=false;
        remove_object();
    }
            
    if (st != 0 && !have_stone){
        have_stone=true;
        add_object();
    }
}

//----------------------------------------
// Seed
//----------------------------------------
namespace
{
    class Seed : public Item {
        CLONEOBJ(Seed);
        enum State { IDLE, GROWING } state;

        bool actor_hit(Actor *a) { return state==IDLE; }
        void on_drop(Actor *a) { start_growing(); }
        void on_stonehit(Stone *) { start_growing(); }
        
        void start_growing() {
            state=GROWING; 
            set_anim("it-seed-growing");
        }

        void animcb() {
            Stone *st = world::MakeStone("st-wood-growing");
            world::SetStone(get_pos(), st);
            world::KillItem(get_pos());
        }

    public:
        Seed() : Item ("it-seed"), state(IDLE) {}

    };
}


//----------------------------------------
// ShogunDot
//----------------------------------------

/** \page it-shogundot Shogun Dot

\subsection shogundota Attributes
- \b size:            1..3  (smallest..largest)
- \b target, \b action:   as usual
*/
namespace
{
    class ShogunDot : public Item {
        CLONEOBJ(ShogunDot);
    public:
        ShogunDot();
    private:
        void set_size(int s) { set_attrib("size", s); }
        int get_size() const;

        void init_model();
        void message(const string &str, const Value &v);
        void stone_change(Stone *st);
        bool actor_hit(Actor *a) { return false; }


        // Variables.
        bool activated;
    };
}

ShogunDot::ShogunDot() 
    : Item("it-shogun"), activated(false) 
{
    set_size(1);
}

int 
ShogunDot::get_size() const 
{ 
    int size = int_attrib("size"); 
    assert(1<=size && size<=3);
    return size;
}

void
ShogunDot::init_model() 
{
    char *models[] = {"it-shogun1", "it-shogun2", "it-shogun3"};
    set_model(models[get_size()-1]);
}

void 
ShogunDot::stone_change(Stone *st) 
{
    if (activated && st == 0) {
        activated = false;
        world::PerformAction(this, false);
    } 
}

#define streq(a,b) (strcmp((a), (b))==0)
void 
ShogunDot::message(const string &str, const Value &v) 
{
    int size=get_size();
    const char *s = str.c_str();

    if (activated && streq(s, "noshogun")) {
        activated=false;
        world::PerformAction(this, false);
    }
    else if ((size==1 && streq(s,"shogun1")) ||
             (size==2  && streq(s,"shogun2")) ||
             (size==3  && streq(s,"shogun3")))
    {
        activated=true;
        world::PerformAction(this, true);
    }
}

//----------------------------------------
// Magnet
//----------------------------------------
namespace
{

    class Magnet : public OnOffItem {
        class Magnet_FF : public world::ForceField {
        public:
            Magnet_FF() : m_active(false), strength(30) {}

            void set_pos(GridPos p) { center = to_vec(p) + V3(0.5,0.5,0); }

            V3 get_force(Actor *a, V3 x, V3 v, double time) 
            {
                if (!m_active)
                    return V3();
                V3 dv = center - x;
                double dist = length(dv);
                if (dist < 0.1)
                    return V3();
                else
                    return strength * dv/(dist*dist);
            }

            bool   m_active;
            V3     center;
            double strength;
        };

        
        CLONEOBJ(Magnet);
    public:
        Magnet() : OnOffItem ("it-magnet") {
            set_attrib("strength",30.0);
        }
    private:
        void init_model() {
            if (is_on())
                set_model("it-magnet-on");
            else
                set_model("it-magnet-off");
        }
        void on_creation() {
            ff.m_active = is_on();
            ff.set_pos(get_pos());
            world::AddForceField(&ff);
            Item::on_creation();
        }
        void on_removal() {
            world::RemoveForceField(&ff);
            Item::on_removal();
        }
        bool actor_hit(Actor *a) { return false; }

        virtual void notify_onoff(bool on) {
            ff.m_active = on;
        }

        
        Magnet_FF ff;
    };
}

//----------------------------------------
// WormHole
//----------------------------------------

/** \page it-wormhole Worm hole

Worm holes teleport actors to another place.  They have a variable
attracting force field.

\subsection wormholea Attributes
- \b targetx:       X coordinate of the destination
- \b targety:       Y coordinate of the destination
- \b strength:      Strength of the force field (default: 50)
- \b range:         Range of the force field

\subsection wormholee Example
\verbatim
set_item("it-wormhole", 1,1, {targetx=5.5, targety=10.5, strength=50, range=5})
\endverbatim
*/

namespace
{
    class WormHole_FF : public world::ForceField {
    public:
        WormHole_FF() : strength(50), range(1000) {}

        void set_pos(GridPos p) { center = to_vec(p) + V3(0.5,0.5,0); }

        V3 get_force(Actor *a, V3 x, V3 v, double time) 
        {
            V3 b = center - x;
            double bb = length(b);
            if (bb>0 && bb < range) {
                return strength*b/(bb*bb);
            }
            return V3();
        }

        V3     center;          // Center of the force field
        double strength;        // Strength of the force
        double range;           // Range of the force
    };

    class WormHole : public Item, public enigma::TimeHandler 
    {
        CLONEOBJ(WormHole);
    public:
        WormHole() : Item("it-wormhole") {
            set_attrib("targetx", 9.5);
            set_attrib("targety", 9.5);
            set_attrib("strength",50.0);
            set_attrib("range", 1000.0);
        }
    private:
        void on_creation() {
            Item::on_creation();
            ff.set_pos(get_pos());
            double_attrib("strength", &ff.strength);
            double_attrib("range", &ff.range);
            world::AddForceField(&ff);
        }
        void on_removal() {
            world::RemoveForceField(&ff);
            Item::on_removal();
        }

        bool actor_hit(Actor *a);

        V3 get_center() const {
            GridPos p = get_pos();
            return V3(p.x+0.5, p.y+0.5,0);
        }
        V3 vec_to_center (V3 v) {return v-get_center();}

        bool near_center_p (Actor *a) {
            return (length(vec_to_center(a->get_pos())) < 0.5/4);
        }
        
        // Variables.
        WormHole_FF ff;
    };
}

bool 
WormHole::actor_hit(Actor *actor)
{
    if (near_center_p(actor)) {
        AddSprite(display::SPRITE_EFFECT, get_center(), "ring-anim");
        double x, y;
        if (double_attrib("targetx", &x) && double_attrib("targety", &y)) {
            play_sound("warp");
            world::WarpActor(actor, x, y);
        }
    }
    return false;
}


//----------------------------------------
// YinYang item
//----------------------------------------
namespace
{
    class YinYang : public Item {
        CLONEOBJ(YinYang);
    public:
        YinYang() : Item("it-yinyang") {}
        
        string get_inventory_model() {
            if (player::CurrentPlayer()==0)
                return  "it-yinyang";
            else
                return "it-yanying";
        }

        ItemAction activate(Actor *a, GridPos) {
            // Switch to other marble 
            player::SwapPlayers();
            sound::PlaySound ("switch");
            return ITEM_KEEP;
        }
    };
}

//----------------------------------------
// Spade
//----------------------------------------
namespace
{
    class Spade : public Item {
        CLONEOBJ(Spade);
    public:
        Spade() : Item("it-spade") {}

        ItemAction activate(Actor *, GridPos p) {
            if (Item *it=GetItem(p)) {
                SendMessage(it, "shovel");
                return ITEM_KEEP;
            }
            return ITEM_DROP;
        }
    };
}

//----------------------------------------
// Pipes
//----------------------------------------
namespace
{
    class Pipe : public Item {
        CLONEOBJ(Pipe);
    public:
        Pipe() : Item("it-pipe") {set_attrib("type", 1);}
        
        string get_inventory_model() {
            char modelname[20];
            sprintf(modelname, "it-pipe%d", int_attrib("type"));
            return modelname;
        }
        void init_model() {
            set_model(get_inventory_model());
        }
    };
}

//----------------------------------------
// Cracks
//----------------------------------------
namespace
{
    class Crack : public Item {
        CLONEOBJ(Crack);
    public:
        Crack() : Item("it-crack") {set_attrib("type", 0.0);}
    private:
        int get_type() const { return int_attrib("type"); }

        void init_model() {
            if (int t=get_type()) {
                char modelname[20];
                sprintf(modelname, "it-crack%d", t);
                set_model(modelname);
            }
            else
                set_model("invisible");
        }
        bool actor_hit(Actor *a) {
            return false;
        }
    };

    class Debris : public Item {
        CLONEOBJ(Debris);
        bool actor_hit(Actor *a) { SendMessage(a, "fall"); return false; }
        void init_model() {set_anim("it-debris");}
        void animcb() {
            GridPos p= get_pos();
            SetFloor(p, MakeFloor("fl-abyss"));
            KillItem(p);
        }
    public:
        Debris():Item("it-debris") {}
    };
}


//----------------------------------------
// Flags
//
// Flags can be used to set a new respawn point for the black or white
// marble.
//----------------------------------------
namespace
{
    class FlagWhite : public Item {
        CLONEOBJ(FlagWhite);
        void on_drop(Actor *a) { }

    public:
        FlagWhite(): Item("it-flagwhite") {}
    };

    class FlagBlack : public Item {
        CLONEOBJ(FlagBlack);
        void on_drop(Actor *a) { }

    public:
        FlagBlack(): Item("it-flagblack") {}
    };
}


void items::Init()
{
    using world::Register;

    Register(new BlackBomb);
    Register(new Brush);
    Register(new Coin);
    Register(new Crack);
    Register(new Debris);
    Register(new Document);
    Register(new Dynamite);
    Register(new ExtraLife);
    Register(new FlagBlack);
    Register(new FlagWhite);
    Register(new Floppy);
    Register(new Hammer);
    Register(new Hill);
    Register(new Hollow);
    Register(new Key);
    Register(new Key_a);
    Register(new Key_b);
    Register(new Key_c);
    Register(new MagicWand);
    Register(new Magnet);
    Register(new Pipe);
    Register(new Seed);
    Register(new ShogunDot);
    Register(new Spade);
    Register(new Spring1);
    Register(new Spring2);
    Register(new Sword);
    Register(new TinyHill);
    Register(new TinyHollow);
    Register(new Trigger);
    Register(new Umbrella);
    Register(new YinYang);
    Register(new WormHole);
}
