/*
 * Copyright (C) 2002,2003 Daniel Heck
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *  
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *  
 * You should have received a copy of the GNU General Public License along
 * with this program; if not, write to the Free Software Foundation, Inc.,
 * 59 Temple Place, Suite 330, Boston, MA 02111-1307, USA.
 *
 * $Id: object_mixins.hh,v 1.2 2003/01/06 12:25:03 dheck Exp $
 */
#ifndef OBJECT_MIXINS_HH
#define OBJECT_MIXINS_HH

#include "objects.hh"

namespace world
{
#define CLONEOBJ(type)                                                  \
        Object* clone() { Object *o= new type; o->init(); return o;}    \
        void dispose() { delete this; }

#define SINGLETONOBJ()                          \
        Object* clone() { return this; }        \
        void dispose() {}

    
#define INSTANCELISTOBJ(type)                                                           \
        typedef std::vector<type*> InstanceList;                                        \
        static InstanceList instances;                                                  \
        Object *clone() { type *o = new type; instances.push_back(o); return o;}        \
        void dispose() {                                                                \
            instances.erase(find(instances.begin(), instances.end(), this));            \
            delete this;                                                                \
        }
    
//----------------------------------------
// MovableStone
//
// This mixin class tries to move the stone to a neighboring field
// when hit hard enough.
//----------------------------------------
    class MovableStone : public Stone {
    protected:
        MovableStone(const char *name) : Stone(name) {}
        
        bool is_movable() { return true; }
        void actor_inside (Actor *a) {
            SendMessage(a, "shatter");
        }

        void actor_hit(const StoneContact &sc)
        {
            ActorInfo *ai = sc.actor->get_actorinfo();
            
            // If the speed component towards the face of the stone is large
            // enough (and actually pointing towards the stone), consider
            // moving the stone.
            Direction dir=contact_face(sc);
            if (dir!=enigma::NODIR && ai->vel * sc.normal < -4)
                MaybeMoveStone(get_pos(), reverse(dir));
        }
    };
    
//----------------------------------------
// OnOffBase
//
// A object that can be on and off and understands the messages "on",
// "off", and "onoff".  Whenever the "on" attribute changes,
// the object's init_model() method is invoked
//----------------------------------------

    template <class T>
    class OnOffBase : public T {
    protected:
        OnOffBase(const char *kind) : T(kind) { set_attrib("on", 0.0); }
        bool is_on() { return int_attrib("on")==1; }

        void set_on(bool newon) {
            if (newon != is_on()) {
                set_attrib("on", Value(newon));
                init_model();
                notify_onoff(newon);
            }
        }

        virtual void notify_onoff(bool on) {}

        void message(const string &m, const Value &) {
            if (m == "on")
                set_on(true);
            else if (m=="off")
                set_on(false);
            else if (m=="onoff")
                set_on(!is_on());
        }
    };

    typedef OnOffBase<Stone> OnOffStone;
    typedef OnOffBase<Item> OnOffItem;
}

#endif
