/*
 * Copyright (C) 2002,2003 Daniel Heck
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *  
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *  
 * You should have received a copy of the GNU General Public License along
 * with this program; if not, write to the Free Software Foundation, Inc.,
 * 59 Temple Place, Suite 330, Boston, MA 02111-1307, USA.
 *
 * $Id: objects.cc,v 1.19 2003/02/14 19:04:51 dheck Exp $
 */
#include "game.hh"
#include "objects.hh"
#include "object_mixins.hh"
#include "display.hh"
#include "player.hh"
#include "lua.hh"
#include "sound.hh"
#include "laser.hh"
#include "actors.hh"

#include "px/tools.hh"
#include "px/dict.hh"

#include <algorithm>
#include <string>
#include <cstdlib>
#include <iostream>
#include <cassert>

using namespace std;
using namespace world;
using namespace enigma;

namespace
{
    player::Inventory *
    get_inventory(const Actor *a)
    {
        if (const Value *v = a->get_attrib("player"))
            return player::GetInventory(to_int(*v));
        return 0;
    }

    bool wielded_item_is(Actor *a, const string &kind)
    {
        if (player::Inventory *inv = get_inventory(a)) 
            if (Item *it = inv->get_item(0))
                return it->get_kind() == kind;
        return false;
    }

}

/* This function is used by all triggers, switches etc. that perform
   some particular action when activated (like opening doors or
   switching lasers on and off). */
void 
world::PerformAction(Object *o, bool onoff)
{
    string action = "idle";
    string target;

    o->string_attrib("action", &action);
    o->string_attrib("target", &target);

    if (action == "callback") {
        lua::CallFunc(target.c_str(), Value(onoff));
    }
    else if (Object *o = GetNamedObject(target)) {
        SendMessage(o, action);
    }
}

//----------------------------------------
// Object impl.
//----------------------------------------

Object::Object(const char *kind)
{ 
    set_attrib("kind", Value(kind));
}

void 
Object::message(const string& msg, const Value &val)
{
}

const char *
Object::get_kind() const
{
    const Value *v = get_attrib("kind");
    assert(v && v->get_type()==Value::STRING);
    return v->get_string();
}


void 
Object::set_attrib(const string& key, const Value& val)
{
    attribs[key] = val;//.insert (key, val);
}

const Value*
Object::get_attrib(const string& key) const
{
    AttribMap::const_iterator i = attribs.find(key);
    if (i == attribs.end())
        return 0;
    else
        return &i->second;
}

bool
Object::string_attrib(const string &name, string *val) const
{
    if (const Value *v = get_attrib(name)) 
    {
        const char *s = to_string(*v);
        if (s != 0) {
            *val = s;
            return true;
        }
    }
    return false;
}

int
Object::int_attrib(const string &name) const
{
    if (const Value *v = get_attrib(name))
        return to_int(*v);
    return 0;
}

bool 
Object::int_attrib(const string &name, int *val) const
{
    if (const Value *v = get_attrib(name)) {
        *val = to_int(*v);
        return true;
    }
    return false;
}


bool
Object::double_attrib(const string &name, double *val) const
{
    if (const Value *v = get_attrib(name))
    {
        *val = to_double(*v);
        return true;
    }
    return false;
}

//----------------------------------------
// GridObject
//----------------------------------------
void 
GridObject::play_sound(const char *name)
{
    sound::PlaySound(name, px::V2(get_pos().x+0.5, get_pos().y+0.5));
}



//======================================================================
// FLOORS
//======================================================================

Floor::Floor(const char *kind, double friction, double mfactor)
    : TGridObject<GRID_FLOOR>(kind) , _friction(friction)
{
    _mousefactor = mfactor;
}

Object *Floor::clone() {
    return this;
}

void Floor::dispose() {
    ; // do nothing
}

void 
Floor::message(const string& msg, const Value &val)
{
    if (msg == "explode")
        SetFloor(get_pos(), MakeFloor("fl-abyss"));
}

//----------------------------------------
// Abyss
//----------------------------------------
namespace
{
    class Abyss : public Floor {
        CLONEOBJ(Abyss);
    public:
        Abyss() : Floor("fl-abyss", 10, 1) {}
    private:
        void actor_enter(Actor* a) {SendMessage(a, "fall");}
        void on_actorhit(Actor* a) {SendMessage(a, "fall");}
    };
}

//----------------------------------------
// Water
//----------------------------------------
namespace
{
    class Water : public Floor {
        CLONEOBJ(Water);
    public:
        Water() : Floor("fl-water", 5, 1) {}
    private:
        void actor_enter(Actor *a) {SendMessage(a, "drown");}
        void on_actorhit(Actor *a) {SendMessage(a, "drown");}
    };
}

//----------------------------------------
// Gradient
//----------------------------------------

/** \page fl-gradient Gradient Floor

This is a sloped floor that accelerates in a particular direction. 

\subsection gradienta Attributes

- \b type: number between 0 and 15 (see below)

\subsection gradiente Examples

*/
namespace
{
    class Gradient : public Floor {
        CLONEOBJ(Gradient);
    public:
        Gradient();
    private:
        int get_type() const;
        enum { MINTYPE=1, MAXTYPE=24 };
        
        // GridObject interface.
        virtual void init_model();

        // Floor interface
        virtual px::V3 get_force(Actor *a);
    };
}

Gradient::Gradient()
    : Floor("fl-gradient", 4, 2)
{
    set_attrib("type", Value(MINTYPE));
}

int Gradient::get_type() const 
{ 
    int t=int_attrib("type");
    if (t < MINTYPE || t>MAXTYPE) {
        enigma::Log << "Gradient: Illegal type="<< int(t) << endl;
        t=MINTYPE;
    }
    return t;
}

void Gradient::init_model()
{
    char mname[20];
    sprintf(mname, "fl-gradient%d", get_type());
    set_model(mname);
}

px::V3 
Gradient::get_force(Actor *a) 
{
    V3 force;
    int t = get_type();

    static int xforce[MAXTYPE-MINTYPE+1] = {
         0,  0,  1, -1,
         1, -1,  1, -1,
         1,  1, -1, -1,
        -1, -1,  1,  1,
         1, -1,  1, -1,
         0,  0,  1, -1
    };
    static int yforce[MAXTYPE-MINTYPE+1] = {
         1, -1,  0,  0,
         1,  1, -1, -1, 
         1, -1,  1, -1,
         1, -1,  1, -1,
        -1, -1,  1,  1,
         1, -1,  0,  0
    };
    force = V3(xforce[t-MINTYPE], yforce[t-MINTYPE], 0);
    force.normalize();
    return 25*force;
}


//----------------------------------------
// Bridge
//----------------------------------------

/** \page fl-bridge Bridge Floor

This Floor can be opened and closed much like a bridge.
The actor can move over this floor if the bridge is closed,
and will fall into abyss when the bridge is open.

\subsection bridgea Attributes

- \b type	The type of the bridge, currently only 'a' is possible

\subsection bridgem Messages
- \b open	open the bridge so actors cannot pass it
- \b close	close the bridge so actors can pass it
- \b openclose	toggle the state of the bridge

*/
namespace
{
    class Bridge : public Floor {
        CLONEOBJ(Bridge);
    public:
        Bridge();
    private:
        enum State { OPEN, CLOSED, OPENING, CLOSING } state;
	
        char get_type() const {
            string type = "a";
            string_attrib("type", &type);
            return type[0];
        }

        void actor_enter(Actor *);
        void on_actorhit(Actor* a) {if (state!=CLOSED) SendMessage(a, "fall");}
        void message(const string &m, const Value &);
        void init_model();

        void change_state( State newstate);
        void animcb();
    };
}

Bridge::Bridge() : Floor("fl-bridge", 5, 1) 
{
    set_attrib("type", "a");
    state=OPEN;
}

void Bridge::message(const string &m, const Value &) 
{
    if (m == "open" && (state==CLOSED || state==CLOSING))
        change_state(OPENING);
    else if (m=="close" && (state==OPEN || state==OPENING))
        change_state(CLOSING);
    else if (m=="openclose") {
        if (state==OPEN || state==OPENING)
            change_state(CLOSING);
        else
            change_state(OPENING);
    }
}

void Bridge::actor_enter(Actor *a)
{
    if (state != CLOSED)
        SendMessage(a, "fall");
}


void Bridge::init_model() 
{
    char mname[25];
    sprintf(mname, "fl-bridge%c-%s", get_type(),
            (state==OPEN) ? "open" : "closed");
    set_model(mname);
}

void Bridge::change_state( State newstate) 
{
    string mname = string("fl-bridge")+get_type();
    switch( newstate) {
    case OPENING:
        if( state == CLOSING)
            get_model()->reverse();
        else
            set_anim(mname+"-opening");
        break;
    case CLOSING:
        if( state == OPENING)
            get_model()->reverse();
        else
            set_anim(mname+"-closing");
        break;
    case OPEN:
    case CLOSED:
        state=newstate;
        init_model();
        break;
    }
    state=newstate;
}

void Bridge::animcb() 
{
    if (state==OPENING)
        change_state(OPEN);
    if (state==CLOSING)
        change_state(CLOSED);
}

#if 0

class WhiteTile : public Floor {
public:
    V3 get_force(Actor *a, V3 mouseforce) {
        if (a->get_attrib("blackball"))
            return Floor::get_force(a, 0);
        else
            return Floor::get_force(a, mouseforce);
    }
};

class BlackTile : public Floor {
public:
    V3 get_force(Actor *a, V3 mouseforce) {
        if (a->get_attrib("whiteball"))
            return Floor::get_force(a, 0);
        else
            return Floor::get_force(a, mouseforce);
    }
};

#endif


//======================================================================
// STONES
//======================================================================

/** \page lstones  Available Stones

Oxyd Stones:
- \ref st-oxyd
- \ref st-fakeoxyd
- \ref st-fart

Movable stones:
- \ref st-brownie
- \ref st-wood
- \ref st-block

Stones that can trigger actions:
- \ref st-switch
- \ref st-laserswitch
- \ref st-key
- \ref st-coinslot
- \ref st-timer

Stones that can change their behaviour:
- \ref st-brick_magic
- \ref st-stonebrush
- \ref st-invisible_magic
- \ref st-break_invisible

Lasers and Mirrors:
- \ref st-laser
- \ref st-pmirror
- \ref st-3mirror

Other stones:
- \ref st-death
- \ref st-swap
- \ref st-bolder
- \ref st-puzzle
- \ref st-stone_break
- \ref st-oneway
- \ref st-oneway_black
- \ref st-oneway_white
- \ref st-chameleon
*/


Stone::Stone(const char * kind) 
    : TGridObject<GRID_STONES>(kind) 
{}

px::V3 
Stone::actor_impulse(const StoneContact &sc)
{
    V3 normal = normalize(sc.normal);
    const V3 &v = sc.actor->get_vel();
    return -2*(v*normal)*normal;
}

StoneResponse
Stone::collision_response(const StoneContact &)
{
    return STONE_REBOUND;
}


void
Stone::actor_hit(const StoneContact &)
{
}

const char *
Stone::collision_sound()
{
    return "st-stone";
}

//----------------------------------------
// SimpleStone
//
// This kind of stone can be defined from Lua programs with
// `DefineSimpleStone'.
//----------------------------------------
namespace
{
    class SimpleStone : public Stone {
    public:
        SimpleStone(const string &knd, const string & snd)
            : Stone(knd.c_str()), sound(snd)
        {}
    private:
        const char *collision_sound() {
            return sound.c_str();
        }

        Object *clone() {
            return new SimpleStone(get_kind(), sound);
        }
        void dispose() { delete this; }

        string sound;
    };
}

//----------------------------------------
// One Way Stone; black, white
//
// This stone can only be passed in one direction. 
//----------------------------------------

/** \page st-oneway Oneway Stone

This stone can be passed by the marble in only one direction
  
\subsection onewaya Attributes

- \c orientation  \n NORTH, EAST, SOUTH, WEST

\subsection onewaye Example
\verbatim
set_stone("st-oneway", 10,10, {orientation=EAST})
\endverbatim
or in short form
\verbatim
oneway( 10, 10, EAST)
\endverbatim

\image html st-oneway-e.png
*/
namespace
{
    class OneWayStone : public Stone {
        CLONEOBJ(OneWayStone);
    public:
	OneWayStone() : Stone("st-oneway") {
	    set_orientation(SOUTH);
	}
    private:
	void init_model() {
            string mname = "st-oneway";
            mname += to_suffix(get_orientation());
            set_model (mname);
	}
	Direction get_orientation() const {
	    return Direction(int_attrib("orientation"));
	}
        void set_orientation(Direction dir) {
            set_attrib("orientation", Value(dir));
        }
	StoneResponse collision_response(const StoneContact&);
        void actor_hit (const StoneContact&);
    };
}

StoneResponse
OneWayStone::collision_response(const StoneContact &sc)
{
    DirectionBits dirs=contact_faces(sc);
    Direction o=get_orientation();
    
    return has_dir(dirs,o) ? STONE_REBOUND : STONE_PASS;
}

void
OneWayStone::actor_hit(const StoneContact &sc)
{
    DirectionBits dirs=contact_faces(sc);
    Direction o=get_orientation();
    
    if (has_dir(dirs,o)) {
        // magic wand as first item?
        if (wielded_item_is(sc.actor, "it-magicwand")) {
            set_orientation(reverse(o));
            init_model();
        }
    } 
}

//----------------------------------------
// One Way Stone; black only
//
// This stone can only be passed in one direction.
//----------------------------------------

namespace
{
    class OneWayStone_black : public Stone {
        CLONEOBJ(OneWayStone_black);
    public:
	OneWayStone_black() : Stone("st-oneway_black") {
	    set_orientation(SOUTH);
	}
    private:
	void init_model() {
            string mname = "st-oneway_black";
            mname += to_suffix(get_orientation());
            set_model (mname);
	}
	Direction get_orientation() const {
	    return Direction(int_attrib("orientation"));
	}
        void set_orientation(Direction dir) {
            set_attrib("orientation", Value(dir));
        }
	StoneResponse collision_response(const StoneContact&);
        void actor_hit (const StoneContact&);
    };
}

StoneResponse
OneWayStone_black::collision_response(const StoneContact &sc)
{
    DirectionBits dirs=contact_faces(sc);
    Direction o=get_orientation();
    if (sc.actor->get_attrib("blackball"))
    return has_dir(dirs,o) ? STONE_REBOUND : STONE_PASS;
    else
         return STONE_REBOUND;
}

void
OneWayStone_black::actor_hit(const StoneContact &sc)
{
    DirectionBits dirs=contact_faces(sc);
    Direction o=get_orientation();

    if (has_dir(dirs,o)) {
        // magic wand as first item?
        if (wielded_item_is(sc.actor, "it-magicwand")) {
            set_orientation(reverse(o));
            init_model();
        }
    }
}

//----------------------------------------
// One Way Stone ; white only
//
// This stone can only be passed in one direction.
//----------------------------------------

namespace
{
    class OneWayStone_white : public Stone {
        CLONEOBJ(OneWayStone_white);
    public:
	OneWayStone_white() : Stone("st-oneway_white") {
	    set_orientation(SOUTH);
	}
    private:
	void init_model() {
            string mname = "st-oneway_white";
            mname += to_suffix(get_orientation());
            set_model (mname);
	}
	Direction get_orientation() const {
	    return Direction(int_attrib("orientation"));
	}
        void set_orientation(Direction dir) {
            set_attrib("orientation", Value(dir));
        }
	StoneResponse collision_response(const StoneContact&);
        void actor_hit (const StoneContact&);
    };
}

StoneResponse
OneWayStone_white::collision_response(const StoneContact &sc)
{
    DirectionBits dirs=contact_faces(sc);
    Direction o=get_orientation();
    if (sc.actor->get_attrib("whiteball"))
    return has_dir(dirs,o) ? STONE_REBOUND : STONE_PASS;
    else
         return STONE_REBOUND;
}

void
OneWayStone_white::actor_hit(const StoneContact &sc)
{
    DirectionBits dirs=contact_faces(sc);
    Direction o=get_orientation();

    if (has_dir(dirs,o)) {
        // magic wand as first item?
        if (wielded_item_is(sc.actor, "it-magicwand")) {
            set_orientation(reverse(o));
            init_model();
        }
    }
}

//----------------------------------------
// Chameleon Stone
//
// This stone takes on the look of the floor beneath it.  Actors can
// move through it, so these stones are perfect for hiding stuff under
// them...
//----------------------------------------

/** \page st-chameleon Chameleon Stone

This stone takes on the look of the floor beneath it.  Actors can
move through it, so these stones are perfect for hiding stuff under
them...

*/
namespace
{
    class ChameleonStone : public Stone {
        CLONEOBJ(ChameleonStone);
    public:
        ChameleonStone() : Stone("st-chameleon"){}
    private:
        void init_model() {
            string modelname = "fl-gray";
            if (Floor *fl = GetFloor(get_pos()))
                modelname = fl->get_kind();
            set_model(modelname);
        }
        StoneResponse collision_response(const StoneContact &sc) {
            return STONE_PASS;
        }
    };
}


//----------------------------------------
// Key Switch, Key_a Switch, Key_b Switch
//
// Attributes:
//
// :keycode         a numerical code; only keys with the same code 
//                  can activate this switch
// :on              1 or 0
// :target,action   as usual
//----------------------------------------

/** \page st-key Key Switch Stone

This stone acts as a lock and can only be activated by
using a key item. You can use keycodes to let keys only
open specific key stones.
  
\subsection keye Example
\verbatim
set_stone( "st-key_b", 14,77, {action="openclose", target="door2"})
set_item("it-key_b", 12,7)
\endverbatim

\image html st-key0.png
*/
namespace
{
    class KeyStone : public OnOffStone
    {
    protected:
        KeyStone(const char *kind, int keycode=0) : OnOffStone(kind) {
            set_attrib("keycode", keycode);
        }
    private:
        void init_model() {set_model(is_on() ? "st-key1" : "st-key0");}
        void actor_hit(const StoneContact &sc);
        const char *collision_sound() { return "st-metal"; }
    };

    class KeyStone_a : public KeyStone {
        CLONEOBJ(KeyStone_a);
    public:
        KeyStone_a() : KeyStone("st-key_a", 1) {}
    };
    class KeyStone_b : public KeyStone {
        CLONEOBJ(KeyStone_b);
    public:
        KeyStone_b() : KeyStone("st-key_b", 2) {}
    };
    class KeyStone_c : public KeyStone {
        CLONEOBJ(KeyStone_c);
    public:
        KeyStone_c() : KeyStone("st-key_c", 3) {}
    };
}

void
KeyStone::actor_hit(const StoneContact &sc) 
{
    if (player::Inventory *inv = get_inventory(sc.actor))
    {
        if (is_on()) {
            Item *key = MakeItem("it-key");
            key->set_attrib("keycode", int_attrib("keycode"));
            inv->add_item(key);
            set_on(false);
            PerformAction(this, is_on());
        }
        else
        {
            Item *it = inv->get_item(0);
            if (it && it->int_attrib("keycode") == int_attrib("keycode"))
            {
                set_on(true);
                PerformAction(this, is_on());
                delete inv->yield_first();
            }
        }
    }
}

//----------------------------------------
// SwapStone
//----------------------------------------

/** \page st-swap Swap Stone

This stone can exchange its position with other neighboring stones
if it is hit hard enough.  In a way, this makes swap stones a kind
of "movable stone", except that they can be only exchanged with
other stones and may not be moved on empty fields.

\subsection swape Example
\verbatim
set_stone("st-swap", 10,10)
\endverbatim

\image html st-swap.png
*/
namespace
{
    class SwapStone : public Stone {
        CLONEOBJ(SwapStone);
    public:
        SwapStone() : Stone("st-swap") {}
    private:
        void actor_hit(const StoneContact &sc)
        {
            Direction hit_dir = reverse(contact_face(sc));
            Actor *a = sc.actor;
            if (a->get_vel() * sc.normal < -4 && hit_dir!=NODIR) {
                GridPos p = get_pos();
                GridPos newp = move(p, hit_dir);
                if (GetStone(newp))
                    SwapStones(p,newp);
            }
        }
    };
}

//----------------------------------------
// Three simple movable stones
//----------------------------------------
namespace
{
    class Greenbrown_move : public MovableStone {
        CLONEOBJ(Greenbrown_move);
        const char *collision_sound() {return "st-stone";}
    public:
        Greenbrown_move() : MovableStone("st-greenbrown_move") {}
    };

    class Marble_move : public MovableStone {
        CLONEOBJ(Marble_move);
        const char *collision_sound() {return "st-thud";}
    public:
        Marble_move() : MovableStone("st-marble_move") {}
    };

    class Rock3_move : public MovableStone {
        CLONEOBJ(Rock3_move);
        const char *collision_sound() {return "st-thud";}
    public:
        Rock3_move() : MovableStone("st-rock3_move") {}
    };
}

//----------------------------------------
// BlockStone
//----------------------------------------

/** \page st-block Block Stone

This stone can is movable.
If moved into water it will disappear and
build new water floor at this place.

\subsection block Example
\verbatim
set_stone("st-block", 10,10)
\endverbatim

\image html st-block.png
*/
namespace
{
    class BlockStone : public MovableStone {
        CLONEOBJ(BlockStone);
    public:
        BlockStone() : MovableStone("st-block") {}
    private:
        V3 get_center() const {
            GridPos p = get_pos();
            return V3(p.x+0.5, p.y+0.5,0);
        }

        void on_move(){
            if (Floor *fl=GetFloor(get_pos())) {
                const string &k = fl->get_kind();
                if (k=="fl-water") {
                    AddSprite(display::SPRITE_EFFECT, get_center(), 
                              "ring-anim");

//                    SetFloor(get_pos(), MakeFloor("fl-water"));
                    KillStone(get_pos());
                }
            }
        }
    };

};

//----------------------------------------
// Stone_break
//----------------------------------------

/** \page st-stone_break Breakable Stone

This stone can be destroyed by an actor having a
hammer.

\subsection stone_breake Example
\verbatim
set_stone("st-stone_break", 10,10)
\endverbatim

\image html st-stone_break.png
*/
namespace
{
    class Stone_break : public Stone {
        CLONEOBJ(Stone_break);
        const char *collision_sound() {return "st-stone";}
    public:
        Stone_break() : Stone("st-stone_break"), state(IDLE) {}
    private:
        enum State { IDLE, BREAK };
        State state;
        void actor_hit(const StoneContact &sc)
        {
          if( state == IDLE)
          {
             if (wielded_item_is(sc.actor, "it-hammer")) {
                play_sound("explosion1");
                state = BREAK;
                set_anim("st-break-anim");
           }
        }
    }
        void animcb() {
            if (state == BREAK) {
            KillStone(get_pos());
         }
      }
   };
}

//----------------------------------------
// Brownie
//----------------------------------------

/** \page st-brownie Brownie Stone

This stone is movable.

\subsection browniee Example
\verbatim
set_stone("st-brownie", 10,10)
\endverbatim

\image html st-brownie.png
*/
namespace
{
    class Brownie : public MovableStone {
        CLONEOBJ(Brownie);
        const char *collision_sound() {return "st-thud";}
    public:
        Brownie() : MovableStone("st-brownie") {}
    };
}

//----------------------------------------
// Brick Magic
//----------------------------------------

/** \page st-brick_magic Magic Brick Stone

This stone does initially looks like a "st-brick".  If touched by the
actor, having a magic wand, it turns into a "st-glass" stone and
allows lasers to go through it.

\subsection brick_magicke Example
\verbatim
set_stone("st-brick_magic", 10,10)
\endverbatim

\image html st-brick.png
*/
namespace
{
    class BrickMagic : public Stone {
        CLONEOBJ(BrickMagic);
        const char *collision_sound() {return "st-stone";}
    public:
	BrickMagic() : Stone("st-brick_magic"), state(BRICK) {}
    private:
        enum State { BRICK, GLASS } state;
	void actor_hit(const StoneContact &sc)
        {
	    if( state == BRICK)
            {
	        if (wielded_item_is(sc.actor, "it-magicwand")) {
		    play_sound("st-magic");
		    state = GLASS;
		    set_model("st-glass");
		    laser::MaybeRecalcLight(get_pos());
                }
            }
	}
	bool on_laserhit(Direction dir) {
	    return state != BRICK;
	}
    };
}

//----------------------------------------
// Stonebrush
//----------------------------------------

/** \page st-stonebrush Brush Stone

This stone is initially invisible. If touched by an actor
having a brush it will be painted an become a "st-rock4".

\subsection stonebrushe Example
\verbatim
set_stone("st-stonebrush", 10,10)
\endverbatim

\image html st-rock4.png
*/
namespace
{
    class Stonebrush : public Stone {
        CLONEOBJ(Stonebrush);
        const char *collision_sound() {return "st-stone";}
    public:
        Stonebrush() : Stone("st-stonebrush"), state(INVISIBLE) {}
    private:
        enum State { INVISIBLE, BRUSH } state;
        void actor_hit(const StoneContact &sc)
        {
            if( state == INVISIBLE)
            {
                if (wielded_item_is(sc.actor, "it-brush")) {
		    play_sound("st-magic");
                    state = BRUSH;
                    set_model("st-rock4");
                }
            }
        }
    };
}

//----------------------------------------
// Break_invisible
//----------------------------------------

/** \page st-break_invisible Brush Stone

This stone is initially invisible. If touched by an actor
having a brush it will be painted an become a "st_stone_break".
This stone can be destroyed by an actor having a hammer.

\subsection break_invisible Example
\verbatim
set_stone("st-break_invisible", 10,10)
\endverbatim

\image html st-stone_break.png
*/
namespace
{
    class Break_invisible : public Stone {
        CLONEOBJ(Break_invisible);
        const char *collision_sound() {return "st-stone";}
    public:
        Break_invisible() : Stone("st-break_invisible"), state(INVISIBLE) {}
    private:
        enum State { INVISIBLE, BRUSH };
        State state;
        void actor_hit(const StoneContact &sc)
        {
            if( state == INVISIBLE)
            {
                if (wielded_item_is(sc.actor, "it-brush")) {
                    play_sound("st-magic");
                    state = BRUSH;
                    set_model("st-stone_break");
                }
            }
        {
            if( state == BRUSH)
            {
                if (wielded_item_is(sc.actor, "it-hammer")) {
                    play_sound("explosion1");
                    set_anim("st-break_invisible-anim");
                 }
             }
        }
    }
        void animcb() {
            if (state == BRUSH) {
                KillStone(get_pos());
            }

        }
    };
}

//----------------------------------------
// Invisible Magic
//----------------------------------------

/** \page st-invisible_magic Magic Invisible Stone

This stone is initially invisible, and laserlight can pass through.
If touched by an actor having a magic wand, it will mutate into a
"st-greenbrown" and laserlight is blocked.

\subsection invisible_magice Example
\verbatim
set_stone("st-invisible_magic", 10,10)
\endverbatim

\image html st-greenbrown.png
*/
namespace
{
    class InvisibleMagic : public Stone {
        CLONEOBJ(InvisibleMagic);
        const char *collision_sound() {return "st-thud";}
    public:
	InvisibleMagic() : Stone("st-invisible_magic"), state(INVISIBLE) {}
    private:
        enum State { INVISIBLE, STONE } state;
	void actor_hit(const StoneContact &sc)
        {
	    if( state == INVISIBLE)
            {
	        if (wielded_item_is(sc.actor, "it-magicwand")) {
	            play_sound("st-magic");
		    state = STONE;
		    set_model("st-greenbrown");
		    laser::MaybeRecalcLight(get_pos());
                }
            }
	}
	bool on_laserhit(Direction dir) {return state==INVISIBLE;}
    };
}

//----------------------------------------
// WoodenStone
//----------------------------------------

/** \page st-wood Wooden Stone

This stone can is movable.  If moved into abyss or water it will
disappear and change into a wooden plank.

\subsection woode Example
\verbatim
set_stone("st-wood", 10,10)
\endverbatim

\image html st-wood.png
*/
namespace
{
    class WoodenStone : public MovableStone {
        CLONEOBJ(WoodenStone);
    public:
        WoodenStone() : MovableStone("st-wood") {}
    private:
        void on_move() {
            if (Floor *fl=GetFloor(get_pos())) {
                const string &k = fl->get_kind();
                if (k == "fl-abyss" || k=="fl-water" || k=="fl-swamp") {
                    SetFloor(get_pos(), MakeFloor("fl-wood"));
                    KillStone(get_pos());
                }
            }
        }
    };

    class WoodenStone_Growing : public Stone {
        CLONEOBJ(WoodenStone_Growing);
    public:
        WoodenStone_Growing() : Stone("st-wood-growing") {}
    private:
        void init_model() { set_anim("st-wood-growing"); }
        void animcb() {
            Stone *st = world::MakeStone("st-wood");
            world::SetStone(get_pos(), st);
        }
    };
}

//----------------------------------------
// ElectricStone
//
// Attributes:
//
// :charge  + - 0
//----------------------------------------
namespace
{
    class ElectricStone : public Stone {
        CLONEOBJ(ElectricStone);
    public:
        ElectricStone() : Stone("st-electric") {
            set_attrib("charge", +1);
        }
    private:
        void init_model() {
        }
    };
}

//----------------------------------------
// ScissorsStone
//----------------------------------------

/** \page st-scissors Scissors stone
   
This stone cuts \c all rubber bands attached to an actor that touches
it.

\image html st-scissors
*/
namespace
{
    class ScissorsStone : public Stone {
        CLONEOBJ(ScissorsStone);
    public:
        ScissorsStone() : Stone("st-scissors") {}
    private:
        void actor_hit(const StoneContact &sc) {
            world::KillRubberBand (sc.actor, (Stone*)0);
            world::KillRubberBand (sc.actor, (Actor*)0);
            play_sound("snip");
            set_anim("st-scissors-snip");
        }
        void animcb() {
            set_model("st-scissors");
        }
    };
}

//----------------------------------------
// RubberBand stone
//----------------------------------------

/** \page st-rubberband Rubberband stone

If hit by a marble, this stone first removes existing connections with
other rubberband stones and then attaches a new elastic between the
marble and itself.  Nothing happens if the marble was already attached
to this particular stone.

This stone can be moved if hit with a magic wand.

\subsection rubberbanda Attributes

- \c length  The natural length of the rubberband (default: 1)
- \c strength The strength of the rubberband (default: 10)

\image html st-rubberband.png
*/
namespace
{
    class RubberBandStone : public MovableStone {
        CLONEOBJ(RubberBandStone);
    public:
        RubberBandStone () : MovableStone ("st-rubberband") {
            set_attrib("length", 1.0);
            set_attrib("strength", 10.0);
        }
    private:
        

        void actor_hit(const StoneContact &sc) {
            double strength = 10.0;
            double length = 1.0;
            double_attrib ("length", &length);
            double_attrib ("strength", &strength);

            if (!world::HasRubberBand (sc.actor, this)) {
                play_sound ("boing");
                world::KillRubberBand (sc.actor, (Stone*)0);
                world::AddRubberBand (sc.actor, this, strength, length);
            }
            if (wielded_item_is (sc.actor, "it-magicwand"))
                MovableStone::actor_hit(sc);
        }
    };
}


//----------------------------------------
// BolderStone
//----------------------------------------

/** \page st-bolder Bolder Stone

The bolder stone will move in one direction until another stone will
block.  When hit with a magic wand, the bolder stone reverse its
direction. When hitting a blocking stone it can activate switches or
oxyd stones.

\subsection boldera Attributes

- \c direction  \n NORTH, EAST, SOUTH, WEST

*/
namespace
{
    class BolderStone : public Stone, public TimeHandler 
    {
        CLONEOBJ(BolderStone);
    public:
        BolderStone() : Stone("st-bolder"), state(IDLE) {
            set_attrib("direction", enigma::NORTH);
        }
    private:
        enum State { IDLE, MOVING } state;

        Direction get_dir() const 
        { 
            return static_cast<Direction>(int_attrib("direction")); 
        }
        void set_dir(Direction d) {set_attrib("direction", d);}

        bool is_movable() { return true; }

        void on_creation() {
            if (state == IDLE)
                g_timer.set_alarm(this, 0.3, true);
            Stone::on_creation();
        }
        void on_removal() {
            if (state == IDLE)
                g_timer.remove_alarm(this);
            Stone::on_removal();
        }

        void alarm() {
            state = MOVING;
	    GridPos p = get_pos();
            if (world::MaybeMoveStone(p, get_dir())) {
                // send a message to the stone that's blocking our way
		Stone *st = GetStone(move(move(p,get_dir()),get_dir()));
		if( st)
	            SendMessage(st, "trigger");
            }
            state = IDLE;
        }

        // Stone interface.
        void init_model() 
        {
            string mname = "st-bolder" + to_suffix(get_dir());
            set_model(mname);
        }

        void actor_hit(const StoneContact &sc)
        {
            Actor *a = sc.actor;

            if (wielded_item_is(a, "it-magicwand")) {
                set_dir(reverse(get_dir()));
                init_model();
            }
        }
        void actor_inside(Actor *a) {SendMessage(a, "shatter");}
    };
}


//----------------------------------------
// PuzzleStones
//----------------------------------------

/** \page st-puzzle Puzzle Stone

These stones can be connected to other stones of the same type.  Any
of the four faces of the stone can have ``socket''.  If the touching
faces of two adjacent stones both have a socket, the two stones link
up and henceforth move as group.

A cluster of puzzle stones may for example look like this:

\verbatim
+---+---+---+---+
|   |   |   |   |
| --+-+-+---+-+ |
|   | | |   | | |
+---+-+-+---+-+-+
    | | |   | | |
    | | |   | | |
    |   |   |   |
    +---+   +---+
\endverbatim

This example actually presents the special case of a ``complete''
cluster.  A cluster is complete if none of its stones has an
unconnected socket.  If such a group is touched with a magic wand, all
its constituents explode.  [They should, anyway. This is not
implemented yet.]

\subsection puzzlea Attributes

- \b connections 
   number between 1 an 16.  Each bit in (connections-1) corresponds to
   a socket on one of the four faces.  You will normally simply use
   one of the Lua constants \c PUZ_0000 to \c PUZ_1111.

\subsection puzzlee Example
<table>
<tr>
<td> \image html st-puzzletempl_0001.png "PUZ_0000"
<td> \image html st-puzzletempl_0002.png "PUZ_0001"
<td> \image html st-puzzletempl_0003.png "PUZ_0010"
<td> \image html st-puzzletempl_0004.png "PUZ_0011"
<tr>
<td> \image html st-puzzletempl_0005.png "PUZ_0100"
<td> \image html st-puzzletempl_0006.png "PUZ_0101"
<td> \image html st-puzzletempl_0007.png "PUZ_0110"
<td> \image html st-puzzletempl_0008.png "PUZ_0111"
<tr>
<td> \image html st-puzzletempl_0009.png "PUZ_1000"
<td> \image html st-puzzletempl_0010.png "PUZ_1001"
<td> \image html st-puzzletempl_0011.png "PUZ_1010"
<td> \image html st-puzzletempl_0012.png "PUZ_1011"
<tr>
<td> \image html st-puzzletempl_0013.png "PUZ_1100"
<td> \image html st-puzzletempl_0014.png "PUZ_1101"
<td> \image html st-puzzletempl_0015.png "PUZ_1110"
<td> \image html st-puzzletempl_0016.png "PUZ_1111"
</tr>
</table>
*/
namespace
{
    class PuzzleStone : public Stone {
        INSTANCELISTOBJ(PuzzleStone);
    public:
        PuzzleStone() : Stone("st-puzzle") {
            set_attrib("connections", 0.0);
        }
    private:
        typedef vector<GridPos> Cluster;

        DirectionBits get_connections();

        static bool visit_dir(vector<GridPos> &stack, 
                              GridPos curpos, Direction dir);
        bool find_cluster(Cluster &);
        void find_row_or_column_cluster(Cluster &c, 
                                        GridPos startpos, Direction dir);
        bool cluster_complete();
        void maybe_move_cluster(Cluster &c, bool is_complete, Direction dir);
        void rotate_cluster(const Cluster &c);

        bool on_laserhit(Direction dir) { 
            return get_connections() == 0; 
        }

        void init_model();
        StoneResponse collision_response(const StoneContact &sc);
        void actor_hit(const StoneContact &sc);
        void actor_inside(Actor *a) {
            if (get_connections() != NODIRBIT)
                SendMessage(a, "shatter");
        }

        // Variables
        bool visited;           // flag for DFS
    };
}

PuzzleStone::InstanceList PuzzleStone::instances;

DirectionBits 
PuzzleStone::get_connections() 
{
    int conn=int_attrib("connections") - 1;
    if (conn >=0 && conn <16)
        return DirectionBits(conn);
    else
        return NODIRBIT;
}

bool
PuzzleStone::visit_dir(vector<GridPos> &stack, GridPos curpos, Direction dir)
{
    GridPos newpos = move(curpos, dir);
    PuzzleStone *pz = dynamic_cast<PuzzleStone*>(GetStone(newpos));
    
    if (!pz)
        return false;

    DirectionBits cfaces = pz->get_connections();

    if (cfaces==NODIRBIT || has_dir(cfaces, reverse(dir))) {
        // Puzzle stone at newpos is connected to stone at curpos
        if (!pz->visited) {
            pz->visited = true;
            stack.push_back(newpos);
        }
        return true;
    } else {
        // The two stones are adjacent but not connected
        return false;
    }
}

/* Use a depth first search to determine the group of all stones that
   are connected to the current stone.  Returns true if the cluster is
   ``complete'' in the sense defined above. */
bool
PuzzleStone::find_cluster(Cluster &cluster)
{
    for (unsigned i=0; i<instances.size(); ++i)
        instances[i]->visited=false;

    vector<GridPos> pos_stack;
    bool is_complete = true;
    pos_stack.push_back(get_pos());
    this->visited = true;
    while (!pos_stack.empty()) 
    {
        GridPos curpos = pos_stack.back();
        pos_stack.pop_back();

        PuzzleStone *pz = dynamic_cast<PuzzleStone*>(GetStone(curpos));
        assert(pz);

        cluster.push_back(curpos);
        DirectionBits cfaces = pz->get_connections();

        if (cfaces==NODIRBIT) 
            cfaces = DirectionBits(NORTHBIT | SOUTHBIT | EASTBIT | WESTBIT);

        if (has_dir(cfaces, NORTH))
            is_complete &= visit_dir(pos_stack, curpos, NORTH);
        if (has_dir(cfaces, EAST))
            is_complete &= visit_dir(pos_stack, curpos, EAST);
        if (has_dir(cfaces, SOUTH))
            is_complete &= visit_dir(pos_stack, curpos, SOUTH);
        if (has_dir(cfaces, WEST))
            is_complete &= visit_dir(pos_stack, curpos, WEST);
    }
    return is_complete;
}

void
PuzzleStone::find_row_or_column_cluster(Cluster &c, GridPos startpos, Direction dir)
{
    GridPos p = startpos;
    while (dynamic_cast<PuzzleStone*>(GetStone(p))) {
        c.push_back(p);
        p.move(dir);
    }
}

void 
PuzzleStone::maybe_move_cluster(Cluster &c, bool is_complete, Direction dir)
{
    sort(c.begin(), c.end());
    Cluster mc(c);              // Moved cluster
    Cluster diff;               // Difference between mc and c

    for (unsigned i=0; i<mc.size(); ++i)
        mc[i].move(dir);

    set_difference(mc.begin(), mc.end(), c.begin(), c.end(),
                   back_inserter(diff));

    // Now check whether all stones can be placed at their new
    // position
    bool move_ok = true;
    for (unsigned i=0; i<diff.size(); ++i)
        if (GetStone(diff[i]) != 0)
            move_ok = false;

    if (!move_ok)
        return;

    // If the floor at the cluster's new position consists exclusively
    // of abyss or water, create a bridge instead of moving the
    // cluster.
    bool create_bridge=false;
    if (is_complete) {
        create_bridge=true;
        for (unsigned i=0; i<mc.size(); ++i) {
            if (Floor *fl = GetFloor(mc[i])) {
                const string &k = fl->get_kind();
                if (!(k=="fl-abyss" || k=="fl-water")) {
                    create_bridge=false;
                    break;
                }
            }
        }
    }

    // Finally, either move the whole cluster or create a bridge
    play_sound("st-move");
    if (create_bridge) {
        for (unsigned i=0; i<c.size(); ++i) {
            KillStone(c[i]);
            SetFloor(mc[i], MakeFloor("fl-gray"));
        }
    } else { 
        vector<Stone*> clusterstones;
        for (unsigned i=0; i<c.size(); ++i)
            clusterstones.push_back(YieldStone(c[i]));
        
        for (unsigned i=0; i<c.size(); ++i)
            SetStone(mc[i], clusterstones[i]);
    }
}

bool 
PuzzleStone::cluster_complete()
{
    Cluster c;
    return find_cluster(c);
}

void 
PuzzleStone::init_model()
{
    string modelname = "st-puzzle";
    char x[10];

    // Every combination of ``sockets'' has its own model.
    int modelno = int_attrib("connections");
    sprintf(x, "%d", modelno);
    modelname += x;
    set_model(modelname);
}

void
PuzzleStone::rotate_cluster(const Cluster &c)
{
    if (c.size() > 1) {
        vector<Stone*> stones;
        for (unsigned i=0; i<c.size(); ++i)
            stones.push_back(YieldStone(c[i]));

        // exclude last list item
        for (unsigned i=0; i<c.size()-1; ++i)
            SetStone(c[i+1], stones[i]);
        SetStone(c[0], stones.back());
    }
}

StoneResponse
PuzzleStone::collision_response(const StoneContact &sc)
{
    if (get_connections() == NODIRBIT)
        return STONE_PASS;
    return STONE_REBOUND;
}

void
PuzzleStone::actor_hit(const StoneContact &sc)
{
    if (get_connections() == NODIRBIT)
        return;

    Actor *a = sc.actor;

    bool move_impulse= (a->get_vel()*sc.normal < -4);
    Direction hit_dir=reverse(contact_face(sc));

    if (wielded_item_is(a, "it-magicwand")) {
        if (cluster_complete()) {
            // make all cluster stones explode
            return;
        } 
        else if (hit_dir != NODIR) {
            Cluster c;
            find_row_or_column_cluster(c, get_pos(), hit_dir);

            bool is_complete = false;

            if (move_impulse)
                maybe_move_cluster(c, is_complete, hit_dir);
            else
                rotate_cluster(c);
        }
    } 
    else if (move_impulse && hit_dir != NODIR) {
        Cluster c;
        bool is_complete = find_cluster(c);
        maybe_move_cluster(c, is_complete, hit_dir);
    }
}


//----------------------------------------
// DoorBase
//
// Base class for everything that behaves like a door, i.e., it has
// four states OPEN, CLOSED, OPENING, CLOSING.
//----------------------------------------
namespace
{
    class DoorBase : public Stone {
    protected:
        DoorBase(const char *name) : Stone(name), state(CLOSED) {}
        enum State { OPEN, CLOSED, OPENING, CLOSING };

        State get_state() const { return state; }
        void set_state(State st) { state=st; }

        State state;

    private:
        virtual string model_basename() { return get_kind(); }
        virtual void init_model();
        virtual string opening_sound() const { return ""; }
        virtual string closing_sound() const { return ""; }

        void change_state(State newstate) ;
        void message(const string &m, const Value &);

        void actor_inside(Actor *a);
        StoneResponse collision_response(const StoneContact &sc);
        void actor_hit(const StoneContact &sc);
        void actor_contact(Actor *a);

        void animcb();
        bool on_laserhit(Direction dir) {return state==OPEN;}
    };
}

void 
DoorBase::message(const string &m, const Value &)
{
    if (m == "open" && (state==CLOSED || state==CLOSING)) 
        change_state(OPENING);
    else if (m=="close" && (state==OPEN || state==OPENING))
        change_state(CLOSING);
    else if (m=="openclose") {
        if (state==OPEN || state==OPENING)
            change_state(CLOSING);
        else 
            change_state(OPENING);
    }
}

void 
DoorBase::init_model() 
{
    string mname = model_basename();
    if (state == CLOSED)
        mname += "-closed";
    else if (state==OPEN)
        mname += "-open";
    set_model(mname);
}

void 
DoorBase::animcb() 
{
    if (state == OPENING)
        change_state(OPEN);
    else if (state == CLOSING)
        change_state(CLOSED);
}


StoneResponse
DoorBase::collision_response(const StoneContact &sc)
{
    return (state == OPEN) ? STONE_PASS:STONE_REBOUND;
}

void
DoorBase::actor_hit(const StoneContact &sc)
{
    if (state == CLOSING)
        SendMessage(sc.actor, "shatter");
}

void 
DoorBase::change_state(State newstate) 
{
    string basename = model_basename();

    switch (newstate) {
    case OPEN: 
        set_model(basename+"-open"); 
        laser::MaybeRecalcLight(get_pos());
        break;
    case CLOSED: 
        set_model(basename+"-closed"); 
        laser::MaybeRecalcLight(get_pos());
        break;
    case OPENING:
        play_sound(opening_sound().c_str());
        if (state == CLOSING)
            get_model()->reverse();
        else
            set_anim(basename+"-opening");
        break;
    case CLOSING:
        play_sound(closing_sound().c_str());
        if (state == OPENING)
            get_model()->reverse();
        else
            set_anim(basename+"-closing");
        break;
    }
    set_state(newstate);
}

void 
DoorBase::actor_inside(Actor *a)
{
    if (state == CLOSING)
        SendMessage(a, "shatter");
}

void 
DoorBase::actor_contact(Actor *a)
{
    if (state == CLOSING)
        SendMessage(a, "shatter");
}

//----------------------------------------
// Door
//
// Attributes:
//
// :type        h or v for a door that opens horizontally or vertically
//----------------------------------------
namespace
{
    class Door : public DoorBase {
        CLONEOBJ(Door);
    public:
        Door() : DoorBase("st-door") {set_attrib("type", "h");}
    private:
        virtual string opening_sound() const { return "dooropen"; }
        virtual string closing_sound() const { return "doorclose"; }
	virtual const char *collision_sound() { return "electric"; }
        string get_type() const {
            string type="h";
            string_attrib("type", &type);
            return type;
        }

        bool on_laserhit(Direction dir);

        string model_basename() { return string("st-door")+get_type(); }
        StoneResponse collision_response(const StoneContact &sc);
    };

    class Door_a : public DoorBase {
        CLONEOBJ(Door_a);
    public:
        Door_a() : DoorBase("st-door_a") {}
    };

    class Door_b : public DoorBase {
        CLONEOBJ(Door_b);
    public:
        Door_b() : DoorBase("st-door_b") {}
    };

    class Door_c : public DoorBase {
        CLONEOBJ(Door_c);
    public:
        Door_c() : DoorBase("st-door_c") {}
    };
}

bool 
Door::on_laserhit(Direction dir) {
    if (get_type() == "h")
        return state==OPEN || dir==EAST || dir==WEST;
    else
        return state==OPEN || dir==NORTH || dir==SOUTH;
}

StoneResponse 
Door::collision_response(const StoneContact &sc)
{
    Direction cf = contact_face(sc);
    if (state == OPEN)
        return STONE_PASS;
    else if (state == CLOSING)
        return STONE_REBOUND;
    else {
        string t = get_type();
        return ((t == "v" && (cf==WEST || cf==EAST)) ||
                (t == "h" && (cf==SOUTH || cf==NORTH)))
            ? STONE_REBOUND 
            : STONE_PASS;
    }
}

//----------------------------------------
// TimerStone
//
// Attributes:
//
// :interval        seconds between two "ticks"
// :loop            
// :action,target   as usual
//----------------------------------------

/** \page st-timer Timer Stone

This stone can be used to trigger periodic events or to trigger one
single event after a certain amount of time.

\subsection timera Attributes

- \b on: 1 if the timer is running
- \b interval:  number of seconds before \b action is performed
- \b loop:      if 1, restart the timer after performing \b action
- \b action, \b target: as usual

\subsection timerm Messages

- \b on, \b off, \b onoff: as usual

\subsection timere Example

\verbatim
-- activate a laser after 5 seconds
set_stone("st-laser", 10,11, {name="laser"})
set_stone("st-timer", 10,10, 
          {loop=0, action="onoff", target="laser", interval=5})
\endverbatim
*/
namespace
{
    class TimerStone : public OnOffStone, public TimeHandler 
    {
        CLONEOBJ(TimerStone);
    public:
        TimerStone() : OnOffStone("st-timer") {
            set_attrib("interval", 1.0);
            set_attrib("loop", 1.0);
            set_attrib("on", 1.0);

            // set_on(true);   DOESN'T WORK! calls init_model()
        }
    private:
        double get_interval() const { 
            double interval = 100;
            double_attrib("interval", &interval);
            return interval;
        }
        void init_model() { set_model(is_on() ? "st-timer" : "st-timeroff"); }

        void on_creation() {
            set_alarm();
            Stone::on_creation();
        }

        void set_alarm() {
            if (is_on())
                g_timer.set_alarm(this, get_interval(), true);
        }

        void alarm() {
            if (is_on())
            {
//                 sound::PlaySound("st-timer");
                PerformAction(this, is_on());
            }
        }
    };
}

//----------------------------------------
// Switch
//
// Attributes:
//
// :on              1 or 0
// :target,action   as usual
//----------------------------------------

/** \page st-switch Switch Stone

On actor hit this stone can trigger actions
  
\image html st-switch1.png
*/
namespace
{
    class SwitchStone : public OnOffStone {
        CLONEOBJ(SwitchStone);
    public:
        SwitchStone() : OnOffStone("st-switch") {}
    private:
        void init_model() {
            set_model(is_on() ? "st-switch1" : "st-switch0");
        }

        void actor_hit(const StoneContact &sc) {
            set_on(!is_on());
            PerformAction(this, is_on());
            play_sound("st-switch");
        }
        const char *collision_sound() { return "st-metal"; }
    };
}

//----------------------------------------
// Laser Switch
//----------------------------------------

/** \page st-laserswitch Laserswitch Stone

Activated by laser light this switch can trigger actions
  
*/
namespace
{
    class LaserSwitch : public laser::PhotoStone {
        CLONEOBJ(LaserSwitch);
    public:
        LaserSwitch();
        ~LaserSwitch();

    private:
        enum State { ON, OFF };
        State state;

        display::Model *model_bak;

        // Stone interface
        void actor_hit(const StoneContact &sc);
        void on_creation();
        void on_removal();
        const char *collision_sound() { return "st-metal"; }

        // PhotoStone interface.
        void notify_laseron() { set_model("st-laserswitch1"); PerformAction(this, true);}
        void notify_laseroff() { set_model("st-laserswitch0"); PerformAction(this, false);}
    };
}

LaserSwitch::LaserSwitch() 
    : PhotoStone("st-laserswitch"), state(OFF), model_bak(0)
{
}

LaserSwitch::~LaserSwitch()
{
    delete model_bak;
}

void 
LaserSwitch::actor_hit(const StoneContact &sc)
{
}

void 
LaserSwitch::on_creation()
{
    if (model_bak) {
        display::SetModel(GridLoc(GRID_STONES, get_pos()), model_bak);
        model_bak=0;
    } 
    else {
        set_model("st-laserswitch0");
    }
    
    photo_activate();
}

void 
LaserSwitch::on_removal()
{
    photo_deactivate();
    model_bak = display::YieldModel(GridLoc(GRID_STONES, get_pos()));
}

//----------------------------------------
// Floppy Switch
//
// Attributes:
//
// :on              1 or 0
// :target,action   as usual
//----------------------------------------
namespace
{
    class FloppyStone : public OnOffStone 
    {
        CLONEOBJ(FloppyStone);
    public:
        FloppyStone() : OnOffStone("st-floppy") {}
    private:
        void init_model() {
            set_model(is_on() ? "st-floppy1" : "st-floppy0");
        }

        void actor_hit(const StoneContact &sc) 
        {
            if (player::Inventory *inv = get_inventory(sc.actor))
            {
                if (is_on()) 
                {
                    inv->add_item(MakeItem("it-floppy"));
                    set_on(false);
                    PerformAction(this, is_on());
                }
                else if (wielded_item_is(sc.actor, "it-floppy")) 
                {
                    DisposeObject(inv->yield_first());
                    set_on(true);
                    PerformAction(this, is_on());
                }
            }
        }
        const char *collision_sound() { return "st-metal"; }
    };
}

//----------------------------------------
// ShogunStone
//
// Attributes:
//
// :holes            1..7
//----------------------------------------
namespace
{
    class ShogunStone : public Stone {
        CLONEOBJ(ShogunStone);
    public:
        ShogunStone() : Stone("st-shogun") {
            set_holes(SMALL);
        }
    private:
        enum Holes { SMALL=1, MEDIUM=2, LARGE=4 };

        void set_holes(Holes h) { set_attrib("holes", h); }
        Holes get_holes() const;
        void notify_item();

        void add_hole(Holes h) {
            set_attrib("holes", get_holes() | h);
            notify_item();
            init_model();
        }

        void on_creation() { 
            init_model(); 
            notify_item();
        }

        void init_model() {
            char x[20];
            sprintf(x, "st-shogun%d", int(get_holes()));
            set_model(x);
        }

        static Holes smallest_hole(Holes s);

        Holes transfer_smallest_hole(GridPos p);
        void actor_hit(const StoneContact &sc);
    };
}

ShogunStone::Holes 
ShogunStone::get_holes() const 
{ 
    int h=int_attrib("holes"); 
    if (h>=1 && h<=7)
        return Holes(h);
    else {
        fprintf(stderr, "Wrong `holes' attribute\n");
        return SMALL;
    }
}

ShogunStone::Holes 
ShogunStone::smallest_hole(Holes s) 
{
    if (s & SMALL) return SMALL;
    if (s & MEDIUM) return MEDIUM;
    if (s & LARGE) return LARGE;
    assert(0);
}

void 
ShogunStone::notify_item() 
{
    if (Item *it=GetItem(get_pos())) {
        switch (get_holes()) {
        case SMALL: SendMessage(it,"shogun1"); break;
        case (MEDIUM | SMALL): SendMessage(it,"shogun2"); break;
        case (LARGE | MEDIUM | SMALL): SendMessage(it,"shogun3"); break;
        default: SendMessage(it, "noshogun"); break;
        }
    }
}

ShogunStone::Holes
ShogunStone::transfer_smallest_hole(GridPos p)
{
    bool hole_transferred=false;
    Holes h = get_holes();
    Holes sh = smallest_hole(h);


    if (Stone *st=GetStone(p)) {
        if (ShogunStone *sst = dynamic_cast<ShogunStone*>(st)) {
            // Transfer the smallest hole to the other shogun stone.
            // Holes must be ``stacked'' properly
            if (sh < smallest_hole(sst->get_holes())) {
                sst->add_hole(sh);
                hole_transferred=true;
            }
        }
    }
    else {
        // empty field? Simply create a new stone on it
        ShogunStone *sst = new ShogunStone();
        sst->set_holes(sh);
        SetStone(p, sst);
        hole_transferred=true;
    }

    return (hole_transferred) ? Holes(h & ~sh) : h;
}

void
ShogunStone::actor_hit(const StoneContact &sc)
{
    Actor *a = sc.actor;
    bool move_impulse= (a->get_vel()*sc.normal < -4);
    Direction hit_dir=reverse(contact_face(sc));

    if (hit_dir == NODIR || !move_impulse)
        return;

    GridPos newpos=move(get_pos(), hit_dir);
    Holes holes = get_holes();
    Holes newholes = transfer_smallest_hole(newpos);

    if (newholes!=holes) {
        play_sound("st-move");
        if (newholes == 0) {
            KillStone(get_pos());
        } else {
            set_holes(newholes);
            notify_item();
            init_model();
        }
    }
}


//----------------------------------------
// FartStone
//----------------------------------------

/** \page st-fart Fart Stone

The fart stone has the unpleasant habit of "blowing off" when triggered
(by actor hit or signal) and will close all oxyd stones.

\subsection fartm Messages

- \b trigger: as usual

*/

namespace
{
    class FartStone : public Stone {
        CLONEOBJ(FartStone);
    public:
        FartStone() : Stone("st-fart"), state(IDLE) {}
    private:
        enum State { IDLE, FARTING, BROKEN };
        State state;

        void fart() {
            if (state == IDLE) {
                Object *ox = world::GetObjectTemplate("st-oxyd");
                SendMessage(ox, "closeall");
                play_sound("fart");
                set_anim("st-farting");
                state = FARTING;
            }
        }
        void destroy() {

        }
        void animcb() { 
            if (state == FARTING) {
                state = IDLE;
                init_model(); 
            }
        }

        void actor_hit(const StoneContact &sc) {
            if (wielded_item_is(sc.actor, "it-hammer")) {
                // ### What we really want is either a broken "fart"
                // ### stone or an animation
                KillStone(get_pos());
            } else {
                fart();
            }
        }
        void message(const string &m, const Value &) {
            if (m=="trigger")
                fart();
        }
    };
}

//----------------------------------------
// Thief Stone
//
// Takes one item from inventory after actor hit.
//
// TODO: It should steal afer some period of time
//----------------------------------------
namespace
{
    class ThiefStone : public Stone {
        CLONEOBJ(ThiefStone);
    public:
        ThiefStone() : Stone("st-thief"), state(IDLE) { }
    private:
        enum State { IDLE, STEALING } state;

        void actor_hit(const StoneContact &sc) {
            if (state==IDLE) {
                set_anim("st-thief-anim");
                if (player::Inventory *inv = get_inventory(sc.actor))
                {
                    int i = int(inv->size() * (rand()/(RAND_MAX+1.0)));
		    delete inv->yield_item(i);
                    
                    play_sound("thief");
                    state = STEALING;
                }
            }
	}
        
        void animcb() {
            if (state == STEALING) {
                state = IDLE;
                init_model();
            }
        }

        const char *collision_sound() {
            return "st-thud";
        }
    };
}


//----------------------------------------
// ActorImpulseStone
//----------------------------------------
namespace
{
    class ActorImpulseStone : public Stone {
        CLONEOBJ(ActorImpulseStone);
    public:
        ActorImpulseStone() : Stone("st-actorimpulse"), state(IDLE) {}
    private:
	enum State { IDLE, PULSING, BROKEN };
	State state;
	
	void actor_hit (const StoneContact &sc) {
            if (state == IDLE) {
                double forcefac = 200;
                sc.actor->add_force (forcefac * normalize(sc.normal));
	        play_sound("impulse");
                set_anim("st-actorimpulse-anim");
                state = PULSING;
            }
	}

        void animcb() {
            if (state == PULSING) {
                state = IDLE;
                init_model();
            }
        }
    };
}


//----------------------------------------
// StoneImpulseStone
//
// Messages:
//
// :trigger
//----------------------------------------
namespace
{
    class StoneImpulseStone : public Stone {
        CLONEOBJ(StoneImpulseStone);
    public:
        StoneImpulseStone() : Stone("st-stoneimpulse"), state(IDLE) {}
    private:
        enum State { IDLE, PULSING, CLOSING, BROKEN };
        State state;

        void stoneimpulse() {
            if (state == IDLE) {
                play_sound("impulse");
                set_anim("st-stoneimpulse-anim1");
                state = PULSING;
            }
        }
        void animcb() {
            if (state == PULSING) {
                state = CLOSING;
                GridPos p = get_pos();

                MaybeMoveStone (move(p,NORTH), NORTH);
                MaybeMoveStone (move(p,EAST), EAST);
                MaybeMoveStone (move(p,SOUTH), SOUTH);
                MaybeMoveStone (move(p,WEST), WEST);
                
                SendMessage(GetStone(move(p, NORTH)), "stoneimpulse");
                SendMessage(GetStone(move(p, EAST)), "stoneimpulse");
                SendMessage(GetStone(move(p, SOUTH)), "stoneimpulse");
                SendMessage(GetStone(move(p, WEST)), "stoneimpulse");
                set_anim ("st-stoneimpulse-anim2");
            } else if (state == CLOSING) {
                state = IDLE;
                init_model();
            }
        }

        void actor_hit(const StoneContact &sc) {
            stoneimpulse();
        }
        void message(const string &m, const Value &) {
            if (m=="trigger" || m=="stoneimpulse")
                stoneimpulse();
        }
    };
}

//----------------------------------------
// Hollow StoneImpulseStone
//
// Messages:
//
// :trigger
//----------------------------------------
namespace
{
    class HollowStoneImpulseStone : public Stone {
        CLONEOBJ(HollowStoneImpulseStone);
    public:
        HollowStoneImpulseStone() : Stone("st-stoneimpulse-hollow"), state(IDLE) {}
    private:
        enum State { IDLE, PULSING, CLOSING, BROKEN };
        State state;

        void stoneimpulse() {
            if (state == IDLE) {
                play_sound("impulse");
                set_anim("st-stoneimpulse-hollow-anim1");
                state = PULSING;
                GridPos p = get_pos();
                MaybeMoveStone( move( p, NORTH), NORTH);
                MaybeMoveStone( move( p, EAST), EAST);
                MaybeMoveStone( move( p, SOUTH), SOUTH);
                MaybeMoveStone( move( p, WEST), WEST);
            }
        }
        void animcb() {
            if (state == PULSING) {
                state = CLOSING;
                GridPos p = get_pos();
                SendMessage(GetStone(move(p, NORTH)), "stoneimpulse");
                SendMessage(GetStone(move(p, EAST)), "stoneimpulse");
                SendMessage(GetStone(move(p, SOUTH)), "stoneimpulse");
                SendMessage(GetStone(move(p, WEST)), "stoneimpulse");
                set_anim("st-stoneimpulse-hollow-anim2");
            } else if (state == CLOSING) {
                state = IDLE;
                init_model();
            }
        }

        StoneResponse collision_response(const StoneContact &sc) {
            if( state == IDLE)
                return STONE_PASS;
            return STONE_REBOUND;
        }
        void message(const string &m, const Value &) {
            if (m=="trigger")
                stoneimpulse();
            if (m=="stoneimpulse")
                stoneimpulse();
        }

        void actor_inside(Actor *a) {
            if( state == PULSING || state == CLOSING)
                SendMessage(a, "shatter");
        }
        
        bool on_laserhit(Direction) { return true; }
    };
}

//----------------------------------------
// OxydStone
//----------------------------------------

/** \page st-oxyd Oxyd Stone

Oxyd stones are characterized by two attributes: Their flavor and
their color.  The flavor only affects the visual representation of
the stone; it can be either 'a' (opening like a flower) or 'b'
(displaying a fade-in animation).  The color attribute determines
the shape on the oxyd stone.

\b Note: You should usually not to create Oxyd stones manually 
with \c set_stone(). Use the predefined \c oxyd() function instead.

\subsection oxyda Attributes

- \b flavor      "a", "b", "c", or "d"
- \b color       number between 0 and 7

\subsection oxydm Messages

- \b closeall    close all oxyd stones
- \b shuffle     interchange the colors of the oxyd stones in the current landscape
- \b trigger     open the stone

<table><tr>
<td>\image html st-oxyda.png "flavor A"
<td>\image html st-oxydb.png "flavor B"
<td>\image html st-oxydc.png "flavor C"
<td>\image html st-oxydd.png "flavor D"
</table>
*/

namespace
{
    class OxydStone : public laser::PhotoStone {
        INSTANCELISTOBJ(OxydStone);
    public:
        OxydStone();
        ~OxydStone();

        static void shuffle_colors();
    private:
        enum State { CLOSED, OPEN, OPENING, CLOSING, BLINKING };
        State state;

        display::Model *model_bak;

        static bool blinking_or_opening(OxydStone *a) { 
            return (a->state==BLINKING || a->state==OPENING); 
        }
        static bool not_open(OxydStone *a) {
            return !(a->state==OPEN || a->state==OPENING);
        }

        // Stone interface
        void actor_hit(const StoneContact &sc);
        void on_creation();
        void on_removal();
        const char *collision_sound() { return "st-stone"; }
        void message(const string &m, const Value &);

        // Animation callback
        void animcb();

        void maybe_open_stone();
        void change_state(State newstate);

        void notify_laseron() { maybe_open_stone(); }
        void notify_laseroff() {}
    };
}

OxydStone::InstanceList OxydStone::instances;

OxydStone::OxydStone() 
: PhotoStone("st-oxyd")
, state(CLOSED)
, model_bak(0)
{
    set_attrib("flavor", "b");
    set_attrib("color", "0");
}

OxydStone::~OxydStone()
{
    delete model_bak;
}

void 
OxydStone::message(const string &m, const Value &)
{
    if (m=="closeall")
        for (unsigned i=0; i<instances.size(); ++i) 
            instances[i]->change_state(CLOSING);
    else if (m=="shuffle") 
        shuffle_colors();
    else if (m=="trigger")
	maybe_open_stone();
}

void 
OxydStone::shuffle_colors()
{
    string icolor, acolor;
    for (unsigned i=0; i<instances.size(); ++i) {
        int a=int(instances.size() * (rand()/(RAND_MAX+1.0)));
        
        instances[i]->string_attrib("color", &icolor);
        instances[a]->string_attrib("color", &acolor);
        
        instances[i]->set_attrib("color", acolor.c_str());
        instances[a]->set_attrib("color", icolor.c_str());
    }
}

void
OxydStone::change_state(State newstate)
{
    string flavor = "a";
    string color = "1";
    string_attrib("flavor", &flavor);
    string_attrib("color", &color);

    string modelname = string("st-oxyd") + flavor + color;

    State oldstate = state;
    state = newstate;
    
    switch (newstate) {
    case CLOSED: 
        set_model(string("st-oxyd")+flavor);
        break;

    case BLINKING:
        set_model(modelname + "-blink");
        break;
	
    case OPEN:
	if (oldstate == CLOSED) {
            play_sound("st-oxydopen");
            play_sound("st-oxydopened");
            set_anim(modelname+"-opening");
        } else {
            set_model(modelname + "-open");
        }
        /* If this was the last closed oxyd stone, finish the
           level */
        if (find_if(instances.begin(),instances.end(),not_open)
            ==instances.end())
        {
            enigma::FinishLevel();
        }
        break;
	
    case OPENING:
        play_sound("st-oxydopen");
	if (oldstate == CLOSED)
            set_anim(modelname + "-opening");
	else if (oldstate == CLOSING)
            get_model()->reverse();

	break;
	
    case CLOSING:
        if (oldstate == CLOSED || oldstate==CLOSING) {
            state = oldstate;
            return;
        }

        play_sound("st-oxydclose");
        if (oldstate == OPENING)
            get_model()->reverse();
	else if (oldstate == BLINKING || oldstate == OPEN) {
            set_anim(modelname + "-closing");
        }
        break;
    }
}

void 
OxydStone::animcb()
{
    if (state == CLOSING)
        change_state(CLOSED);
    else if (state == OPENING)
        change_state(BLINKING);
    else if (state == OPEN)
        change_state(OPEN); // set the right model
}

void
OxydStone::maybe_open_stone()
{
    if (state == CLOSED || state == CLOSING) {
        int mycolor = int_attrib("color");

        // Is another oxyd stone currently blinking?
        InstanceList::iterator i;
        i=find_if(instances.begin(), instances.end(), blinking_or_opening);

        if (i != instances.end()) {
            // If colors match, open both stones. Close one of them
            // otherwise
            if (mycolor == (*i)->int_attrib("color")) {
                change_state(OPEN);
                (*i)->change_state(OPEN);
            } else {
                (*i)->change_state(CLOSING);
                change_state(OPENING);
            }
        } 
        else {
            // no blinking stone? -> make this one blink
            change_state(OPENING);
        }
    }
}

void 
OxydStone::actor_hit(const StoneContact &sc)
{
    maybe_open_stone();
}

void 
OxydStone::on_creation()
{
    if (model_bak) {
        display::SetModel(GridLoc(GRID_STONES, get_pos()), model_bak);
        model_bak=0;
    } 
    else 
    {
        string flavor = "a";
        string_attrib("flavor", &flavor);
        set_model(string("st-oxyd") + flavor);
    }
    
    photo_activate();
}

void 
OxydStone::on_removal()
{
    photo_deactivate();
    model_bak = display::YieldModel(GridLoc(GRID_STONES, get_pos()));
}

//----------------------------------------
// FakeOxydStone
//----------------------------------------

/** \page st-fakeoxyd Fake Oxyd Stone

These stones look like real Oxyd stones, but they only blink a little
when touched and do not open or have other special abilities.

\image html st-fakeoxyd-blink_0001.png
*/
namespace
{
    class FakeOxydStone : public Stone {
        CLONEOBJ(FakeOxydStone);
    public:
        FakeOxydStone() : Stone("st-fakeoxyd"), state(IDLE) {}
    private:
        enum State { IDLE, BLINKING } state;
        void actor_hit(const StoneContact &sc) {
            if (state == IDLE) {
                set_anim("st-fakeoxyd-blink");
                state = BLINKING;
            }
        }
        const char *collision_sound() {
            return "st-fakeoxyd";
        }
        void animcb() {
            set_model("st-fakeoxyd");
            state = IDLE;
        }
    };
}

//----------------------------------------
// Grate1
//----------------------------------------
namespace
{
    class Grate1 : public Stone {
        CLONEOBJ(Grate1);
    public:
        Grate1() : Stone("st-grate1") {}
        StoneResponse collision_response(const StoneContact &sc) {
            return STONE_PASS;
        }
        bool on_laserhit(Direction) { return true; }
    };
}

//----------------------------------------
// Grate2
//----------------------------------------
namespace
{
    class Grate2 : public Stone {
        CLONEOBJ(Grate2);
    public:
        Grate2() : Stone("st-grate2") {}
        StoneResponse collision_response(const StoneContact &sc) {
            return STONE_PASS;
        }
        bool on_laserhit(Direction) { return true; }
    };
}

//----------------------------------------
// Marble_hole
//----------------------------------------
namespace
{
    class Marble_hole : public Stone {
        CLONEOBJ(Marble_hole);
    public:
        Marble_hole() : Stone("st-marble_hole") {}
        StoneResponse collision_response(const StoneContact &sc) {
            return STONE_PASS;
        }
        bool on_laserhit(Direction) { return true; }
    };
}

//----------------------------------------
// Greenbrown_hole
//----------------------------------------
namespace
{
    class Greenbrown_hole : public Stone {
        CLONEOBJ(Greenbrown_hole);
    public:
        Greenbrown_hole() : Stone("st-greenbrown_hole") {}
        StoneResponse collision_response(const StoneContact &sc) {
            return STONE_PASS;
        }
        bool on_laserhit(Direction) { return true; }
    };
}

//----------------------------------------
// Rock3_hole
//----------------------------------------
namespace
{
    class Rock3_hole : public Stone {
        CLONEOBJ(Rock3_hole);
    public:
        Rock3_hole() : Stone("st-rock3_hole") {}
        StoneResponse collision_response(const StoneContact &sc) {
            return STONE_PASS;
        }
        bool on_laserhit(Direction) { return true; }
    };
}

//----------------------------------------
// Black stones
//----------------------------------------
namespace
{
    class BlackStone : public Stone {
        CLONEOBJ(BlackStone);
    public:
	BlackStone() : Stone("st-black1") {}
	BlackStone(const char *name) : Stone( name) {}
        StoneResponse collision_response(const StoneContact &sc) {
	    if (sc.actor->get_attrib("blackball")) 
                return STONE_PASS;
	    else
                return STONE_REBOUND;
        }
    };

    class BlackStone2 : public BlackStone {
        CLONEOBJ(BlackStone2);
    public:
        BlackStone2() : BlackStone("st-black2") {}
    };

    class BlackStone3 : public BlackStone {
        CLONEOBJ(BlackStone3);
    public:
        BlackStone3() : BlackStone("st-black3") {}
    };
    
    class BlackStone4 : public BlackStone {
        CLONEOBJ(BlackStone4);
    public:
        BlackStone4() : BlackStone("st-black4") {}
    };

}

//----------------------------------------
// White stones
//----------------------------------------
namespace
{
    class WhiteStone : public Stone {
        CLONEOBJ(WhiteStone);
    public:
        WhiteStone() : Stone("st-white1") {}
        WhiteStone(const char *name) : Stone( name) {}
        StoneResponse collision_response(const StoneContact &sc) {
	    if (sc.actor->get_attrib("whiteball")) 
                return STONE_PASS;
	    else
                return STONE_REBOUND;
        }
    };

    class WhiteStone2 : public WhiteStone {
        CLONEOBJ(WhiteStone2);
    public:
        WhiteStone2() : WhiteStone("st-white2") {}
    };

    class WhiteStone3 : public WhiteStone {
        CLONEOBJ(WhiteStone3);
    public:
        WhiteStone3() : WhiteStone("st-white3") {}
    };

    class WhiteStone4 : public WhiteStone {
        CLONEOBJ(WhiteStone4);
    public:
        WhiteStone4() : WhiteStone("st-white4") {}
    };

}

//----------------------------------------
// YinYang stone 1
//----------------------------------------
namespace
{
    class YinYangStone1 : public Stone {
        CLONEOBJ(YinYangStone1);
    public:
	YinYangStone1() : Stone("st-yinyang1"), state(NOCOLOR) {}
    private:
        enum State { NOCOLOR, WHITE, BLACK } state;

        void init_model() {
            switch (state) {
            case NOCOLOR: set_model("st-yinyang1"); break;
            case WHITE: set_model("st-white1"); break;
            case BLACK: set_model("st-black1"); break;
            }
        }
        
	void actor_hit(const StoneContact &sc)
        {
	    if (state==NOCOLOR) {
                if (sc.actor->get_attrib("blackball")) {
		    state = WHITE;
                    init_model();
                    play_sound("st-magic");
		}
                else if (sc.actor->get_attrib("whiteball")) {
		    state = BLACK;
		    init_model();
                    play_sound("st-magic");
		}
	    }
        }
        
	StoneResponse collision_response(const StoneContact &sc)
        {
	    if ((state==BLACK && sc.actor->get_attrib("blackball")) ||
                (state==WHITE && sc.actor->get_attrib("whiteball")))
                return STONE_PASS;
            return STONE_REBOUND;
	}
    };
}

//----------------------------------------
// YinYang stone 2
//----------------------------------------
namespace
{
    class YinYangStone2 : public Stone {
        CLONEOBJ(YinYangStone2);
    public:
	YinYangStone2() : Stone("st-yinyang2"), state(NOCOLOR) {}
    private:
        enum State { NOCOLOR, WHITE, BLACK } state;

        void init_model() {
            switch (state) {
            case NOCOLOR: set_model("st-yinyang2"); break;
            case WHITE: set_model("st-white1"); break;
            case BLACK: set_model("st-black1"); break;
            }
        }
        
	void actor_hit(const StoneContact &sc)
        {
	    if (state==NOCOLOR) {
                if (sc.actor->get_attrib("blackball")) {
		    state = BLACK;
		    init_model();
                    play_sound("st-magic");
		}
                else if (sc.actor->get_attrib("whiteball")) {
		    state = WHITE;
                    init_model();
                    play_sound("st-magic");
		}
	    }
        }
        
	StoneResponse collision_response(const StoneContact &sc)
        {
	    if ((state==BLACK && sc.actor->get_attrib("blackball")) ||
                (state==WHITE && sc.actor->get_attrib("whiteball")))
                return STONE_PASS;
            return STONE_REBOUND;
	}
    };
}


//----------------------------------------
// BombStone
//
// These stones add a bomb to the player's inventory when touched.
//----------------------------------------
namespace
{
    class BombStone : public Stone {
        CLONEOBJ(BombStone);
    public:
        BombStone() : Stone("st-bombs") {}
    private:
        void actor_hit(const StoneContact &sc);
    };
}

void
BombStone::actor_hit(const StoneContact &sc)
{
    if (player::Inventory *inv = get_inventory(sc.actor))
    {
        if (!inv->is_full())
        {
            Item *it = MakeItem("it-blackbomb");
            inv->add_item(it);
        }
    }
}

//----------------------------------------
// GlassStone
//
// Laser light may pass this stone, but actors may not.
//----------------------------------------
namespace
{
    class GlassStone : public Stone {
        CLONEOBJ(GlassStone);
    public:
        GlassStone() : Stone("st-glass") {}
    private:
        bool on_laserhit(Direction dir) { return true; }
    };
}

//----------------------------------------
// MagicStone
//----------------------------------------
namespace
{
    class MagicStone : public Stone {
        CLONEOBJ(MagicStone);
    public:
        MagicStone() : Stone("st-magic") {}
    private:

    };
}
        

//----------------------------------------
// DeathStone
//----------------------------------------

/** \page st-death Death's Head Stone

Simply kills all actors that touch it (except for actors that are
immune to these stones).

\image html st-death.png
*/
namespace
{
    class DeathStone : public Stone {
        CLONEOBJ(DeathStone);
    public:
        DeathStone() : Stone("st-death"), active(false) {}
    private:
        bool active; 
        void actor_hit(const StoneContact &sc) 
        {
	    if (!active) {
		SendMessage(sc.actor, "shatter");
		active=true;
		set_anim("st-death-munch");
	    }
        }
	void animcb() { set_model("st-death"); active=false; }
    };
}


//----------------------------------------
// Coin slot Switch
//----------------------------------------

/** \page st-coinslot Coin Slot Stone 

Coin Slots are switches that must be activated by inserting coins.
The more coins you put in, the longer the switch will stay activated

\subsection coinslota Attributes

- \b cost            how many on-time bought for coin1
- \b interval        on-time remaining
- \b target, \b action   as usual

\subsection coinslote Example

\verbatim
set_stone("st-coinslot", 20,13,{action="openclose", target="door", cost=1})
\endverbatim
*/

namespace
{
    class CoinSlot : public OnOffStone, public TimeHandler {
        CLONEOBJ(CoinSlot);
    public:
        CoinSlot();
    private:
        // Variables.
        enum State { ACTIVE, INACTIVE } state;
        double remaining_time;

        // Private methods.
        void init_model() {
            set_model(state==ACTIVE ? "st-coinslot" : "st-coinslot"); 
            // st-coinslot-filled
        }

        void actor_hit(const StoneContact &sc);
        const char *collision_sound() { return "st-metal"; }

        void animcb() { init_model(); }
        void tick(double dtime);
        void change_state(State newstate);
    };
}

CoinSlot::CoinSlot() 
    : OnOffStone("st-coinslot"), state(INACTIVE), remaining_time(0)
{
    set_attrib("cost", 1.0);
}

void
CoinSlot::change_state(State newstate)
{
    if (state == newstate) return;

    switch (newstate) {
    case ACTIVE:
        PerformAction(this, true);
        g_timer.activate(this);
        break;
    case INACTIVE:
        PerformAction(this, false);
        g_timer.deactivate(this);
        break;
    }
    state = newstate;
}

void
CoinSlot::tick(double dtime)
{
    assert(remaining_time > 0);
    
    remaining_time -= dtime;
    if (remaining_time <= 0)
    {
        change_state(INACTIVE);
    }
}

void
CoinSlot::actor_hit(const StoneContact &sc) 
{
    if (wielded_item_is(sc.actor, "it-coin")) 
    {
        if (player::Inventory *inv = get_inventory(sc.actor))
        {
	    play_sound("st-coinslot");
            Item *it = inv->yield_first();

            double coin_value = 0;
            it->double_attrib("value", &coin_value);
            remaining_time += coin_value;

            set_anim("st-coin2slot");
            change_state(ACTIVE);
            delete it;
        }
    }
}

//----------------------------------------------------------------------
// OBJECT REPOSITORY
//----------------------------------------------------------------------
namespace
{
    class ObjectRepos : public px::Nocopy {
    public:
        ObjectRepos();
        ~ObjectRepos();
        void add_templ(Object *o);
        bool has_templ(const string &name);
        Object *make(const string &name);
        Object *get_template(const string &name);
    private:
        typedef px::Dict<Object*> ObjectMap;
        ObjectMap objmap;           // repository of object templates
        int stonecount, floorcount, itemcount;
    };
}

ObjectRepos::ObjectRepos()
{
    stonecount=floorcount=itemcount=0;

    // Floors (most floors are defined in init.lua)
    add_templ(new Abyss);
    add_templ(new Water);
    add_templ(new Bridge);
    add_templ(new Gradient);

    // Stones
    add_templ(new BlackStone);
    add_templ(new BlackStone2);
    add_templ(new BlackStone3);
    add_templ(new BlackStone4);
    add_templ(new BolderStone);
    add_templ(new BombStone);
    add_templ(new BrickMagic);
    add_templ(new Brownie);
    add_templ(new ChameleonStone);
    add_templ(new CoinSlot);
    add_templ(new DeathStone);
    add_templ(new Door);
    add_templ(new Door_a);
    add_templ(new Door_b);
    add_templ(new Door_c);
    add_templ(new FakeOxydStone);
    add_templ(new FartStone);
    add_templ(new FloppyStone);
    add_templ(new GlassStone);
    add_templ(new Grate1);
    add_templ(new Grate2);
    add_templ(new HollowStoneImpulseStone);
    add_templ(new InvisibleMagic);
//    add_templ(new KeyStone);
    add_templ(new KeyStone_a);
    add_templ(new KeyStone_b);
    add_templ(new KeyStone_c);
    add_templ(new MagicStone);
    add_templ(new OneWayStone);
    add_templ(new OneWayStone_black);
    add_templ(new OneWayStone_white);    
    add_templ(new OxydStone);
    add_templ(new PuzzleStone);
    add_templ(new RubberBandStone);
    add_templ(new ShogunStone);
    add_templ(new ActorImpulseStone);
    add_templ(new StoneImpulseStone);
    add_templ(new SwapStone);
    add_templ(new SwitchStone);
    add_templ(new LaserSwitch);
    add_templ(new ScissorsStone);
    add_templ(new ThiefStone);
    add_templ(new TimerStone);
    add_templ(new WhiteStone);
    add_templ(new WhiteStone2);
    add_templ(new WhiteStone3);
    add_templ(new WhiteStone4);
    add_templ(new WoodenStone);
    add_templ(new WoodenStone_Growing);
    add_templ(new YinYangStone1);
    add_templ(new YinYangStone2);

    add_templ(new BlockStone);
    add_templ(new Stone_break);
    add_templ(new Greenbrown_hole);
    add_templ(new Greenbrown_move);
    add_templ(new Marble_hole);
    add_templ(new Marble_move);
    add_templ(new Rock3_hole);
    add_templ(new Rock3_move);
    add_templ(new Stonebrush);
    add_templ(new Break_invisible);
}

ObjectRepos::~ObjectRepos()
{
    px::delete_map(objmap.begin(), objmap.end());
}

// Add an new Object template to `objmap'. 
void 
ObjectRepos::add_templ(Object *o)
{
    string kind = o->get_kind();
    if (!has_templ(kind))
//         cerr << "add_templ: redefinition of object `" <<kind<< "'.\n";
//     else
        objmap.insert(kind, o);
}

bool 
ObjectRepos::has_templ(const string &name)
{
    return objmap.find(name) != objmap.end();
}

Object *
ObjectRepos::make(const string &name)
{
    ObjectMap::iterator i=objmap.find(name);
    if (i==objmap.end())
        return 0;
    else
        return i->second->clone();
}

Object *
ObjectRepos::get_template(const string &name)
{
    if (objmap.has_key(name))
        return objmap[name];
    else
        return 0;
}

//----------------------------------------------------------------------
// FUNCTIONS
//----------------------------------------------------------------------
namespace
{
    ObjectRepos *repos;
}

void 
world::Register(Object *obj)
{
    if (!repos)
        repos = new ObjectRepos;
    repos->add_templ(obj);
}

Object *
world::MakeObject(const char *kind)
{
    static Object *last_templ = 0;
    static string last_kind;

    if (last_kind!=kind) {
        last_kind = kind;
        last_templ = repos->get_template(kind);
    }

    Object *o = 0;
    if (last_templ)
        o=last_templ->clone();
    if (!o)
        fprintf(stderr, "MakeObject: unkown object name `%s'\n",kind);
    return o;
}

Object *
world::GetObjectTemplate(const std::string &kind)
{
    if (!repos->has_templ(kind)) {
        cerr << "GetObjectTemplate: unkown object name `" <<kind<< "'.\n";
        return 0;
    } else
        return repos->get_template(kind);
}


Floor*
world::MakeFloor(const char *kind)
{
    return dynamic_cast<Floor*>(MakeObject(kind));
}

Item *
world::MakeItem(const char *kind)
{
    return dynamic_cast<Item*>(MakeObject(kind));
}

Stone *
world::MakeStone (const char *kind)
{
    return dynamic_cast<Stone*>(MakeObject(kind));
}


void
world::DisposeObject(Object *o)
{ 
    if (o != 0) {
        UnnameObject(o);
        o->dispose();
    }
}

void 
world::SendMessage(Object *o, const std::string &msg)
{
    if (o)
        o->message(msg, Value());
}

void 
world::DefineSimpleStone(const std::string &kind, 
                         const std::string &sound)
{
    Register(new SimpleStone(kind,sound));
}

void 
world::DefineSimpleFloor(const std::string &kind,
                         double friction, double mousefactor)
{
    Register(new Floor(kind.c_str(), friction, mousefactor));
}
