/*
 * Copyright (C) 2002 Daniel Heck
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *  
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *  
 * You should have received a copy of the GNU General Public License along
 * with this program; if not, write to the Free Software Foundation, Inc.,
 * 59 Temple Place, Suite 330, Boston, MA 02111-1307, USA.
 */
#include "enigma.hh"
#include "lua.hh"
#include "options.hh"
#include "system.hh"
#include "px/dict.hh"

#include <stdio.h>
#include <stdlib.h>
#include <string>

using std::string;
using namespace options;

namespace options
{
    int    BitsPerPixel    = 16; 
    double MouseSpeed      = 5.0;
    double MouseDamping    = 25.0;
    double FrictionFactor  = 1.0;
    bool   WizardMode      = false;
    bool   FullScreen      = true;
    bool   ShowFPS         = false;
    bool   UseAlpha        = true;
    bool   Nograb          = false;

    double SoundVolume      = 1.0;
    double MusicVolume      = 1.0;
    double StereoSeparation = 10.0;
}

namespace
{
    px::Dict<LevelStatus> level_status_dict;
}

static string
levelkey(const string &levelpack, const string &levelname)
{
    return string("#")+levelname; //levelpack+"#"+levelname;
}

/* Get the status of a particular level in a particular pack.
   Returns 0 if no record for this level exists. */
LevelStatus *
options::GetLevelStatus(const string &levelpack, const string &levelname)
{
    px::Dict<LevelStatus>::iterator i;
    i = level_status_dict.find (levelkey (levelpack, levelname));
    return (i != level_status_dict.end()) ? &i->second : 0;
}

/* Set the status of a particular level.  The previous entry (if
   available) for this level is discarded. */
void
options::SetLevelStatus(const string &levelpack, const string &levelname,
                        const LevelStatus &stat)
{
    px::Dict<LevelStatus>::iterator i;
    string key = levelkey (levelpack, levelname);
    i = level_status_dict.find (key);
    if (i != level_status_dict.end())
        i->second = stat;
    else
        level_status_dict.insert(key,stat);
}

void
options::SetLevelTime(const string &pack, const string &level, int difficulty, int time)
{
    SetLevelFinished(pack, level, difficulty);
    LevelStatus *stat = GetLevelStatus(pack, level);
    assert(stat);
    if (difficulty == 1) {
        if (stat->par_easy > time)
            stat->par_easy = time;
    } else if (difficulty == 2) {
        if (stat->par_hard > time)
            stat->par_hard = time;
    }
}

void
options::SetLevelFinished(const string &pack, const string &level, int difficulty)
{
    assert(difficulty==1 || difficulty==2);
    if (LevelStatus *stat = GetLevelStatus(pack, level)) {
        stat->finished |= difficulty;
    } else
        SetLevelStatus(pack, level, LevelStatus(-1, -1, difficulty));
}

/* Determine name of the user's personal configuration file. */
static string
Personal_ConfigurationFileName()
{
    string fname = "enigmarc.lua";

    if (getenv ("HOME") != 0)
        fname = sysdep::expand_path ("~/.enigmarc");
    
    return fname;
}

static string
System_ConfigurationFileName()
{
    return enigma::FindDataFile ("enigma_conf.lua");
}

bool
options::Save ()
{
    string fname = Personal_ConfigurationFileName();
    FILE *fp = fopen(fname.c_str(), "wt");
    
    if (fp) {
        fprintf (fp, "options.MouseSpeed = %f\n",       MouseSpeed);
        fprintf (fp, "options.MouseDamping = %f\n",     MouseDamping);
        fprintf (fp, "options.FrictionFactor = %f\n",   FrictionFactor);
        fprintf (fp, "options.FullScreen = %d\n",       FullScreen);
        fprintf (fp, "options.UseAlpha = %d\n",         UseAlpha);
        fprintf (fp, "options.SoundVolume = %f\n",      SoundVolume);
        fprintf (fp, "options.MusicVolume = %f\n",      MusicVolume);
        fprintf (fp, "options.StereoSeparation = %f\n", StereoSeparation);

        // Save level information
        px::Dict<LevelStatus>::iterator i=level_status_dict.begin();
        for (; i!=level_status_dict.end(); ++i) {
            fprintf (fp, "options.LevelStat(\"%s\", {%d, %d, %d})\n",
                     i->first.c_str(),
                     i->second.par_easy, i->second.par_hard,
                     i->second.finished);
        }
        return fclose (fp);
    }
    return false;
}

static bool
load_from_file (const char *fname)
{
    switch (lua_dofile (lua::state, fname)) {
    case 0:                     // Everything ok
    case LUA_ERRFILE:           // File not found is not fatal
        return true;
    default:
        return false;           // All other errors are
    }
}

bool
options::Load ()
{
    string fname;
    bool error = false;

    fname = System_ConfigurationFileName();
    error |= load_from_file (fname.c_str());

    fname = Personal_ConfigurationFileName();
    error |= load_from_file (fname.c_str());

    return error;
}

