/*
 * Copyright (C) 2002,2003 Daniel Heck
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *  
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License along
 * with this program; if not, write to the Free Software Foundation, Inc.,
 * 59 Temple Place, Suite 330, Boston, MA 02111-1307, USA.
 *
 * $Id: oxyd.cc,v 1.5 2003/02/02 18:29:47 dheck Exp $
 */

/*
 * This files contains the interface between Enigma and the original
 * Oxyd games. Stuff like converting level descriptions or sound
 * effects is done here.
 */

#include "oxyd.hh"
#include "enigma.hh"
#include "sound.hh"

#include "oxydlib/DatFile.h"
#include "oxydlib/VecUtils.h"
#include "oxydlib/FileUtils.h"

#include <string>
#include <cstdio>
#include <cstdlib>
#include <cassert>
#include <iostream>

using namespace std;
using namespace enigma;
using namespace enigma::oxyd;

namespace
{
    struct GameInfo {
        GameInfo();
        GameInfo (const string &game_, const string &datfile_);

        // Variables.
        string      game;
        string      datfile;
        bool        is_present;
        string      datfile_path;
    };

    map <string,string> soundfx_map;
}

GameInfo::GameInfo()
    : is_present(false)
{}


GameInfo::GameInfo (const string &game_, const string &datfile_)
    : game(game_), datfile(datfile_), is_present(false)
{
    string fname;
    if (FindFile (datfile, datfile_path)) {
        enigma::Log << "Found " << game << " data file\n"; 
        is_present = true;
    }
}


//----------------------------------------
// local variables
//----------------------------------------
namespace
{
    vector<GameInfo>  games;
    DatFile          *datfile = 0;
}

void enigma::oxyd::Init()
{
    games.clear();
    games.resize(OxydVersion_Count);
    games[OxydVersion_PerOxyd]    = GameInfo("Per.Oxyd",    "peroxyd.dat");
    games[OxydVersion_Oxyd1]      = GameInfo("Oxyd 1",      "oxyd1ibm.dat");
    games[OxydVersion_OxydMagnum] = GameInfo("Oxyd magnum", "oxydmibm.dat");
    games[OxydVersion_OxydExtra]  = GameInfo("Oxyd extra",  "oxydex.dat");
}

bool 
enigma::oxyd::FoundOxyd (OxydVersion ver)
{
    return games[ver].is_present;
}

bool 
enigma::oxyd::OpenDatFile (OxydVersion ver)
{
    assert(datfile == 0);

    if (FoundOxyd(ver)) {
        string fname = games[ver].datfile_path;
        ByteVec data;
        readFile (fname, &data);

        datfile = new DatFile;
        string errmsg;
        if (!parseDatFile (data, ver, datfile, &errmsg)) {
            enigma::Log << "Error loading " << fname << ": " << errmsg << endl;
            delete datfile;
        } else {
            enigma::Log << "Loaded "<< fname << endl;
        }
    }
    return false;
}

void 
enigma::oxyd::UseOxydSounds (bool ok)
{
    if (!ok) {
        soundfx_map.clear();
        return;
    }

    soundfx_map["intro"]         = "OXINTRO.SDD";
    soundfx_map["exit"]          = "OXEXIT.SDD";
    soundfx_map["finished"]      = "OXFINITO.SDD";

    soundfx_map["boink"]         = "OXJUMP.SDD";
    soundfx_map["st-oxydopen"]   = "OXMEMOP.SDD";
    soundfx_map["st-oxydopened"] = "OXMEMOK.SDD";
    soundfx_map["st-oxydclose"]  = "OXMEMCL.SDD";
    soundfx_map["st-mirrorturn"] = "OXTURN.SDD";
    soundfx_map["st-coinslot"]   = "OXMONEY.SDD";
    soundfx_map["st-laseron"]    = "OXLASER.SDD";
    soundfx_map["dooropen"]      = "OXMOTOR.SDD";
    soundfx_map["doorclose"]     = "OXMOTOR.SDD";
    soundfx_map["fart"]          = "OXUNTITL.SDD";
    soundfx_map["pickup"]        = "OXINVENT.SDD";
    soundfx_map["invrotate"]     = "OXINVROT.SDD";
    soundfx_map["it-triggerdown"]= "OXSWON.SDD";
    soundfx_map["it-triggerup"]  = "OXSWOFF.SDD";
    soundfx_map["shatter"]       = "OXKLIRR.SDD";

    soundfx_map["explosion2"]    = "OXCRASH1.SDD";
    soundfx_map["explosion1"]    = "OXCRASH2.SDD";
    soundfx_map["impulse"]       = "OXWOUOU.SDD";
    soundfx_map["st-move"]       = "OXMOVE.SDD";
    soundfx_map["rotate-left"]   = "OXMAGIC4.SDD";
    soundfx_map["rotate-right"]  = "OXMAGIC3.SDD";
    soundfx_map["st-magic"]      = "OXMAGIC.SDD";
    soundfx_map["puller"]        = "OXPULLER.SDD";

    soundfx_map["st-stone"]      = "OXKLICK1.SDD";
    soundfx_map["st-metal"]      = "OXKLICK2.SDD";
    soundfx_map["st-thud"]       = "OXKLICK4.SDD";
    soundfx_map["st-thud"]       = "OXKLICK5.SDD";
    soundfx_map["electric"]      = "OXKLICK6.SDD";
    soundfx_map["drown"]         = "OXBLOOP.SDD";
    soundfx_map["thief"]         = "OXTHIEF.SDD";
    soundfx_map["warp"]          = "OXTRANS.SDD";
    soundfx_map["crack"]         = "OXCRACK.SDD";
    soundfx_map["squish"]        = "OXMATSCH.SDD";
    soundfx_map["snip"]           = "OXCUT.SDD";

    // ???
    soundfx_map["klick3"]        = "OXKLICK3.SDD";
    soundfx_map["drop"]          = "OXDROP.SDD";
    soundfx_map["boing"]         = "OXBOING.SDD";
    soundfx_map["bold"]          = "OXBOLD.SDD";
    soundfx_map["magic2"]        = "OXMAGIC2.SDD";
    soundfx_map["magic3"]        = "OXMAGIC3.SDD";
}


Mix_Chunk *
enigma::oxyd::LoadSound (const std::string &name)
{
    Mix_Chunk *chunk = 0;
    if (datfile) {
        string chunkname = soundfx_map[name];
        const ByteVec *snddata = datfile->getChunk(chunkname);
        if (snddata) {
            enigma::Log << "Loaded sound file " << name << " =^= " << chunkname<<endl;
            const int offset = 16;
            return sound::ChunkFromRaw (&(*snddata)[0], snddata->size()-offset,
                                        6000, AUDIO_S8, 1);
        }
    }
    return chunk;
}
