/*
 * Copyright (C) 2002,2003 Daniel Heck
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *  
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License along
 * with this program; if not, write to the Free Software Foundation, Inc.,
 * 59 Temple Place, Suite 330, Boston, MA 02111-1307, USA.
 *
 * $Id: player.cc,v 1.9 2003/02/14 19:04:49 dheck Exp $
 */
#include "player.hh"
#include "display.hh"
#include "game.hh"
#include "sound.hh"
#include "objects.hh"
#include "actors.hh"
#include "px/tools.hh"
#include <algorithm>
#include <iostream>
#include <cassert>

using namespace player;
using namespace std;
using namespace enigma;

namespace
{
    class PlayerInfo {
    public:
        PlayerInfo();

        string          name;
        Inventory       inventory;
        vector<Actor*>  actors;
        bool            out_of_lives;
        bool            respawn_shortly;
        double          dead_dtime; // number of seconds the player is
                                    // already dead
    };

    typedef vector<PlayerInfo> PlayerList;

    struct RespawnInfo {
        RespawnInfo (Actor *a, double t) 
        : actor(a), time_left(t) 
        {}

        Actor *actor;           // The actor to respawn
        double time_left;       // Time left before respawning
    };


    struct LevelLocalData {
        // Functions
        void respawn_dead_actors(double dtime);
        void resurrect_actor (Actor *a);
        void remove_extralife (Actor *a);

        // Variables
        vector<RespawnInfo>  respawn_list;
    } leveldat;
}



PlayerInfo::PlayerInfo () 
: out_of_lives(false)
, respawn_shortly (false)
, dead_dtime(0)
{}



//----------------------------------------
// Local variables
//----------------------------------------

namespace
{
    PlayerList players;
    unsigned   icurrent_player               = 0;
    bool       current_player_inhibit_pickup = false;
  
}

void
LevelLocalData::respawn_dead_actors(double dtime)
{
    for (unsigned i=0; i<respawn_list.size(); ) {
        RespawnInfo &info = respawn_list[i];

        info.time_left -= dtime;
        if (info.time_left < 0) {
            remove_extralife(info.actor);
            info.actor->respawn();
            respawn_list.erase(respawn_list.begin()+i);
            continue;           // don't increment i
        }
        ++i;
    }
}

void
LevelLocalData::resurrect_actor (Actor *a)
{
    const double RESPAWN_TIME = 1.5;

    SendMessage(a, "resurrect");
    respawn_list.push_back(RespawnInfo(a, RESPAWN_TIME));
}

void 
LevelLocalData::remove_extralife (Actor *a)
{
    int iplayer = a->int_attrib("player");
    Inventory *inv = player::GetInventory(iplayer);
    for (int i=0; i<inv->size(); ++i)
        if (0 == strcmp(inv->get_item(i)->get_kind(), "it-extralife")) {
            delete inv->yield_item(i);
            break;
        }
}


//----------------------------------------
// Inventory impl.
//----------------------------------------
unsigned const Inventory::max_items = 12;

Inventory::Inventory()
    : items()
{}

Inventory::~Inventory() 
{ 
    clear();
}

void
Inventory::clear()
{
    px::delete_sequence(items.begin(), items.end()); 
    items.clear();
}

Item *
Inventory::get_item(int idx)
{
    return idx >= size() ? 0 : items[idx];
}

Item *
Inventory::yield_item(int idx)
{
    if (0 <= idx && idx < size())
    {
        Item *it = items[idx];
        items.erase(items.begin()+ idx);
        redraw();
        return it;
    }
    return 0;
}

void 
Inventory::add_item(Item *i)
{
    items.insert(items.begin(), i);
    redraw();
}

void 
Inventory::redraw()
{
    display::GetStatusBar()->update_inventory(this);
}

void 
Inventory::rotate()
{
    if (!items.empty())
    {
        std::rotate(items.begin(), items.begin()+1, items.end());
        redraw();
    }
}

void 
Inventory::activate_first()
{
    if (!items.empty())
    {
        Item *it = items[0];
        Actor *ac = players[icurrent_player].actors[0];
        const world::ActorInfo &ai = world::GetActorInfo(ac);

        GridPos p(static_cast<int>(ai.pos[0]), 
                  static_cast<int>(ai.pos[1]));

        switch (it->activate(ac, p)) {
        case world::ITEM_DROP:
            // only drop if no item underneath and actor allows it
            if (world::GetItem(p) == 0 && ac->can_drop_items()) {
                items.erase(items.begin());
                redraw();
                world::SetItem(p, it);
                it->on_drop(ac);
            }
            break;
        case world::ITEM_KILL:
            delete it;
            items.erase(items.begin());
            redraw();
            break;
        case world::ITEM_KEEP:
            break;
        }
    }
}

Item * 
Inventory::yield_first()
{
    if (!items.empty())
    {
        Item *it = items[0];
        items.erase(items.begin());
        redraw();
        return it;
    }
    return 0;
}


//----------------------------------------
// Global functions
//----------------------------------------

void 
player::NewGame(int nplayers)
{
    assert(nplayers > 0);

    players.clear();
    players.resize(nplayers);

    for (int i=0; i<nplayers; ++i)
    {
        Inventory *inv = GetInventory(i);
        inv->add_item(world::MakeItem("it-extralife"));
        inv->add_item(world::MakeItem("it-extralife"));
    }
}

void
player::NewWorld()
{
    // Clear up the inventories of all players: keep only extra lifes.
    for (unsigned iplayer=0; iplayer<players.size(); ++iplayer)
    {
        Inventory *inv = GetInventory(iplayer);
        int nextralifes=0;
        for (int i=0; i<inv->size(); ++i)
            if (0 == strcmp(inv->get_item(i)->get_kind(), "it-extralife"))
                nextralifes += 1;
        inv->clear();
        for (int i=0; i<nextralifes; ++i)
            inv->add_item(world::MakeItem("it-extralife"));
        players[iplayer].actors.clear();
    }
    SetCurrentPlayer(0);
    leveldat = LevelLocalData();
}

void
player::LevelFinished()
{
    for (unsigned i=0; i<players.size(); ++i) {
        for (unsigned j=0; j<players[i].actors.size(); ++j) {
            Actor *a = players[i].actors[j];
            SendMessage(a, "disappear");
        }
    }
}

Inventory *
player::GetInventory(int iplayer)
{
    return &players[iplayer].inventory;
}

int
player::CurrentPlayer()
{
    return icurrent_player;
}

void
player::SetCurrentPlayer(unsigned iplayer)
{
    if (iplayer >=players.size())
        fprintf(stderr, "SetCurrentPlayer: no such player %d\n", iplayer);
    else {
        icurrent_player = iplayer;
        display::GetStatusBar()->set_inventory(GetInventory(iplayer));
        if (!players[iplayer].actors.empty())
            if (Actor *a = players[iplayer].actors[0])
                display::FollowSprite(a->get_spriteid());
    }
}

void 
player::Suicide()
{
    for (unsigned i=0; i<players.size(); ++i) {
        vector<Actor *> &al = players[icurrent_player].actors;
        for (unsigned i=0; i<al.size(); ++i) {
            SendMessage(al[i], "shatter");
        }
    }
}

void 
player::AddActor(int iplayer, Actor *a)
{
    assert(iplayer >= 0 && (unsigned)iplayer < players.size());
    players[iplayer].actors.push_back(a);
    if (iplayer==icurrent_player)
        display::FollowSprite(a->get_spriteid());
}

void 
player::SwapPlayers()
{
    if (players.size() >= 2) {
        SetCurrentPlayer(icurrent_player==0 ? 1:0);
    } else
        fprintf(stderr, "No two players to swap!\n");
}


/* Try to resurrect actor `a' beloning to player `pl'.  Return true if
   this was possible. */
static bool
resurrect_actor (unsigned pl, Actor *a)
{
    if (enigma::ConserveLevel) {
        Inventory *inv = &players[pl].inventory;

//    return enigma::ConserveLevel && respawn_player(iplayer,a))
        for (int i=0; i<inv->size(); ++i)
            if (0 == strcmp(inv->get_item(i)->get_kind(), "it-extralife")) {
                leveldat.resurrect_actor(a);
                return true;
            }
    } else {
        
    }
    return false;
}


/* 
** Returns true if all actors are dead and they cannot be respawned
** (i.e., their owners don't have any extra lifes left). 
*/
static bool 
AllPlayersDead()
{
    bool all_players_dead=true;
    unsigned old_player = icurrent_player;
    for (unsigned i=0; i<players.size(); ++i) {
        bool actors_dead=true;
        for (unsigned j=0; j<players[i].actors.size(); ++j) {
            Actor *a = players[i].actors[j];
            if (!a->is_dead() || resurrect_actor(i,a))
                all_players_dead = actors_dead = false;
        }
        if (actors_dead && i==icurrent_player) {
            // Switch to next player if this one is dead
            unsigned nextplayer = (icurrent_player+1) % players.size();
            SetCurrentPlayer (nextplayer);
        }
    }
    if (all_players_dead)
        SetCurrentPlayer(old_player);
    return all_players_dead;
}

void
player::Tick(double dtime)
{
    // Update position of active actor for stereo sound
    Actor *ac = players[icurrent_player].actors[0];
    const world::ActorInfo &ai = world::GetActorInfo(ac);
    sound::SetListenerPosition (px::V2(ai.pos[0], ai.pos[1]));

    // Respawn actors that have been dead for a certain amount of time
    leveldat.respawn_dead_actors(dtime);

    // Update the respawn list or restart the game when all actors are
    // dead and no extra lifes are left.
    if (AllPlayersDead()) {
        enigma::RestartGame();
    }
}

void 
player::InhibitPickup(bool flag)
{
    current_player_inhibit_pickup = flag;
}

void 
player::PickupItem(Actor *a, GridPos p)
{
    int iplayer=-1;
    a->int_attrib("player", &iplayer);
    if (iplayer < 0 || (unsigned)iplayer >= players.size()) {
//        cerr << "PickupItem: illegal 'player' entry\n";
        return;
    }

    if (iplayer==icurrent_player && current_player_inhibit_pickup)
    {
        // do not pick up items if current player doesn't want to
        return;
    }

    if (a->is_flying())
    {
        // do not pick up items while flying
        return;
    }

    Inventory &inv = players[iplayer].inventory;
    if (!inv.is_full())
    {
        Item *item = world::YieldItem(p);
        inv.add_item(item);
	sound::PlaySound("pickup");
    }
}

void 
player::ActivateItem()
{
    Inventory &inv = players[icurrent_player].inventory;
    inv.activate_first();
}

void 
player::RotateInventory()
{
    sound::PlaySound("invrotate");
    Inventory &inv = players[icurrent_player].inventory;
    inv.rotate();
}
