/*
 * Copyright (C) 2002,2003 Daniel Heck
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *  
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License along
 * with this program; if not, write to the Free Software Foundation, Inc.,
 * 59 Temple Place, Suite 330, Boston, MA 02111-1307, USA.
 *
 * $Id: video.cc,v 1.4 2003/01/12 19:37:19 dheck Exp $
 */
#include "video.hh"
#include "error.hh"
#include "geom.hh"
#include "tools.hh"
#include "sdl.hh"

#include "SDL_image.h"
#include "SDL_syswm.h"

#include <cassert>
#include <string>
#include <memory>
#include <cstdio>

using namespace std;
using namespace px;


class FBDrawable : public px::Drawable {
public:
    FBDrawable (FrameBuffer *framebuf) : fb(framebuf) {}

    // Drawable interface.
    Uint32 map_color(int r, int g, int b) { return fb->map_color(r,g,b); }
    Uint32 map_color(int r, int g, int b, int a) { return fb->map_color(r,g,b,a); }

    Uint32 get_pixel(int x, int y) { return 0; }
    
    Rect size() const { return fb->size(); }

    void blit(int x, int y, Surface* src, const Rect& sr) {fb->blit(x,y,src,sr);}
    void blit(int x, int y, Surface* src) {fb->blit(x,y,src); }
    
    void box(int x, int y, int w, int h, Uint32 color)
    {
        SDL_Rect dr;
        dr.x = x; dr.y = y; dr.w = w; dr.h = h;
        SDL_FillRect(fb->get_surface(), &dr, color);
    }

protected:
    FrameBuffer *fb;
};

/*
 * Generic "Drawable" implementation for framebuffers.
 */
template <class PIXELT>
class FBDrawable_Gen : public FBDrawable {
public:
    FBDrawable_Gen (FrameBuffer *framebuf) : FBDrawable(framebuf) {}

    PIXELT *pixel_pointer(int x, int y)
    {
        return static_cast<PIXELT*>(fb->pixel_pointer(x,y));
    }

    void set_pixel(int x, int y, Uint32 color) 
    {
        *pixel_pointer(x, y) = color;
    }

    void set_pixels(int n, const int* xlist, const int* ylist, Uint32 color)
    {
        const int *xp = xlist, *yp = ylist;
        for (int i=n; i > 0; --i) 
        {
            int x = *xp++, y = *yp++;
            *pixel_pointer(x,y) = color;
        }
    }

    void hline(int x, int y, int w, Uint32 color) 
    {
        PIXELT* dst = pixel_pointer(x, y);
        for (; w > 0; --w)
            *dst++ = color;
    }
    
    void vline(int x, int y, int h, Uint32 color) 
    {
        PIXELT* dst = pixel_pointer(x, y);
        int off = fb->pitch()/fb->bypp();
        for (; h > 0; --h) 
        {
            *dst = color;
            dst += off;
        }
    }
};





class FBDrawable_24 : public FBDrawable {
public:
    FBDrawable_24 (FrameBuffer *framebuf) : FBDrawable(framebuf) {}

    void set_pixel(int x, int y, Uint32 color) {
        Uint8* p = static_cast<Uint8*>(fb->pixel_pointer(x, y));
        SDL_GetRGB(color, fb->get_surface()->format, p,p+1,p+2);
    }

    Uint32 get_pixel(int x, int y) {
        Uint8* p = static_cast<Uint8*>(fb->pixel_pointer(x, y));
        return SDL_MapRGB(fb->get_surface()->format, p[0],p[1],p[2]);
    }
};


//----------------------------------------
// FrameBuffer implementation.
//----------------------------------------

void FrameBuffer::lock() 
{
    if (SDL_MUSTLOCK(m_surface))
        SDL_LockSurface(m_surface);
}

void FrameBuffer::unlock() 
{
    if (SDL_MUSTLOCK(m_surface))
        SDL_UnlockSurface(m_surface);
}

void 
FrameBuffer::blit(int x, int y, Surface* src, const Rect& sr) 
{
    SDL_Rect r1;
    SDL_Rect r2;
    r1.x = sr.x; r1.y = sr.y;
    r1.w = sr.w; r1.h = sr.h;
    
    r2.x = x; r2.y = y;
    SDL_BlitSurface(src->m_surface, &r1, m_surface, &r2);
}

void 
FrameBuffer::blit(int x, int y, Surface* src) 
{
    SDL_Rect dr;
    dr.x = x; dr.y = y;
    SDL_BlitSurface(src->m_surface, 0, m_surface, &dr);
}



Uint32 
FrameBuffer::map_color(int r, int g, int b)
{
    return SDL_MapRGB(m_surface->format, r, g, b);
}

Uint32 
FrameBuffer::map_color(int r, int g, int b, int a)
{
    return SDL_MapRGBA(m_surface->format, r, g, b, a);
}

//----------------------------------------
// Surface implementation
//----------------------------------------
Surface::Surface (SDL_Surface* sfc) : FrameBuffer(sfc)
{
    switch (bipp()) {
    case 8: drawable = new FBDrawable_Gen<Uint8> (this); break;
    case 16:drawable = new FBDrawable_Gen<Uint16>(this); break;
    case 32:drawable = new FBDrawable_Gen<Uint32>(this); break;
    case 24:drawable = new FBDrawable_24(this); break;
    default:
        break;
//         throw XVideo("Invalid bit depth in surface.");
    }
}

Surface::~Surface()
{
    delete drawable;
}

void 
Surface::set_color_key (int r, int g, int b) 
{
    Uint32 color = map_color(r,g,b);
    SDL_SetColorKey(get_surface(), SDL_SRCCOLORKEY | SDL_RLEACCEL, color); 
}

void
Surface::set_alpha(int a)
{
    SDL_SetAlpha(get_surface(), SDL_SRCALPHA, a);
}

//----------------------------------------
// Screen implementation
//----------------------------------------

/* `Xlib.h' also defines a type named `Screen' so we have to specify
   the namespace explicitly and cannot simple use a using-declaration. */

px::Screen::Screen(SDL_Surface *s) 
    : Surface(s), m_dirtyrects(), update_all_p(false)
{}

void 
px::Screen::update_all() 
{
    //SDL_UpdateRect(get_surface(), 0, 0, 0, 0); 
    update_all_p = true;
}
    
void 
px::Screen::update_rect(const Rect& r) 
{
    if (m_dirtyrects.size() < 200)
        m_dirtyrects.push_back(r);
    else
        update_all();
}
    
void 
px::Screen::set_caption(const char* str) 
{
    SDL_WM_SetCaption(str, 0);
}
    
void 
px::Screen::flush_updates() 
{
    if (update_all_p) {
        SDL_UpdateRect(get_surface(), 0, 0, 0, 0);	// update everything
        update_all_p=false;
    } 
    else if (!m_dirtyrects.empty()) {
        m_dirtyrects.intersect(Rect(0, 0, width(), height()));

        vector<SDL_Rect> rects(m_dirtyrects.size());

        RectList::iterator j=m_dirtyrects.begin();;
        for (unsigned i=0; i<rects.size(); ++i, ++j)
            sdl::copy_rect (rects[i], *j);
        SDL_UpdateRects(get_surface(), rects.size(), &rects[0]);
    } 
    m_dirtyrects.clear();
}

//----------------------------------------
// Functions
//----------------------------------------

/* Convert the surface to the native surface format.  A pointer to a
   new surface is returned; the old one must be deleted by hand if it
   is no longer needed.  */
Surface*
px::DisplayFormat(Surface* s) 
{
    if (SDL_Surface* s2 = SDL_DisplayFormat(s->get_surface()))
        return new Surface(s2);
    return 0;
}


px::Screen *
px::OpenScreen(int w, int h, int bipp)
{
    SDL_Surface* sfc = SDL_SetVideoMode(w, h, bipp, SDL_SWSURFACE);
    return new Screen(sfc);
}

Surface *
px::Duplicate(Surface *s)
{
    if (s==0) return 0;
    SDL_Surface *sdls = s->get_surface();
    SDL_Surface *copy = SDL_ConvertSurface(sdls, sdls->format, SDL_SWSURFACE);
    if (sdls->format->palette != 0)
        SDL_SetColors(copy, sdls->format->palette->colors, 0, 
                      sdls->format->palette->ncolors);
    return new Surface(copy);
}

void
px::TintRect(Surface *s, Rect rect, Uint8 r, Uint8 g, Uint8 b, Uint8 a)
{
    std::auto_ptr<Surface> copy(Grab(s, rect));
    if (copy.get()) {
        copy->set_alpha(a);

        GC gc(s->get_drawable());
        set_color(gc, r,g,b);
        box(gc, rect);
        blit(gc, rect.x, rect.y, copy.get());
    }
}


Surface *
px::Grab(Surface *s, Rect r)
{
    if (s==0) return 0;
    SDL_Surface *sdls = s->get_surface();
    SDL_Surface *copy = SDL_CreateRGBSurface (SDL_SWSURFACE, r.w, r.h,
                                              sdls->format->BitsPerPixel,
                                              sdls->format->Rmask,
                                              sdls->format->Gmask,
                                              sdls->format->Bmask,
                                              sdls->format->Amask);
    if (sdls->format->palette != 0)
        SDL_SetColors(copy, sdls->format->palette->colors, 0, 
                      sdls->format->palette->ncolors);

    Surface *new_s = new Surface(copy);
    new_s->blit(0,0, s, r);
    return new_s;
}

#undef LoadImage
Surface* 
px::LoadImage (const char* filename)
{
    if (SDL_Surface *tmp = IMG_Load(filename))
    {
        SDL_Surface* img;

        if (tmp->flags & SDL_SRCALPHA) {
            SDL_SetAlpha(tmp, SDL_RLEACCEL, 0);
            img = SDL_DisplayFormatAlpha(tmp);
        }
        else if (tmp->flags & SDL_SRCCOLORKEY) {
            SDL_SetColorKey(tmp, SDL_SRCCOLORKEY | SDL_RLEACCEL,
                            tmp->format->colorkey);
            img = SDL_DisplayFormat(tmp);
        }
        else
            img = SDL_DisplayFormat(tmp);
        if (img == 0)
            return new Surface(tmp);
        SDL_FreeSurface(tmp);
        return new Surface(img);
    }
    return 0;
}

Surface *
px::MakeSurface(int w, int h, int bipp, const RGBA_Mask &mask)
{
    SDL_Surface* sfc;
    sfc = SDL_CreateRGBSurface(SDL_SWSURFACE, w, h, bipp,
                               mask.r, mask.g, mask.g, mask.a);
    if (sfc == 0)
        return 0; //throw XVideo(string("Couldn't create surface: ") + SDL_GetError());
    return new Surface(sfc);
}


/* Create a surface from image data that is already somewhere in
   memory.  */
Surface *
px::MakeSurface(void* data, int w, int h, int bipp, int pitch,
                const RGBA_Mask &mask)
{
    SDL_Surface* sfc;
    sfc = SDL_CreateRGBSurfaceFrom(SDL_SWSURFACE, w, h, bipp,
                                   pitch, mask.r, mask.g, mask.b, mask.a);
    if (sfc == 0)
        return 0; //throw XVideo(string("Couldn't create surface:") + SDL_GetError());
    return new Surface(sfc);
}	

