/*
 * Copyright (C) 2002,2003 Daniel Heck
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *  
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License along
 * with this program; if not, write to the Free Software Foundation, Inc.,
 * 59 Temple Place, Suite 330, Boston, MA 02111-1307, USA.
 *
 * $Id: world.cc,v 1.12 2003/02/14 19:04:48 dheck Exp $
 */
#include "actors.hh"
#include "world.hh"
#include "objects.hh"
#include "items.hh"
#include "laser.hh"
#include "display.hh"
#include "player.hh"
#include "sound.hh"
#include "px/px.hh"
#include "px/dict.hh"
#include "lua.hh"
#include "options.hh"

#include <iostream>
#include <algorithm>
#include <functional>
#include <numeric>
#include <cassert>

using namespace std;
using namespace world;
using namespace px;

namespace test
{
    enum ShapeType
    {
        SHAPE_Sphere,
        SHAPE_Square,
        SHAPE_Rectangle,
        SHAPE_Polygon,
    };

    class Shape
    {
        ShapeType type;
    public:
        Shape (ShapeType t) : type(t) {}
        virtual ~Shape() {}

        ShapeType get_type() const { return type; }
    };

    struct Sphere : public Shape
    {
        Sphere (const V2 &c, double r) 
            : Shape(SHAPE_Sphere), center(c), radius(r)
        {}

        // Variables.
        V2 center;
        double radius;
    };
}

//----------------------------------------
// Sphere
//----------------------------------------
namespace
{
    class Sphere {
        V3 center;
        double radius;
    public:
        Sphere(const V3 &c, double r)
            : center(c), radius(r)
        {}
        bool contains(const V3 &p) const {
            return length(center-p) < radius;
        }
        bool intersection_point(const Sphere &s, V3 &ip) {
            double rr = radius + s.radius;
            ip = center + normalize(s.center-center)*rr/2;
            return (length(center - s.center) <= rr);
        }
    };
}

//----------------------------------------
// Rubber band
//----------------------------------------
namespace world
{
    class RubberBand {
    public:
        RubberBand (Actor *a1, Actor *a2, double strength=10, double length=2);
        RubberBand (Actor *a1, Stone *st, double strength=10, double length=2);
        ~RubberBand();

        void tick(double);
        V3 get_force(Actor *a);

        Actor *get_actor() const { return actor; }
        Actor *get_actor2() const { return actor2; }
        Stone *get_stone() const { return stone; }
    private:
        V2 get_p1() const;
        V2 get_p2() const;

        // Variables.
        Actor *actor, *actor2;
        Stone *stone;
        display::RubberID model;
        double strength, length;
    };
}

RubberBand::RubberBand (Actor *a1, Actor *a2, double strength_, double length_)
: actor(a1), actor2(a2), stone(0), model(0)
, strength(strength_), length(length_)
{
    assert(actor);
    assert(length >= 0);
    model = display::AddRubber(get_p1(),get_p2());
}

RubberBand::RubberBand (Actor *a1, Stone *st, double strength_, double length_)
: actor(a1), actor2(0), stone(st), model(0)
, strength(strength_), length(length_)
{
    assert(actor);
    assert(length >= 0);
    model = display::AddRubber(get_p1(), get_p2());
}

RubberBand::~RubberBand() {
    model.kill();
}

void 
RubberBand::tick(double)
{
    V2 v = get_p2()-get_p1();
    double vv = px::length(v);

    if (vv > length) {
        V2 force = v * strength*(vv-length)/vv;
        V3 f(force[0],force[1],0);
        actor->add_force(f);
        if (actor2)
            actor2->add_force(-f);
    }

    model.update_first (get_p1());
    model.update_second (get_p2());
}

V3 
RubberBand::get_force(Actor *a)
{
    return V3();
}

V2 
RubberBand::get_p1() const
{
    return V2(actor->get_pos()[0], actor->get_pos()[1]);
}

V2 
RubberBand::get_p2() const
{
    if (!stone)
        return V2(actor2->get_pos()[0], actor2->get_pos()[1]);
    else
        return V2(stone->get_pos().x+0.5, stone->get_pos().y+0.5);
}



//----------------------------------------
// Field
//----------------------------------------
namespace
{
    class Field {
    public:
	Field();
	~Field();

        Floor *floor;
        Item *item;
        Stone *stone;
    };
}


Field::Field()
{
    floor=0;
    item=0;
    stone=0;
}

Field::~Field()
{
    DisposeObject(floor);
    DisposeObject(item);
    DisposeObject(stone);
}

/*
** This class implements the "force field" that accelerates objects
** when the mouse is moved.  Only objects that have the "mouseforce"
** attribute set are affected by this force field.
*/
namespace
{
    class MouseForce : public ForceField {
    public:
 	void set_force(V3 f) { force=f; }
 	void add_force(V3 f) { force+=f; }
	
	V3 get_force(Actor *a, V3 x, V3 v, double time) 
        {
            int iplayer=0;
            if (a->int_attrib("player", &iplayer) && iplayer != player::CurrentPlayer())
                return V3();

            if (a->is_flying() || a->is_dead() )
                return V3();

            if (a->int_attrib("mouseforce"))
                return force;
            else
                return V3();
	}

	void tick(double dtime) {
            force=V3();
	}
    private:
	V3 force;
    };
}

namespace
{
    typedef px::Array2<Field> FieldArray;
    typedef std::vector<Actor*> ActorList;

    typedef vector<ForceField*> ForceList;
    typedef vector<StoneContact> StoneContactList;
}

StoneContact::StoneContact(Actor *a, GridPos p,
                           const V3 &cp, const V3 &n)
: actor(a), stonepos(p) 
, response(STONE_REBOUND) 
, contact_point(cp), normal(n) 
, is_collision(false) 
, fake_collision(false)
, is_contact(true)
{}

StoneContact::StoneContact()
: is_collision(false)
, fake_collision(false)
, is_contact(false)
{}

DirectionBits 
world::contact_faces(const StoneContact &sc)
{
    using namespace enigma;

    int dirs = NODIRBIT;

    if (sc.normal[0] < 0) 
        dirs |= WESTBIT;
    else if (sc.normal[0] > 0) 
        dirs |= EASTBIT;
    if (sc.normal[1] < 0) 
        dirs |= NORTHBIT;
    else if (sc.normal[1] > 0)
        dirs |= SOUTHBIT;

    return DirectionBits(dirs);
}

Direction
world::contact_face(const StoneContact &sc)
{
    using namespace enigma;
    if (sc.normal == V3(-1,0,0)) 
        return WEST;
    else if (sc.normal == V3(1,0,0)) 
        return EAST;
    else if (sc.normal == V3(0,-1,0))
        return NORTH;
    else if (sc.normal == V3(0,1,0))
        return SOUTH;
    else 
        return NODIR;
}

//======================================================================
// VARIABLES
//======================================================================

/* Objects can be named with NameObject and referenced with
   GetNamedObject.  This dictionary maps the names to the actual
   objects. */
static px::Dict<Object *> named_objects;

namespace
{
    MouseForce *mouseforce;

    class Level {
    public:
        Level(int ww, int hh) : fields(ww,hh) {
            w = ww;
            h = hh;
        }
        ~Level() {
            fields = FieldArray(0,0);
            for_each(actorlist.begin(), actorlist.end(), mem_fun(&Actor::dispose));
            delete_sequence (rubbers.begin(), rubbers.end());
        }

        bool contains (GridPos p) {
            return (p.x>=0 && p.y>=0 && p.x<w && p.y<h);
        }

        Field *get_field (GridPos p) {
            if (this->contains(p))
                return &fields(p.x,p.y);
            else
                return 0;
        }


        // Variables.
        FieldArray fields;      // Contains floors, items, etc.
        int        w, h;   	// Width and height of the level 
        ForceList  forces;
        ActorList  actorlist;   // List of movable, dynamic objects
        vector<RubberBand *> rubbers;
    }; 

    Level *level;
}

enigma::Timer world::g_timer;


//======================================================================
// PHYSICS SIMULATION
//======================================================================

const double MINVEL = 0.0001;

/* Calculate the total acceleration on an actor A at time TIME.  The
   actor's current position X and velocity V are also passed.  Note
   that the position and velocity entries in ActorInfo will be updated
   only after a *successful* time step, so they cannot be used
   here. */
static V3
get_accel (Actor *a, const V3 & x, const V3 & v, double time)
{
    V3 f;
    GridPos p((int)x[0], (int)x[1]);

    if (a->is_on_floor())
    {
        if (Floor *floor = GetFloor(p))
        {
            // Mouse force
            f += mouseforce->get_force(a,x,v,time)*floor->mousefactor();

            // Friction
            double vv=length(v);
            if (vv > 0) {
                V3 frictionf = v * options::FrictionFactor*floor->friction();
		frictionf /= vv;
                frictionf *= pow(vv, 0.8);
                f -= frictionf;
            }
            f += floor->get_force(a);
        }

        if (Item *item = GetItem(p))
        {
            f += item->get_force(a);
        }
    }

    // all other force fields
    for (ForceList::iterator i=level->forces.begin(); 
         i != level->forces.end(); ++i)
    {
        f += (*i)->get_force(a, x, v, time);
    }

    ActorInfo *ai = a->get_actorinfo();
    f += ai->forceacc2;

    return f / ai->mass;
}


/* This function performs one step in the numerical integration of an
   actor's equation of motion.  TIME ist the current absolute time and
   H the size of the integration step. */
static void
advance_actor(Actor *a, double time, double h)
{
    const double MAXVEL = 40;
    
    ActorInfo &ai = *a->get_actorinfo();
    V3 accel = get_accel(a, ai.pos, ai.vel, time);
    
    ai.pos += h*ai.vel;
    ai.vel += h*accel;

    // Limit maximum velocity
    double q = length(ai.vel) / MAXVEL;
    if (q > 1)
        ai.vel /= q;
}

/* If the actor and the stone are not in contact, `contact' is filled
   with information about the closest feature on the stone and
   `is_contact' is set to false. */
static void
find_contact_with_stone(Actor *a, StoneContact &contact, int x, int y)
{
    contact.actor = a;
    contact.stonepos = GridPos(x, y);
    contact.is_contact = false;

    Stone *stone = world::GetStone(contact.stonepos);
    if (!stone)
        return;
    
    const ActorInfo &ai = *a->get_actorinfo();
    double r = ai.radius;
    double ax = ai.pos[0];
    double ay = ai.pos[1];
    const double contact_e = 0.01;
    const double erad = 2.0/32; // edge radius

    // Closest feature == north or south face of the stone?
    if (ax>x+erad && ax<x+1-erad) {
        double dist = r+5;

        // south
        if (ay>y+1) {
            contact.contact_point = V3(ax, y+1,0);
            contact.normal        = V3(0,+1,0);
            dist                  = ay-(y+1);
        }
        // north
        else if (ay<y) {
            contact.contact_point = V3(ax, y,0);
            contact.normal        = V3(0,-1,0);
            dist                  = y-ay;
        }
	contact.is_contact = (dist-r < contact_e);
    }
    // Closest feature == west or east face of the stone?
    else if (ay>y+erad && ay<y+1-erad) {
        double dist=r+5;
        if (ax>x+1) { // east
            contact.contact_point = V3(x+1,ay,0);
            contact.normal        = V3(+1,0,0);
            dist                  = ax-(x+1);
        }
        else if (ax<x) { // west
            contact.contact_point = V3(x,ay,0);
            contact.normal        = V3(-1,0,0);
            dist                  = x-ax;
        }
	contact.is_contact = (dist-r < contact_e);
    }
    // Closest feature == any of the four corners
    else {
        int xcorner=(ax >= x+1-erad);
        int ycorner=(ay >= y+1-erad);
        double cx[2] = {erad, -erad};

        V3 corner(x+xcorner+cx[xcorner], y+ycorner+cx[ycorner], 0);
        V3 b=V3(ax,ay,0) - corner;

        contact.is_contact    = (length(b)-r-erad < contact_e);
        contact.normal        = normalize(b);
        contact.contact_point = corner + contact.normal*erad;
    }
    contact.is_collision  = contact.normal*ai.vel < 0;
    contact.response = stone->collision_response(contact);
    contact.sound = stone->collision_sound();
}

/* Enigma's stones have rounded corners; this leads to realistic
   behaviour when a marble rebounds from the corner of a single stone.
   But it's annyoing when there are two adjacent stones and a marble
   rebounds from any one (or even both) corners, because it changes
   the direction of the marble quite unpredictably.

   This function modifies the contact information for two adjacent
   stones in such a way that the two collisions are treated as a
   single collision with a flat surface.
 */
static void
maybe_join_contacts (StoneContact & a, StoneContact &b)
{
//    double maxd = 4.0/32;   // Max distance between joinable collisions

    if ((a.is_contact || b.is_contact)  &&
	a.is_collision && b.is_collision &&
        a.response==STONE_REBOUND && b.response==STONE_REBOUND) // &&
//        length(a.contact_point - b.contact_point) <= maxd)
    {
        if (a.is_contact && b.is_contact)
            b.fake_collision = true; // Don't rebound from `b'
        
        DirectionBits fa = contact_faces(a);
        DirectionBits fb = contact_faces(b);

        StoneContact &c = (a.is_contact) ? a : b;
        switch (fa & fb) {
        case NORTHBIT: c.normal = V3(0,-1,0); break;
        case EASTBIT: c.normal = V3(1,0,0); break;
        case SOUTHBIT: c.normal = V3(0,1,0); break;
        case WESTBIT: c.normal = V3(-1,0,0); break;
        case NODIRBIT:
            //fprintf(stderr, "Strange: contacts have no direction in common\n");
            break;
        default:
            //fprintf(stderr, "Strange: contacts have multiple directions in common\n");
            break;
        }
    }
}

static void
find_stone_contacts (Actor *a, StoneContactList &cl)
{
    ActorInfo &ai = *a->get_actorinfo();
    px::Rect r(int(ai.pos[0]-0.5), int(ai.pos[1]-0.5),
	       int(ai.pos[0]+0.5), int(ai.pos[1]+0.5));
    r.w -= r.x;
    r.h -= r.y;
    
    StoneContact contacts[2][2];
    
    // Test for collisions with the nearby stones
    for (int i=0; i<=r.w; i++)
        for (int j=0; j<=r.h; j++)
            find_contact_with_stone(a, contacts[i][j], r.x+i, r.y+j);

    maybe_join_contacts (contacts[0][0], contacts[1][0]);
    maybe_join_contacts (contacts[0][0], contacts[0][1]);
    maybe_join_contacts (contacts[1][0], contacts[1][1]);
    maybe_join_contacts (contacts[0][1], contacts[1][1]);

    for (int i=0; i<=r.w; i++)
        for (int j=0; j<=r.h; j++)
            if (contacts[i][j].is_contact)
                cl.push_back(contacts[i][j]);
}


/* This function is called for every contact between an actor and a
   stone.  It is here that the stones are informed about contacts.
   Note that a stone may kill itself during a call to actor_hit() or
   actor_contact() so we must not refer to it after having called
   these functions.  (Incidentally, this is why StoneContact only
   contains a 'stonepos' and no 'stone' entry.)
*/
static V3
handle_stone_contact(V3 normal, StoneContact &sc)
{
    Actor *a = sc.actor;
    ActorInfo *ai = a->get_actorinfo();

    if (sc.is_collision) {
        if (Stone *stone = world::GetStone(sc.stonepos))
            stone->actor_hit(sc);
        if (!sc.fake_collision && sc.response != STONE_PASS) {
            normal += sc.normal;
            normal /= length(normal);
            
            if (length(ai->vel) > 0.1) {
                AddSprite(display::SPRITE_EFFECT, sc.contact_point, "ring-anim");
                sound::PlaySound(sc.sound.c_str(), px::V2(sc.contact_point[0], 
                                                          sc.contact_point[1]));
            }
        }
    }
    else if (sc.is_contact) {
        if (Stone *stone = world::GetStone(sc.stonepos))
            stone->actor_contact(sc.actor);
    }
    return normal;
}

static V3 
reflect(V3 v, V3 normal)
{
    double n = length(normal);
    if (n != 0)
        normal /= n;
    return V3(v-2*(v*normal)*normal);
}

static void
handle_contacts(Actor *a)
{
    // Contacts with stones
    StoneContactList cl;
    find_stone_contacts(a, cl);
    V3 normal = accumulate(cl.begin(), cl.end(), V3(), handle_stone_contact);

    ActorInfo &ai = *a->get_actorinfo();

    ai.vel = reflect(ai.vel, normal);

    // Contacts with other actors
    unsigned s=level->actorlist.size();
    unsigned i;
    for (i=0; i<s && level->actorlist[i] != a; ++i)
        ;
    for (unsigned j=i+1; j<s; ++j)
    {
        ActorInfo &v = *level->actorlist[i]->get_actorinfo();
        ActorInfo &w = *level->actorlist[j]->get_actorinfo();

        V3 p1 = v.pos;
        V3 p2 = w.pos;
        double r1 = v.radius;
        double r2 = w.radius;

        bool contactp = length(p1-p2) < r1+r2;
        if (contactp)
        {
            V3 n = normalize(p1-p2); // normal to contact surface
            V3 v1 = n * (v.vel*n);
            V3 v2 = n * (w.vel*n);
            bool collisionp = (v2-v1)*(p2-p1) < 0;
            if (collisionp) {
                double m1 = v.mass;
                double m2 = w.mass;
                double M = m1+m2;
                
                v.vel += -v1 + (v2*(2*m2) + v1*(m1-m2)) / M;
                w.vel += -v2 + (v1*(2*m1) + v2*(m2-m1)) / M;
                sound::PlaySound("ballcollision");
            }
        }
    }
}

static void
move_actor(Actor *a)
{
//     if (Stone *st = GetStone(field))
//         st->actor_inside(a);
    a->move();
}

static void
tick_actor(Actor *a, double dtime)
{
    ActorInfo &ai = * a->get_actorinfo();
    ai.oldpos = ai.pos;
    ai.oldvel = ai.vel;

    double rest_time = dtime;
    double dt = 0.005; //dtime;

    handle_contacts(a);
    while (rest_time > 0 && !ai.grabbed)
    {
        advance_actor(a, dtime-rest_time, dt);
                
        double l = length(ai.pos-ai.oldpos);

        // time step too large? try again
        if (l > 0.04 && dt > 1/1000.) {
            dt /= 2;
            ai.pos = ai.oldpos;
            ai.vel = ai.oldvel;
            continue;
        }

        // ok, accept this step
        move_actor(a);
        handle_contacts(a);

        rest_time -= dt;
        ai.oldpos = ai.pos;
        ai.oldvel = ai.vel;

        // time step too small?
        if (l < 0.01) {
            dt *= 2;
            dt = dt > rest_time? rest_time : dt;
        }
    }
}

 // world:: global functions

static void
clear_world()
{
    delete level;
    level = 0;
    delete mouseforce;
    mouseforce=0;
}

void
world::Create(int w, int h)
{
    clear_world();
    level = new Level(w,h);
    mouseforce = new MouseForce;
    display::NewWorld(w, h);
    player::NewWorld();
    g_timer.clear();
}

void
world::Shutdown()
{
    clear_world();
}

bool
world::Load(const string &name)
{
    string filename("levels/");
    filename += name;
    filename += ".lua";


    // Restart the Lua environment, so symbol definitions from
    // different levels do not get in each other's way.
    lua::Shutdown();
    lua::Init();
    lua::Dofile("init.lua");


    if (lua::Dofile(filename) != 0) {
        fprintf(stderr, "Error while loading landscape '%s'\n",name.c_str());
        clear_world();
        return false;
    }
    laser::RecalcLight();
    laser::RecalcLightNow(); // recalculate laser beams if necessary

    for (ActorList::iterator i=level->actorlist.begin();
         i!=level->actorlist.end(); ++i)
    {
        int iplayer=-1;
        Actor *a=*i;
        if (a->int_attrib("player", &iplayer)) {
            player::AddActor(iplayer,a);
        }
    }

    return true;
}

void
world::SetMouseForce(V3 f)
{
    mouseforce->add_force(f);
}

void
world::NameObject(Object *obj, const std::string &name)
{
    named_objects.insert(name, obj); // [name] = obj;
}

void 
world::UnnameObject(Object *obj)
{
    assert(obj);
    string name;
    if (obj->string_attrib("name", &name)) {
        named_objects.remove(name);
        obj->set_attrib("name", "");
    }
}


Object *
world::GetNamedObject(const std::string &name)
{
    if (named_objects.find(name) != named_objects.end())
        return named_objects[name];
    else
        return 0;
}

//----------------------------------------
// Force fields
//----------------------------------------


void
world::AddForceField(ForceField *ff)
{
    level->forces.push_back(ff);
}

void
world::RemoveForceField(ForceField *ff)
{
    ForceList::iterator i=find(level->forces.begin(), level->forces.end(), ff);
    if (i != level->forces.end())
        level->forces.erase(i);
}

//----------------------------------------
// Rubber bands
//----------------------------------------

void
world::AddRubberBand (Actor *a, Stone *st, double strength,double length)
{
    level->rubbers.push_back(new RubberBand (a, st, strength, length));
}


void
world::AddRubberBand (Actor *a, Actor *a2, double strength,double length)
{
    level->rubbers.push_back(new RubberBand (a, a2, strength, length));
}

void
world::KillRubberBand (Actor *a, Stone *st)
{
    assert(a);
    for (unsigned i=0; i<level->rubbers.size(); ) {
        RubberBand &r = *level->rubbers[i];
        if (r.get_actor() == a && r.get_stone() != 0)
            if (r.get_stone()==st || st==0) {
                delete &r;
                level->rubbers.erase(level->rubbers.begin()+i);
                continue;       // don't increment i
            }
        ++i;
    }
}

void
world::KillRubberBand (Actor *a, Actor *a2)
{
    assert (a);
    for (unsigned i=0; i<level->rubbers.size(); ) {
        RubberBand &r = *level->rubbers[i];
        if (r.get_actor() == a && r.get_actor2() != 0)
            if (r.get_actor2()==a2 || a2==0) {
                delete &r;
                level->rubbers.erase(level->rubbers.begin()+i);
                continue;       // don't increment i
            }
        ++i;
    }
}

void 
world::KillRubberBands (Stone *st)
{
   for (unsigned i=0; i<level->rubbers.size(); ) {
        RubberBand &r = *level->rubbers[i];
        if (r.get_stone() != 0 && r.get_stone()==st) {
            delete &r;
            level->rubbers.erase(level->rubbers.begin()+i);
            continue;       // don't increment i
        }
        ++i;
    }
}

bool 
world::HasRubberBand (Actor *a, Stone *st)
{
    for (unsigned i=0; i<level->rubbers.size(); ++i) {
        RubberBand &r = *level->rubbers[i];
        if (r.get_actor() == a && r.get_stone() == st)
            return true;
    }
    return false;
}


namespace
{
//----------------------------------------
// Layer
//
// Changes to a layer are cached and only applied at the end of a
// tick. [### not implemented]
//----------------------------------------
    template <class T>
    class Layer {
        T *defaultval;
    public:
        Layer(T* deflt = 0) : defaultval(deflt) {}
        virtual ~Layer() {}

        T *get(GridPos p) {
            if (Field *f=level->get_field(p))
                return raw_get(*f);
            else 
                return defaultval;
        }

        T *yield(GridPos p) {
            if (Field *f=level->get_field(p)) {
                T *x = raw_get(*f);
                if (x) {
                    raw_set(*f, 0);
                    x->removal();
                }
                return x;
            } else
                return defaultval;
        }

        void set(GridPos p, T *x) {
            if (x) {
                if (Field *f=level->get_field(p)) {
                    dispose(raw_get(*f));
                    raw_set(*f, x);
                    x->creation(p);
                }
                else
                    dispose(x);
            }
        }
        void kill(GridPos p) { dispose(yield(p)); }

        virtual T* raw_get (Field &) = 0;
        virtual void raw_set (Field &, T *) = 0;

    private:
        virtual void dispose(T *x) { if (x) DisposeObject(x); }
    };

    /*
    ** Floor layer
    */
    class FloorLayer : public Layer<Floor> {
    private:
        Floor *raw_get (Field &f) { return f.floor; }
        void raw_set (Field &f, Floor *x) { f.floor = x;}
    };


    /*
    ** Item layer
    */
    class ItemLayer : public Layer<Item> {
    private:
        Item *raw_get (Field &f) { return f.item; }
        void raw_set (Field &f, Item *x) { f.item = x;}
    };

    /*
    ** Stone layer
    */
    class StoneLayer : public Layer<Stone> 
    {
    public:
        StoneLayer() : Layer<Stone>(&borderstone) {}
    private:
        Stone *raw_get (Field &f) { return f.stone; }
        void raw_set (Field &f, Stone *st) { f.stone = st;}
        void dispose (Stone *st) { 
            if (st) {
                KillRubberBands(st);
                DisposeObject(st); 
            }
        }

        /* This stone is used as the virtual border of the playing area.
           It is immovable and indestructible and makes sure the player's
           marble cannot leave the level. */
        class BorderStone : public Stone {
        public:
            BorderStone() : Stone("borderstone") {}
            Object *clone() { return this; }
            void dispose() {}
        };

        BorderStone borderstone;
    };

}

namespace
{
    FloorLayer      fl_layer;
    ItemLayer       it_layer;
    StoneLayer      st_layer;
    vector<GridPos> changed_stones;
}


//----------------------------------------
// Floor manipulation.
//----------------------------------------

void world::KillFloor(GridPos p) {fl_layer.kill(p);}
Floor *world::GetFloor(GridPos p) {return fl_layer.get(p);}
void world::SetFloor(GridPos p, Floor* st) {fl_layer.set(p,st);}

//----------------------------------------
// Stone manipulation.
//----------------------------------------

namespace
{
    void stone_change(GridPos p) {
        Stone *st = GetStone(p);
        if (Item *it = GetItem(p))
            it->stone_change(st);
        if (Floor *fl = GetFloor(p))
            fl->stone_change(st);
	laser::MaybeRecalcLight(p);
    }
}


void 
world::KillStone(GridPos p) 
{
    st_layer.kill(p);
//    stone_change(p);
    changed_stones.push_back(p);
}

Stone *
world::GetStone(GridPos p) {return st_layer.get(p);}

Stone *
world::YieldStone(GridPos p)
{
    Stone *st = st_layer.yield(p);
    laser::MaybeRecalcLight(p);
    changed_stones.push_back(p);
    return st;
}

void
world::SetStone(GridPos p, Stone* st)
{
    st_layer.set(p,st);
    laser::MaybeRecalcLight(p);
    changed_stones.push_back(p);
}

bool
world::MaybeMoveStone(GridPos p, Direction dir)
{
    GridPos newp = move(p,dir);
    if (GetStone(newp) == 0) 
    {
        Stone *st = GetStone(p);
        if (st && st->is_movable()) 
        {
            st->play_sound ("st-move");
            SetStone(newp, YieldStone(p));
            st->on_move();
            if (Item *it = GetItem(newp))
                it->on_stonehit(st);
            return true;
        }
    }
    return false;
}

void 
world::SwapStones(GridPos p, GridPos newp)
{
    Stone *tmp = st_layer.yield(newp);
    if (0 != strcmp(tmp->get_kind(), "borderstone")) {
        tmp->play_sound ("st-move");
        SetStone(newp, st_layer.yield(p));
        SetStone(p, tmp);
    }
}

//----------------------------------------
// Item manipulation.
//----------------------------------------

void world::KillItem(GridPos p) {
    laser::MaybeRecalcLight(p);
    it_layer.kill(p);
}
Item *world::GetItem(GridPos p) {return it_layer.get(p);}
Item *world::YieldItem(GridPos p) {
    laser::MaybeRecalcLight(p);
    return it_layer.yield(p);
}
void world::SetItem(GridPos p, Item* st) {
    laser::MaybeRecalcLight(p);
    it_layer.set(p,st);
}

//----------------------------------------
// Actor manipulation.
//----------------------------------------

void
world::AddActor(double x, double y, Actor* a)
{
    V3 pos(x,y,0);
    level->actorlist.push_back(a);
    a->get_actorinfo()->pos = pos;
    a->on_creation(pos);
    ReleaseActor(a);
}

Actor *
world::YieldActor(Actor *a)
{
    ActorList::iterator i=find(level->actorlist.begin(), level->actorlist.end(), a);
    if (i!=level->actorlist.end()) {
        level->actorlist.erase(i);
        GrabActor(a);
        return a;
    }
    return 0;
}

void 
world::WarpActor(Actor *a, double newx, double newy)
{
    V3 newpos = V3(newx, newy, 0);
    a->get_actorinfo()->vel = V3();
    a->warp(newpos);
}


void
world::GrabActor(Actor *a)
{
    a->get_actorinfo()->grabbed = true;
}

void
world::ReleaseActor(Actor *a)
{
    a->get_actorinfo()->grabbed = false;
}

const ActorInfo &
world::GetActorInfo(Actor *a)
{
    return *a->get_actorinfo();
}

void
world::Tick(double dtime)
{
    const double timestep = 15.0/1000;
    static double timeaccu = 0.0;

    timeaccu += dtime;
    
    while (timeaccu >= timestep) {
        for (ActorList::iterator i=level->actorlist.begin(); i!=level->actorlist.end(); ++i) {
            ActorInfo *ai = (*i)->get_actorinfo();
            if (!ai->grabbed && !(*i)->is_dead())
            {
                ai->forceacc2 = ai->forceacc;
                ai->forceacc = V3();
                tick_actor(*i, timestep);
            }
        }
        // Tell floors and items about new stones.
        for (unsigned i=0; i<changed_stones.size(); ++i)
            stone_change(changed_stones[i]);
        changed_stones.clear();


        mouseforce->tick(timestep);
        for_each(level->forces.begin(), level->forces.end(),
                 bind2nd(mem_fun(&ForceField::tick), timestep));

        for_each (level->rubbers.begin(), level->rubbers.end(),
                 bind2nd(mem_fun(&RubberBand::tick), timestep));

        g_timer.tick(timestep);
        laser::RecalcLightNow();   // recalculate laser beams if necessary
        timeaccu -= timestep;
        timeaccu = 0;
    }
}

void 
world::Init()
{
    actors::Init();
    laser::Init();
    items::Init();
}
